//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 46
// sandbox with simulated ball
template<class T> void Function_46(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    if(e.extra_T.m<=0) e.extra_T.Append(1);
    if(e.extra_T.m<=1) e.extra_T.Append(1);
    if(e.extra_T.m<=2) e.extra_T.Append(1);
    e.particles.Store_Fp(true);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    T sand_depth=0.35*e.m;
    T air_height=0.65*e.m;
    T sand_length=1*e.m;
    T sand_width=1*e.m;
    T wall_thickness=0.2*e.m;
    RANGE<TV> sand(TV(-sand_width/2,-sand_depth,-sand_length/2),TV(sand_width/2,0,sand_length/2));
    RANGE<TV> sand_box(sand.Thickened(wall_thickness));
    RANGE<TV> sand_box_interior(sand);
    sand_box_interior.max_corner(1)=sand_box.max_corner(1);
    RANGE<TV> domain(sand);
    domain.max_corner(1)=air_height;
    domain=domain.Thickened(0.05*e.m);
    e.grid.Initialize(IV(domain.Edge_Lengths()*e.resolution),domain);
    e.Set_Weights(e.order);
    auto sand_box_walls=
        new IMPLICIT_OBJECT_INTERSECTION<TV>(
                new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sand_box),
                new IMPLICIT_OBJECT_INVERT<TV>(
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sand_box_interior)));
    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(sand_box_walls);}
    else{
        e.Add_Collision_Object(sand_box_walls,MPM::BC_NOSLIP,0);}

    T density=(T)1582*e.unit_rho;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.3;

    e.Seed_Particles(sand,0,0,density,e.particles_per_cell,r0); 
    // SAND PROPERTIES
    ARRAY<int> sand_particles(e.particles.X.m);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    e.Add_Drucker_Prager(E,nu,(T)21,&sand_particles);
    e.Set_Lame_On_Particles(E,nu);
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    T g=9.81*e.m/(e.s*e.s);
    for(int i=0;i<e.particles.X.m;i++) (*color_attribute)(i)=VECTOR<T,3>(.8,.7,.7);
    {
        int N_sand=e.particles.number;
        VECTOR<T,3> velocity(TV(5,-5,0)*e.extra_T(0)*e.m/e.s);
        VECTOR<T,3> angular_velocity(TV(0,0,10)*e.extra_T(1)/e.s);
        T y0=0.15*e.m;
        T t_impact=(velocity(1)+sqrt(sqr(velocity(1))+2*y0*g))/g;
        SPHERE<TV> sphere(TV(-velocity(0)*t_impact,y0,-velocity(2)*t_impact),0.1*e.extra_T(2)*e.m);
        T density=11340*e.unit_rho*e.scale_mass;
        e.Seed_Particles(sphere,[=](const TV& X){return velocity+angular_velocity.Cross(X-sphere.center);},
                [=](const TV&){return MATRIX<T,3>::Cross_Product_Matrix(angular_velocity);}
                ,density,e.particles_per_cell,r0);
        int N_box_particles=e.particles.number-N_sand;
        LOG::cout<<N_sand<<" sand particles "<<N_box_particles<<" ball particles\n";
        ARRAY<int> ball_particles(N_box_particles);
        for(int k=0;k<ball_particles.m;k++) ball_particles(k)=k+N_sand;
        e.Add_Fixed_Corotated(16e7*e.unit_p*e.scale_E,0.44,&ball_particles);
    }
    e.Add_Gravity(TV(0,-g,0));

}
template void Function_46(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_46(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
