//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Implicit_Objects/ANALYTIC_IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_DILATE.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INTERSECTION.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_INVERT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_UNION.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 47
// sandbox with collision object ball
template<class T> void Function_47(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    REGION_ID r0(0);
    if(e.extra_T.m<=0) e.extra_T.Append(1);
    if(e.extra_T.m<=1) e.extra_T.Append(1);
    if(e.extra_T.m<=2) e.extra_T.Append(1);
    e.particles.Store_Fp(true);
    if(!e.no_implicit_plasticity) e.use_implicit_plasticity=true;
    T sand_depth=0.35*e.m;
    T air_height=0.65*e.m;
    T sand_length=e.extra_T(1)*e.m;
    T sand_width=e.extra_T(1)*e.m;
    T wall_thickness=0.2*e.m;
    T lip_height=0.05*e.m;
    RANGE<TV> sand(TV(-sand_width/2,-sand_depth,-sand_length/2),TV(sand_width/2,0,sand_length/2));
    RANGE<TV> domain(sand);
    domain.max_corner(1)=air_height;
    T dx=domain.Edge_Lengths().Min()/e.resolution;
    RANGE<TV> sand_box_interior(sand);
    RANGE<TV> sand_box(sand_box_interior.Thickened(wall_thickness));
    sand_box_interior.max_corner(1)=sand_box.max_corner(1)=lip_height;
    domain=domain.Thickened(0.5*e.m);
    RANGE<TV> outer_box(domain.Thickened(wall_thickness));
    e.grid.Initialize(IV(domain.Edge_Lengths()/dx),domain);
    e.Set_Weights(e.order);
    domain=domain.Thickened(-0.05*e.m);

    RANGE<TV> sides[6];
    for(int i=0;i<3;i++){
        sides[2*i]=sand_box.Thickened(-lip_height);
        sides[2*i+1]=sand_box.Thickened(-lip_height);
        sides[2*i].min_corner(i)=sand_box_interior.max_corner(i)+lip_height;
        sides[2*i+1].max_corner(i)=sand_box_interior.min_corner(i)-lip_height;}
    
    auto sand_box_walls=
        new IMPLICIT_OBJECT_DILATE<TV>(
                new IMPLICIT_OBJECT_UNION<TV>(
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sides[0]),
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sides[1]),
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sides[3]),
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sides[4]),
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(sides[5])),
                lip_height);
    auto outer_box_walls=
        new IMPLICIT_OBJECT_INTERSECTION<TV>(
                new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(outer_box),
                new IMPLICIT_OBJECT_INVERT<TV>(
                    new ANALYTIC_IMPLICIT_OBJECT<RANGE<TV> >(domain)));
    if(e.use_penalty_collisions){
        e.Add_Penalty_Collision_Object(sand_box_walls);
        e.Add_Penalty_Collision_Object(outer_box_walls);}
    else{
        e.Add_Collision_Object(sand_box_walls,MPM::BC_SLIP,0);
        e.Add_Collision_Object(outer_box_walls,MPM::BC_SLIP,0);}

    T y0=0.15*e.m;
    T v0=-1*e.m/e.s*e.extra_T(0);
    T g=9.81*e.m/(e.s*e.s);
    T radius=0.1*e.extra_T(2)*e.m;
    SPHERE<TV> sphere(TV(0,y0,0),radius);
    T stop_time=(v0+sqrt(sqr(v0)+2*(y0-radius+sand_depth-0.05*e.m)*g))/g;
    e.Add_Collision_Object(sphere,MPM::BC_SEP,e.friction,
            [=](T time){
                if(time>stop_time) time=stop_time;
                return FRAME<TV>(TV(0,v0*time-0.5*g*sqr(time),0));},
            [=](T time){
                if(time>=stop_time) return TWIST<TV>(TV(),typename TV::SPIN());
                return TWIST<TV>(TV(0,v0-g*time,0),typename TV::SPIN());});

    T density=(T)1582*e.unit_rho;
    T E=35.37e6*e.unit_p*e.scale_E,nu=.2;

    e.Seed_Particles(sand,0,0,density,e.particles_per_cell,r0); 
    // SAND PROPERTIES
    ARRAY<int> sand_particles(e.particles.X.m);
    for(int p=0;p<e.particles.X.m;p++) sand_particles(p)=p;
    e.Add_Drucker_Prager(E,nu,(T)21,&sand_particles);
    e.Set_Lame_On_Particles(E,nu);
    ARRAY_VIEW<VECTOR<T,3> >* color_attribute=e.particles.template Get_Array<VECTOR<T,3> >("color");
    for(int i=0;i<e.particles.X.m;i++) (*color_attribute)(i)=VECTOR<T,3>(.8,.7,.7);
    e.Add_Gravity(TV(0,-g,0));

}
template void Function_47(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_47(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
