//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Grid_Tools/Grids/CELL_ITERATOR.h>
#include <Geometry/Basic_Geometry/CYLINDER.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Basic_Geometry/PLANE.h>
#include <Geometry/Basic_Geometry/RANGE_TESSELLATION.h>
#include <Geometry/Basic_Geometry/RING.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Basic_Geometry/TORUS.h>
#include <Geometry/Meshes/TRIANGULATED_SURFACE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 87
// Falling objects into a box
template<class T> void Function_87(STANDARD_TESTS_BASE<VECTOR<T,3> >& e)
{
    typedef VECTOR<T,3> TV;
    typedef VECTOR<int,3> IV;
    e.reflection_bc=true;
    e.use_full_reflection=true;
    if(!e.extra_int.m) e.extra_int={3,3,3};
    if(!e.extra_T.m) e.extra_T={(T).2};

    PHYSBAM_ASSERT(e.extra_int(0)*e.extra_T(0)<=1); // No overlaps
    PHYSBAM_ASSERT(e.extra_int(2)*e.extra_T(0)<=1);

    T pole_radius=(T).2;
    T height_per_layer=(T)2/std::max(e.extra_int(0),e.extra_int(2));
    T top_of_stack=2+pole_radius+height_per_layer*e.extra_int(1);
    int iy=ceil(top_of_stack);

    e.Set_Grid(RANGE<TV>(TV(-1,0,-1),TV(1,iy,1))*e.m,IV(2,iy,2),IV(),2);
    T density=2*e.unit_rho*e.scale_mass;

    CYLINDER<T> pole(TV(-(T)1.1,2,.2),TV((T)1.1,2,.2),pole_radius);

    IMPLICIT_OBJECT<TV>* pole_io=Make_IO(pole);
    if(e.use_penalty_collisions) e.Add_Penalty_Collision_Object(pole_io);
    else e.Add_Collision_Object(pole_io,MPM::BC_SEP,.3);

    SPHERE<TV> sphere(TV(),e.m);
    RANGE<TV> box=RANGE<TV>::Centered_Box()/sqrt((T)3);
    TORUS<T> torus(TV(),TV(0,1,0),(T).3,(T).7);
    RING<T> ring(TV(0,-0.28,0),TV(0,0.28,0),(T).96,(T)0.4);
    CYLINDER<T> cylinder(TV(0,-.8,0),TV(0,.8,0),(T).6);
    LOG::printf("s %P  b %P  t %P  r %P  c %P\n",
        sphere.Bounding_Box().Edge_Lengths(),
        box.Bounding_Box().Edge_Lengths(),
        torus.Bounding_Box().Edge_Lengths(),
        ring.Bounding_Box().Edge_Lengths(),
        cylinder.Bounding_Box().Edge_Lengths());

    T rs=(T)e.resolution/128;
    int lg_rs=rint(log(rs)/log(2));
    int res_s=6+lg_rs;
    IV res_b=IV()+rint(32*rs);
    int res_t_i=rint(64*rs),res_t_o=rint(64*rs);
    int res_r_t=rint(80*rs),res_r_h=rint(10*rs),res_r_r=rint(10*rs);
    int res_c_h=rint(32*rs),res_c_t=rint(16*rs);

    auto save=[&e](auto* ts,std::string file)
    {
        e.viewer_dir.Make_Common_Directory();
        ts->Write_Obj(e.viewer_dir.Common_File(file+".obj"));
        delete ts;
    };

    save(Generate_Triangles(sphere,res_s),"sphere-orig");
    save(Generate_Triangles(box,res_b),"box-orig");
    save(Generate_Triangles(torus,res_t_i,res_t_o),"torus-orig");
    save(Generate_Triangles(ring,res_r_t,res_r_h,res_r_r),"ring-orig");
    save(Generate_Triangles(cylinder,res_c_h,res_c_t),"cylinder-orig");

    T scale=e.extra_T(0);
    sphere=Scaled(sphere,TV(),scale);
    box=Scaled(box,TV(),scale);
    torus=Scaled(torus,TV(),scale);
    ring=Scaled(ring,TV(),scale);
    cylinder=Scaled(cylinder,TV(),scale);

    save(Generate_Triangles(sphere,res_s),"sphere-scaled");
    save(Generate_Triangles(box,res_b),"box-scaled");
    save(Generate_Triangles(torus,res_t_i,res_t_o),"torus-scaled");
    save(Generate_Triangles(ring,res_r_t,res_r_h,res_r_r),"ring-scaled");
    save(Generate_Triangles(cylinder,res_c_h,res_c_t),"cylinder-scaled");

    RANGE<TV> seed_box=e.grid.domain;
    seed_box.min_corner.y=2*e.m+pole.radius;
    seed_box.max_corner.y=top_of_stack;

    REGION_ID region(0);
    GRID<TV> placement_grid(IV(e.extra_int),seed_box);
    for(CELL_ITERATOR<TV> it(placement_grid);it.Valid();it.Next())
    {
        FRAME<TV> f(it.Location());
        e.random.Fill_Uniform(f.r);
        TWIST<TV> t;
        e.random.Fill_Uniform(t.linear,-.1,.1);
        e.random.Fill_Uniform(t.angular,-.1,.1);
        int index=e.random.Get_Uniform_Integer(0,4);
        auto V=[t](const TV& X) {return t.linear+t.angular.Cross(X);};
        auto dV=[t](const TV& X) {return MATRIX<T,TV::m>::Cross_Product_Matrix(t.angular);};
        switch(index)
        {
            case 0:
                e.Seed_Particles_To_Analytic_Surface(Transformed(sphere,f),
                    V,dV,density,e.particles_per_cell,region++,true,res_s);
                break;
            case 1:
                e.Seed_Particles_To_Analytic_Surface(Transformed(box,f),
                    V,dV,density,e.particles_per_cell,region++,true,res_b);
                break;
            case 2:
                e.Seed_Particles_To_Analytic_Surface(Transformed(torus,f),
                    V,dV,density,e.particles_per_cell,region++,true,res_t_i,res_t_o);
                break;
            case 3:
                e.Seed_Particles_To_Analytic_Surface(Transformed(ring,f),
                    V,dV,density,e.particles_per_cell,region++,true,res_r_t,res_r_h,res_r_r);
                break;
            case 4:
                e.Seed_Particles_To_Analytic_Surface(Transformed(cylinder,f),
                    V,dV,density,e.particles_per_cell,region++,true,res_c_h,res_c_t);
                break;
        }
    }

#if 0            
    auto seed_io=Intersect(Make_IO(seed_box),Invert(Make_IO(pole)));
    OBJECT_PLACEMENT<TV> object_placement(random,seed_io);
    object_placement.max_seed_tries=1000;
    object_placement.max_intersect_tries=100;
    object_placement.min_sep=grid.dX(0);

    typename OBJECT_PLACEMENT<TV>::OBJECT o;
    o.scale={(T).2,(T).3};
    o.max_angular_vel=.1/s;
    o.velocity_range=RANGE<TV>::Centered_Box()*.1*m/s;

    o.box=sphere.Bounding_Box();
    object_placement.objects.Append(o);

    o.box=box.Bounding_Box();
    object_placement.objects.Append(o);

    o.box=torus.Bounding_Box();
    object_placement.objects.Append(o);

    o.box=ring.Bounding_Box();
    object_placement.objects.Append(o);

    o.box=cylinder.Bounding_Box();
    object_placement.objects.Append(o);

    ARRAY<typename OBJECT_PLACEMENT<TV>::SEEDING> objs;
    object_placement.Seed_Objects(objs,25);

    for(int i=0;i<objs.m;i++)
    {
        FRAME<TV> f=objs(i).frame;
        TWIST<TV> t=objs(i).twist;
        T s=objs(i).scale;
 
        auto V=[t](const TV& X) {return t.linear+t.angular.Cross(X);};
        auto dV=[t](const TV& X) {return MATRIX<T,TV::m>::Cross_Product_Matrix(t.angular);};
        switch(objs(i).index)
        {
            case 0:
                e.Seed_Particles_To_Analytic_Surface(Transformed(Scaled(sphere,TV(),s),f),V,dV,density,
                    particles_per_cell,true,5);
                break;
            case 1:
                e.Seed_Particles_To_Analytic_Surface(Transformed(Scaled(box,TV(),s),f),V,dV,density,
                    particles_per_cell,true,IV()+5);
                break;
            case 2:
                e.Seed_Particles_To_Analytic_Surface(Transformed(Scaled(torus,TV(),s),f),V,dV,density,
                    particles_per_cell,true,10,20);
                break;
            case 3:
                e.Seed_Particles_To_Analytic_Surface(Transformed(Scaled(ring,TV(),s),f),V,dV,density,
                    particles_per_cell,true,40,5,5);
                break;
            case 4:
                e.Seed_Particles_To_Analytic_Surface(Transformed(Scaled(cylinder,TV(),s),f),V,dV,density,
                    particles_per_cell,true,5);
                break;
        }
    }
#endif

    e.Add_Fixed_Corotated(1e3*e.unit_p*e.scale_E,0.3);
    e.Add_Gravity(e.m/(e.s*e.s)*TV(0,-9.8,0));
    e.Add_Walls(-1,MPM::BC_SEP,0,0*e.m,false);
}
template void Function_87(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_87(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
