//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include "STANDARD_TESTS_BASE.h"
#include <Core/Log/FINE_TIMER.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Tools/Parsing/PARSE_ARGS.h>
#include <Grid_Tools/Grids/CELL_ITERATOR.h>
#include <Grid_Tools/Grids/NODE_ITERATOR.h>
#include <Geometry/Basic_Geometry/SPHERE.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT.h>
#include <Geometry/Meshes/TETRAHEDRALIZED_VOLUME.h>
#include <Geometry/Meshes/TRIANGULATED_AREA.h>
#include <Geometry/Seeding/POISSON_DISK.h>
#include <Deformables/Constitutive_Models/COROTATED.h>
#include <Deformables/Constitutive_Models/COROTATED_FIXED.h>
#include <Deformables/Constitutive_Models/ISOTROPIC_CONSTITUTIVE_MODEL.h>
#include <Deformables/Constitutive_Models/NEO_HOOKEAN.h>
#include <Deformables/Constitutive_Models/ST_VENANT_KIRCHHOFF_HENCKY_STRAIN.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
#include <Solids/Solids/SOLID_BODY_COLLECTION.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Forces/MPM_DRUCKER_PRAGER.h>
#include <Hybrid_Methods/Forces/MPM_FINITE_ELEMENTS.h>
#include <Hybrid_Methods/Forces/MPM_GRAVITY.h>
#include <Hybrid_Methods/Forces/MPM_PLASTIC_FINITE_ELEMENTS.h>
#include <Hybrid_Methods/Forces/MPM_PLASTICITY_CLAMP.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
#include <Hybrid_Methods/Seeding/MPM_PARTICLE_SOURCE.h>
#ifdef USE_OPENMP
#include <omp.h>
#endif
namespace PhysBAM{
template<class TV> extern void
Collect_Tests(HASHTABLE<std::string,void(*)(STANDARD_TESTS_BASE<TV>&)>& test_list);
//#####################################################################
// Constructor
//#####################################################################
template<class TV> STANDARD_TESTS_BASE<TV>::
STANDARD_TESTS_BASE(const STREAM_TYPE stream_type_input,PARSE_ARGS& parse_args)
    :MPM_EXAMPLE<TV>(stream_type_input),test_name("<none>"),resolution(32),user_resolution(false),stored_last_frame(0),
    user_last_frame(false),order(2),seed(1234),particles_per_cell(1<<TV::m),regular_seeding(false),
    no_regular_seeding(false),scale_mass(1),scale_E(1),scale_speed(1),
    penalty_collisions_stiffness((T)1e4),penalty_collisions_separation((T)1e-4),penalty_collisions_length(1),
    penalty_damping_stiffness(0),use_penalty_collisions(false),use_plasticity(true),
    use_theta_c(false),use_theta_s(false),use_hardening_factor(false),use_max_hardening(false),
    theta_c(0),theta_s(0),hardening_factor(0),max_hardening(0),use_implicit_plasticity(false),no_implicit_plasticity(false),
    hardening_mast_case(0),use_hardening_mast_case(false),override_output_directory(false),
    m(1),s(1),kg(1),forced_collision_type(-1),friction(0),friction_is_set(false),sigma_Y(0),use_cohesion(false),
    dump_collision_objects(false),poisson_disk(*new POISSON_DISK<TV>(1)),
    tests(solid_body_collection.deformable_body_collection)
{
    T framerate=0;
    bool no_affine=false;
    bool use_separate=false,use_slip=false,use_stick=false;
    bool no_sound_speed_cfl=false,no_strong_cfl=false;
    parse_args.Extra(&test_name,"name","example name to run");
    parse_args.Add("-restart",&restart,"frame","restart frame");
    parse_args.Add("-resolution",&resolution,&user_resolution,"resolution","grid resolution");
    parse_args.Add("-substeps",&write_substeps_level,"level","output-substep level");
    parse_args.Add("-substeps_delay",&substeps_delay_frame,"frame","delay substeps until after this frame");
    parse_args.Add("-num_substeps",&num_substeps,"num_substeps","Set number of substeps per frame, fix dt for substeps");
    parse_args.Add("-last_frame",&last_frame,&user_last_frame,"frame","number of frames to simulate");
    parse_args.Add("-test_diff",&test_diff,"test analytic derivatives");
    parse_args.Add("-threads",&threads,"threads","Number of threads");
    parse_args.Add("-sph",&sph_rel,"sphere resolution","Sphere Resolution");
    parse_args.Add("-o",&viewer_dir.output_directory,&override_output_directory,"dir","Output directory");
    parse_args.Add("-d",&data_directory,"dir","data directory");
    parse_args.Add("-framerate",&framerate,"rate","Number of frames per second");
    parse_args.Add("-frame_dt",&frame_dt,"rate","Number of frames per second");
    parse_args.Add("-min_dt",&min_dt,"dt","Minimum time step size");
    parse_args.Add("-max_dt",&max_dt,"dt","Maximum time step size");
    parse_args.Add("-order",&order,"order","Interpolation basis order");
    parse_args.Add_Not("-no_affine",&opt.use_affine,"Use affine PIC");
    parse_args.Add("-affine",&opt.use_affine,"Use affine PIC");
    parse_args.Add("-compute_sound_speed",&compute_sound_speed,"Compute Sound Speed");
    parse_args.Add("-use_reflect",&reflection_bc,"Use reflection as boundary conditions");
    parse_args.Add("-midpoint",&opt.use_midpoint,"Use midpoint rule");
    parse_args.Add("-explicit",&use_symplectic_euler,"Use forward euler for grid update");
    parse_args.Add("-print_stats",&print_stats,"Print momentum/energy stats");
    parse_args.Add("-only_write_particles",&only_write_particles,"Only write particle data (ignore grid data, restart data etc)");
    parse_args.Add("-flip",&opt.flip,&no_affine,"frac","Flip ratio");
    parse_args.Add("-cfl",&cfl,"cfl","CFL number");
    parse_args.Add("-cfl_F",&cfl_F,"cfl","CFL number for F update");
    parse_args.Add("-cfl_c",&cfl_sound,"cfl","CFL number for sound speed");
    parse_args.Add("-use_cfl_F",&use_strong_cfl,"limit dt based on final change in X and F");
    parse_args.Add("-use_cfl_c",&use_sound_speed_cfl,"limit dt based on final change in X and F");
    parse_args.Add("-no_cfl_F",&no_strong_cfl,"Do not limit dt based on final change in X and F");
    parse_args.Add("-no_cfl_c",&no_sound_speed_cfl,"Do not limit dt based on final change in X and F");
    parse_args.Add("-newton_tolerance",&newton_tolerance,"tol","Newton tolerance");
    parse_args.Add("-newton_iterations",&newton_iterations,"iter","Newton iterations");
    parse_args.Add("-solver_tolerance",&solver_tolerance,"tol","Solver tolerance");
    parse_args.Add("-solver_iterations",&solver_iterations,"iter","Solver iterations");
    parse_args.Add("-seed",&seed,"seed","Random number seed");
    parse_args.Add("-particles_per_cell",&particles_per_cell,"num","Number of particles per cell");
    parse_args.Add("-scale_mass",&scale_mass,"scale","Scale mass of particles");
    parse_args.Add("-scale_E",&scale_E,"scale","Scale stiffness");
    parse_args.Add("-scale_speed",&scale_speed,"scale","Scale initial speed of simulated object");
    parse_args.Add("-penalty_stiffness",&penalty_collisions_stiffness,"tol","penalty collisions stiffness");
    parse_args.Add("-penalty_separation",&penalty_collisions_separation,"tol","penalty collisions separation");
    parse_args.Add("-penalty_length",&penalty_collisions_length,"tol","penalty collisions length scale");
    parse_args.Add("-penalty_damping",&penalty_damping_stiffness,"tol","penalty damping stiffness");
    parse_args.Add("-regular_seeding",&regular_seeding,"use regular particle seeding");
    parse_args.Add_Not("-no_regular_seeding",&no_regular_seeding,"use regular particle seeding");
    parse_args.Add("-use_exp_F",&opt.use_quad_F,"Use an approximation of the F update that prevents inversion");
    parse_args.Add("-use_plasticity",&use_plasticity,"Use plasticity in the F update");
    parse_args.Add("-theta_c",&theta_c,&use_theta_c,"theta_c","Critical compression coefficient for plasticity");
    parse_args.Add("-theta_s",&theta_s,&use_theta_s,"theta_s","Critical stretch coefficient for plasticity");
    parse_args.Add("-hardening",&hardening_factor,&use_hardening_factor,"hardening factor","Hardening factor for plasticity");
    parse_args.Add("-max_hardening",&max_hardening,&use_max_hardening,"max hardening coefficient","Maximum hardening coefficient for plasticity");
    parse_args.Add("-use_penalty_collisions",&use_penalty_collisions,"Use penalty collisions objects");
    parse_args.Add("-use_implicit_plasticity",&use_implicit_plasticity,"Use implicit plasticity");
    parse_args.Add("-no_implicit_plasticity",&no_implicit_plasticity,"Disable implicit plasticity");
    parse_args.Add("-mast_case",&hardening_mast_case,&use_hardening_mast_case,"mast_case","The case number from the Mast thesis for hardening parameters");
    parse_args.Add("-m",&m,"scale","meter scale");
    parse_args.Add("-s",&s,"scale","second scale");
    parse_args.Add("-kg",&kg,"scale","kilogram scale");
    parse_args.Add("-slip",&use_slip,"force slip collisions");
    parse_args.Add("-stick",&use_stick,"force stick collisions");
    parse_args.Add("-separate",&use_separate,"force separating collisions");
    parse_args.Add("-friction",&friction,&friction_is_set,"friction","Coefficient of friction"); // need changes in testcase
    parse_args.Add("-cohesion",&sigma_Y,&use_cohesion,"cohesion","sigma_Y in Drucker Prager");
    parse_args.Add_Not("-no_vc",&use_vc,"Use v_cp volume correction term for Drucker Prager");
    parse_args.Add("-T",&extra_T,"float","extra float argument");
    parse_args.Add("-I",&extra_int,"int","extra int argument");
    parse_args.Add("-dump_collisions",&dump_collision_objects,"dump out collision objects");
    parse_args.Add("-test_output_prefix",&test_output_prefix,&use_test_output,"","prefix to use for test output");
    parse_args.Add("-strong_cfl",&use_strong_cfl,"limit dt based on final change in X and F");
    parse_args.Add("-sound_cfl",&use_sound_speed_cfl,"limit dt based on soudn speed");
    parse_args.Add("-sound_cfl_async",&this->use_async_sound_speed_cfl,"limit dt based on soudn speed");
    parse_args.Add("-single_particle_cfl",&this->use_single_particle_cfl,"limit dt based on single particle stability");
    parse_args.Add("-cfl_p",&this->cfl_single_particle,"cfl","limit dt based on single particle stability");
    parse_args.Add("-reflection_bc",&reflection_bc_flags,"flags","Flags indicating which walls should be reflection BC");
    parse_args.Add("-test_c",&this->test_sound_speed,"test sound speed calculations");
    parse_args.Add("-dilation_only",&opt.dilation_only,"Discard non-volumetric portions of F");
    parse_args.Add("-use_reflect_friction",&this->use_full_reflection,"Use frictional reflection bc");
    parse_args.Add("-use_particle_collisions",&this->use_particle_collisions,"Use particle collisions");
    parse_args.Add("-sand_color",&use_sand_color,"Use sand color for rendering");
    parse_args.Add_Not("-no_affine_cfl",&this->use_affine_cfl,"No not use B_p for affine CFL.");
    parse_args.Add("-cpic",&opt.use_cpic,"Use Cp instead of grad-u.");
    parse_args.Add("-verbose_cfl",&this->verbose_cfl,"Be verbose about CFL computations.");
    parse_args.Add("-test_periodic",&use_periodic_test_shift,"test periodic bc");
    parse_args.Add("-damp_p_velocity",&this->set_damp_particle_velocity,"damp particle velocity on each step");
    parse_args.Add("-test_dt",&this->test_dt,"Print unique particle values, and calculate corresponding dt");
    parse_args.Add("-fix_particles_loc",&fix_particles_location,"Fix particles location");
    parse_args.Add("-suppress_output",&suppress_output,"Suppress Output");
    parse_args.Add_Not("-no_C_bound",&this->C_bound,"Disable C bound on grid velocity");
    parse_args.Add("-mu",&input_mu,"mu","Assign mu");
    parse_args.Add("-lam",&input_lambda,"lam","Assign lam");
    parse_args.Add("-E",&input_E,"E","Assign E"); // need changes in testcase
    parse_args.Add("-nu",&input_nu,"nu","Assign nu");
    parse_args.Add("-write_save_velocity",&this->write_save_velocity,"write velocity_save");
    parse_args.Parse();
    PHYSBAM_ASSERT((int)use_slip+(int)use_stick+(int)use_separate<=1);

    if(use_slip) forced_collision_type=MPM::BC_SLIP;
    if(use_stick) forced_collision_type=MPM::BC_NOSLIP;
    if(use_separate) forced_collision_type=MPM::BC_SEP;

    if(use_symplectic_euler)
    {
        if(!no_sound_speed_cfl) use_sound_speed_cfl=true;
        if(!no_strong_cfl) use_strong_cfl=true;
    }

    unit_p=kg*pow<2-TV::m>(m)/(s*s);
    unit_rho=kg*pow<-TV::m>(m);
    unit_mu=kg*pow<2-TV::m>(m)/s;
    min_dt*=s;
    max_dt*=s;
    penalty_collisions_separation*=m;
    penalty_collisions_length*=m;
    newton_tolerance*=kg*m/(s*s); // TODO: check me.

    if(this->use_async_sound_speed_cfl)
        use_sound_speed_cfl=true;
    
    if(no_affine) opt.use_affine=false;

    if(framerate) frame_dt=1/framerate;
    frame_dt*=s;
    if(num_substeps!=0)
    {
        max_dt=frame_dt/num_substeps;
        min_dt=frame_dt/num_substeps;
        LOG::printf("maxdt %P framedt %P\n",max_dt,frame_dt);
    }

    opt.use_quad_F=false;

#ifdef USE_OPENMP

    omp_set_num_threads(threads);
#pragma omp parallel
#pragma omp single
    {
        if(omp_get_num_threads()!=threads) PHYSBAM_FATAL_ERROR();
        LOG::cout<<"Running on "<<threads<<" threads"<<std::endl;
    }
#else
    PHYSBAM_ASSERT(threads==1);
#endif

    gather_scatter.threads=threads;
    stored_last_frame=last_frame;
    random.Set_Seed(seed);

    particles.Store_Fp(use_plasticity);
    particles.Store_C(opt.use_affine);
    this->write_output_files.Append([this]()
        {
            FILE_OSTREAM output;
            Safe_Open_Output(output,stream_type,viewer_dir.Frame_File("restart_data_stb"));
            Write_Binary(output,random);
            for(auto p:sources) 
                Write_Binary(output,*p);
        });
    this->read_output_files.Append([this]()
        {
            FILE_ISTREAM input;
            Safe_Open_Input(input,viewer_dir.Frame_File("restart_data_stb"));
            Read_Binary(input,random);
            for(auto p:sources) 
                Read_Binary(input,*p);
        });

    if(use_periodic_test_shift){
        auto shift_func=[this](int sign){
                TV shift=TV(sign*this->periodic_test_shift)*grid.dX;
                for(int i=0;i<particles.X.m;i++)
                    particles.X(i)=wrap(particles.X(i)+shift,grid.domain.min_corner,grid.domain.max_corner);};
        Add_Callbacks(true,"time-step",[=](){shift_func(1);});
        Add_Callbacks(false,"time-step",[=](){shift_func(-1);});
        Add_Callbacks(false,"initialize",[this](){
                RANDOM_NUMBERS<T> local_random;
                for(int i=0;i<TV::m;i++)
                    periodic_test_shift(i)=local_random.Get_Uniform_Integer(0,grid.numbers_of_cells(i)-1);
                LOG::printf("shift %P\n",periodic_test_shift);});}

    default_region_colors.Append({1,1,1});
    default_region_colors.Append({1,0,0});
    default_region_colors.Append({1,1,0});
    default_region_colors.Append({0,1,0});
    default_region_colors.Append({0,1,1});
    default_region_colors.Append({0,0,1});
    default_region_colors.Append({1,0,1});
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> STANDARD_TESTS_BASE<TV>::
~STANDARD_TESTS_BASE()
{
    if(destroy) destroy();
    for(auto p:sources)
    {
        delete p->io;
        delete p;
    }
    delete &poisson_disk;
}
//#####################################################################
// Function Initialize
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Initialize()
{
    Collect_Tests(test_list);
    auto* func=test_list.Get_Pointer(test_name);
    if(!func)
    {
        ARRAY<std::string> tests;
        test_list.Get_Keys(tests);
        tests.Sort();
        LOG::printf("No test case '%P' exists.  Existing tests: %P\n",test_name,tests);
        PHYSBAM_FATAL_ERROR("");
    }

    (*func)(*this);

    // TODO add argument -test_dt
    if(this->test_dt) Test_Dt();

    for(auto* s:solid_body_collection.deformable_body_collection.structures)
        s->Update_Number_Nodes();
    if(forced_collision_type!=-1)
        for(int i=0;i<collision_objects.m;i++)
            collision_objects(i)->type=(MPM::BC_TYPE)forced_collision_type;
}
//#####################################################################
// Function Test_Dt
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Test_Dt()
{
    // don't use threads can mess up with this
    HASHTABLE<std::string> h;
    for(int i=0; i<particles.number; i++)
        h.Set(LOG::sprintf(
                "TESTPRINT D %P order %P %P m %.16P volume %.16P mu %.16P lambda %.16P dx %P F %P velocity %P C %P\n",
                TV::m,order,opt.use_affine?"apic":"pic",particles.mass(i),
                particles.volume(i),particles.mu0(i),particles.lambda0(i),
                grid.dX(0),particles.F(i),particles.V(i),
                opt.use_affine?particles.C(i):MATRIX<T,TV::m>()));
    ARRAY<std::string> a;
    h.Get_Keys(a);
    a.Sort();
    for(const auto&s:a) LOG::cout<<s;
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles_Poisson(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,T particles_per_cell,
    REGION_ID r,ARRAY_VIEW<const TV> seeds,bool add_seeds)
{
    ARRAY<TV> X;
    if(seeds.m) X=seeds;
    poisson_disk.Sample(random,object,X);

    T volume=grid.dX.Product()/particles_per_cell;
    T mass=density*volume;
    int start=add_seeds?0:seeds.m;
    for(int i=start;i<X.m;i++)
        Add_Particle(X(i),V,dV,mass,volume,r);
    LOG::printf("Seeded %P particles; total: %P\n",X.m-start,particles.X.m);
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles_Uniform(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
    const GRID<TV>& seed_grid,REGION_ID r)
{
    int num_particles=particles.X.m;
    T volume=seed_grid.dX.Product();
    T mass=density*volume;
    for(CELL_ITERATOR<TV> it(seed_grid);it.Valid();it.Next()){
        TV X=it.Location();
        if(object.Lazy_Inside(X))
            Add_Particle(X,V,dV,mass,volume,r);}
    LOG::printf("Seeded %P particles; total: %P\n",particles.X.m-num_particles,particles.X.m);
}
//#####################################################################
// Function Seed_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Seed_Particles(IMPLICIT_OBJECT<TV>& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,
    T particles_per_cell,REGION_ID r)
{
    if(!regular_seeding)
        return Seed_Particles_Poisson(object,V,dV,density,particles_per_cell,r);

    object.Update_Box();
    RANGE<IV> range=grid.Clamp_To_Cell(object.Box(),3).Intersect(grid.Cell_Indices());
    TV LB=grid.Node_X(range.min_corner);
    TV UB=grid.Node_X(range.max_corner);
    T scale=pow<1,TV::m>((T)particles_per_cell);
    GRID<TV> seed_grid(IV(TV(range.Edge_Lengths())*scale),RANGE<TV>(LB,UB));
    Seed_Particles_Uniform(object,V,dV,density,seed_grid,r);
}
//#####################################################################
// Function Perturb
//#####################################################################
template<class TV> auto STANDARD_TESTS_BASE<TV>::
Perturb(T a) -> T
{
    return random.Get_Uniform_Number(1-a,1+a);
}
//#####################################################################
// Function Uniform
//#####################################################################
template<class TV> auto STANDARD_TESTS_BASE<TV>::
Uniform(T a,T b) -> T
{
    return random.Get_Uniform_Number(a,b);
}
//#####################################################################
// Function Add_Particle
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Particle(const TV& X,const TV& V,const MATRIX<T,TV::m>& dV,const T mass,const T volume,REGION_ID r)
{
    int p=particles.Add_Element();
    particles.valid(p)=true;
    particles.X(p)=X;
    particles.V(p)=V;
    particles.F(p)=MATRIX<T,TV::m>()+1;
    if(particles.store_Fp) particles.Fp(p).Set_Identity_Matrix();
    if(particles.store_C) particles.C(p)=dV;
    if(particles.store_S) particles.S(p)=SYMMETRIC_MATRIX<T,TV::m>()+1;
    particles.mass(p)=mass;
    particles.volume(p)=volume;
    particles.region(p)=Value(r);

    if(use_sand_color)
    {
        T color=random.Get_Number();
        if(color<(T)0.85) particles.color(p)={1,(T).66,(T).37};
        else if(color<(T)0.95) particles.color(p)={(T).42,(T).33,(T).12};
        else particles.color(p)={1,1,1};
    }
    else if(default_region_colors.m)
        particles.color(p)=default_region_colors(Value(r)%default_region_colors.m);
    else particles.color(p)={1,1,1};
    return p;
}
//#####################################################################
// Function Add_Particle
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Particle(const TV& X,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,
    const T mass,const T volume,REGION_ID r)
{
    return Add_Particle(X,V?V(X):TV(),dV?dV(X):MATRIX<T,TV::m>(),mass,volume,r);
}
//#####################################################################
// Function Add_Lambda_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Add_Lambda_Particles(ARRAY<int>* affected_particles,T E,T nu,T density,REGION_ID r,bool no_mu,T porosity,T saturation_level)
{
    ARRAY<int> lambda_particles(affected_particles->m);
    T volume_lambda=particles.volume(0)*porosity*saturation_level;
    T mass_lambda=density*volume_lambda;
    T lambda=E*nu/((1+nu)*(1-2*nu));
    for(int k=0;k<affected_particles->m;k++){
        int p=particles.Add_Element();
        lambda_particles(k)=p;
        int i=(*affected_particles)(k);
        particles.valid(p)=true;
        particles.X(p)=particles.X(i);
        particles.V(p)=particles.V(i);
        particles.F(p)=particles.F(i);
        if(particles.store_Fp) particles.Fp(p)=particles.Fp(i); 
        if(particles.store_C) particles.C(p)=particles.C(i);
        if(particles.store_S) particles.S(p)=particles.S(i);
        particles.mass(p)=mass_lambda;
        particles.volume(p)=volume_lambda;
        particles.mu(p)=(T)0;
        particles.mu0(p)=(T)0;
        particles.lambda(p)=lambda;
        particles.lambda0(p)=lambda;
        particles.color(p)=VECTOR<T,3>(1,(T).5,0);}
    Add_Fixed_Corotated(E,nu,&lambda_particles,no_mu);
}
//#####################################################################
// Function Add_Gravity
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Gravity(TV g,ARRAY<int>* affected_particles)
{
    return Add_Force(*new MPM_GRAVITY<TV>(force_helper,g,gather_scatter,affected_particles));
}
//#####################################################################
// Function Add_Fixed_Corotated
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Fixed_Corotated(T E,T nu,ARRAY<int>* affected_particles,bool no_mu)
{
    COROTATED_FIXED<T,TV::m>* coro=new COROTATED_FIXED<T,TV::m>(E,nu);
    if(no_mu){nu=0;coro->Zero_Out_Mu();}
    ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& constitutive_model=*coro;
    MPM_FINITE_ELEMENTS<TV>& fe=*new MPM_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles);
    Set_Lame_On_Particles(E,nu,affected_particles);
    return Add_Force(fe);
}
//#####################################################################
// Function Add_Corotated
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Corotated(T E,T nu,ARRAY<int>* affected_particles,bool no_mu)
{
    COROTATED<T,TV::m>* coro=new COROTATED<T,TV::m>(E,nu);
    if(no_mu){nu=0;coro->constant_mu=0;coro->constant_beta=0;}
    ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& constitutive_model=*coro;
    MPM_FINITE_ELEMENTS<TV>& fe=*new MPM_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles);
    Set_Lame_On_Particles(E,nu,affected_particles);
    return Add_Force(fe);
}
//#####################################################################
// Function Add_Neo_Hookean
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Neo_Hookean(T E,T nu,ARRAY<int>* affected_particles)
{
    ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& constitutive_model=*new NEO_HOOKEAN<T,TV::m>(E,nu);
    MPM_FINITE_ELEMENTS<TV>& fe=*new MPM_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles);
    Set_Lame_On_Particles(E,nu,affected_particles);
    return Add_Force(fe);
}
//#####################################################################
// Function Add_St_Venant_Kirchhoff_Hencky_Strain
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_St_Venant_Kirchhoff_Hencky_Strain(T E,T nu,
    ARRAY<int>* affected_particles,bool no_mu)
{
    ST_VENANT_KIRCHHOFF_HENCKY_STRAIN<T,TV::m>* hencky=new ST_VENANT_KIRCHHOFF_HENCKY_STRAIN<T,TV::m>(E,nu);
    if(no_mu){nu=0;hencky->Zero_Out_Mu();}
    ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& constitutive_model=*hencky;
    MPM_FINITE_ELEMENTS<TV>& fe=*new MPM_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles);
    Set_Lame_On_Particles(E,nu,affected_particles);
    return Add_Force(fe);
}
//#####################################################################
// Function Add_Drucker_Prager
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Drucker_Prager(T E,T nu,T a0,T a1,T a3,T a4,
    ARRAY<int>* affected_particles,bool no_mu,T sigma_Y)
{
    ST_VENANT_KIRCHHOFF_HENCKY_STRAIN<T,TV::m>* hencky=new ST_VENANT_KIRCHHOFF_HENCKY_STRAIN<T,TV::m>(E,nu);
    if(no_mu){nu=0;hencky->Zero_Out_Mu();}
    ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& constitutive_model=*hencky;
    MPM_DRUCKER_PRAGER<TV>* plasticity=new MPM_DRUCKER_PRAGER<TV>(particles,0,a0,a1,a3,a4);
    plasticity->use_implicit=(use_implicit_plasticity&&!no_implicit_plasticity);
    plasticity->use_vc=use_vc;
    PHYSBAM_ASSERT(!plasticity->use_implicit || !use_symplectic_euler);
    PARTICLE_GRID_FORCES<TV>* fe=0;
    if(plasticity->use_implicit){
        fe=new MPM_PLASTIC_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles,*plasticity);
        this->asymmetric_system=true;}
    else{
        MPM_FINITE_ELEMENTS<TV>* mfe=new MPM_FINITE_ELEMENTS<TV>(force_helper,constitutive_model,gather_scatter,affected_particles);
        plasticity->gather_scatter=&mfe->gather_scatter;
        fe=mfe;}
    plasticity->Initialize_Particles(affected_particles,sigma_Y);
    plasticity_models.Append(plasticity);
    Set_Lame_On_Particles(E,nu,affected_particles);
    update_dp_func=[=](const ARRAY<int>& a){plasticity->Initialize_Particles(&a);};
    return Add_Force(*fe);
}
//#####################################################################
// Function Add_Drucker_Prager
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Drucker_Prager(T E,T nu,T phi_F,
    ARRAY<int>* affected_particles,bool no_mu,T sigma_Y)
{
    return Add_Drucker_Prager(E,nu,phi_F,0,0,0,affected_particles,no_mu,sigma_Y);
}
//#####################################################################
// Function Add_Drucker_Prager
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Drucker_Prager_Case(T E,T nu,int case_num,
    ARRAY<int>* affected_particles,bool no_mu)
{
    ARRAY<ARRAY<T>> mast_constants({
        {35,0,0.2,10},
        {35,4,0.29,10},
        {35,9,0.3,10},
        {35,13,0.27,10},
        {35,0,0.2,6.57},
        {35,0,0.2,3.33},
        {35,0,0.2,0},
        {38.33,0,0.2,13.33},
        {41.67,0,0.2,16.67},
        {45,0,0.2,20}});
    ARRAY<T>& a=mast_constants(case_num);
    return Add_Drucker_Prager(E,nu,a(0),a(1),a(2),a(3),affected_particles,no_mu);
}
//#####################################################################
// Function Add_Clamped_Plasticity
//#####################################################################
template<class TV> int STANDARD_TESTS_BASE<TV>::
Add_Clamped_Plasticity(ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>& icm,T theta_c,
    T theta_s,T max_hardening,T hardening_factor,
    ARRAY<int>* affected_particles)
{
    MPM_PLASTICITY_CLAMP<TV>& plasticity=*new MPM_PLASTICITY_CLAMP<TV>(particles,0,theta_c,theta_s,
        max_hardening,hardening_factor);
    plasticity.use_implicit=(use_implicit_plasticity&&!no_implicit_plasticity);
    PHYSBAM_ASSERT(!plasticity.use_implicit || !use_symplectic_euler);
    PARTICLE_GRID_FORCES<TV>* fe=0;
    if(plasticity.use_implicit){
        fe=new MPM_PLASTIC_FINITE_ELEMENTS<TV>(force_helper,icm,gather_scatter,affected_particles,plasticity);
        this->asymmetric_system=true;}
    else{
        MPM_FINITE_ELEMENTS<TV>* mfe=new MPM_FINITE_ELEMENTS<TV>(force_helper,icm,gather_scatter,affected_particles);
        plasticity.gather_scatter=&mfe->gather_scatter;
        fe=mfe;}
    plasticity.Initialize_Particles(affected_particles);
    plasticity_models.Append(&plasticity);
    return Add_Force(*fe);
}
//#####################################################################
// Function Add_Walls
//#####################################################################
// -x +x -y +y [ -z +z ], as bit flags
template<class TV> void STANDARD_TESTS_BASE<TV>::
Add_Walls(int flags,MPM::BC_TYPE type,T friction,T inset,bool penalty)
{
    if(forced_collision_type!=-1) type=(MPM::BC_TYPE)forced_collision_type;
    if(friction_is_set) this->friction=friction;
    if(reflection_bc) this->reflection_bc_friction=friction;
    RANGE<TV> range=grid.domain.Thickened(grid.dX*(ghost*2+1));
    for(int a=0;a<TV::m;a++)
        for(int s=0;s<2;s++)
            if(flags&(1<<(a*2+s))){
                if(reflection_bc) side_bc_type(2*a+s)=type;
                else{
                    RANGE<TV> wall=range;
                    if(s)wall.max_corner(a)=grid.domain.min_corner(a)+inset;
                    else wall.min_corner(a)=grid.domain.max_corner(a)-inset;
                    if(!reflection_bc){
                        if(penalty) Add_Penalty_Collision_Object(wall);
                        else Add_Collision_Object(wall,type,friction);}}}
}
//#####################################################################
// Function Seed_Lagrangian_Particles
//#####################################################################
template<class TV> template<class T_STRUCTURE> T_STRUCTURE& STANDARD_TESTS_BASE<TV>::
Seed_Lagrangian_Particles(T_STRUCTURE& object,std::function<TV(const TV&)> V,
    std::function<MATRIX<T,TV::m>(const TV&)> dV,T density,REGION_ID r,
    bool use_constant_mass,bool destroy_after)
{
    int old_particles_number=particles.number;
    T_STRUCTURE& new_object=tests.Copy_And_Add_Structure(object,0,destroy_after);
    tests.Set_Mass_Of_Particles(new_object,density,use_constant_mass);
    for(int p=old_particles_number;p<particles.number;p++){
        TV X=particles.X(p);
        particles.valid(p)=true;
        if(V) particles.V(p)=V(X);
        if(particles.store_C && dV) particles.C(p)=dV(X);
        particles.F(p)=MATRIX<T,TV::m>()+1;
        if(particles.store_Fp) particles.Fp(p).Set_Identity_Matrix();
        if(particles.store_S) particles.S(p)=SYMMETRIC_MATRIX<T,TV::m>()+1;
        particles.volume(p)=particles.mass(p)/density;
        particles.region(p)=Value(r);
        if(default_region_colors.m)
            particles.color(p)=default_region_colors(Value(r)%default_region_colors.m);
        else particles.color(p)={1,1,1};
    }
    LOG::printf("seeded %i lagrangian particles\n",particles.number-old_particles_number);
    return new_object;
}
//#####################################################################
// Function Add_Penalty_Collision_Object
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Add_Penalty_Collision_Object(IMPLICIT_OBJECT<TV>* io,const T coefficient_of_friction)
{
    PHYSBAM_FATAL_ERROR();
}
//#####################################################################
// Function Set_Lame_On_Particles
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Set_Lame_On_Particles(T E,T nu,ARRAY<int>* affected_particles)
{
    particles.Store_Lame(true);
    T mu=E/(2*(1+nu));
    T lambda=E*nu/((1+nu)*(1-2*nu));
    if(input_mu!=0)mu=input_mu;
    if(input_lambda!=0)lambda=input_lambda;
    if(affected_particles)
#pragma omp parallel for
        for(int k=0;k<affected_particles->m;k++){
            int p=(*affected_particles)(k);
            particles.mu(p)=mu;
            particles.mu0(p)=mu;
            particles.lambda(p)=lambda;
            particles.lambda0(p)=lambda;}
    else{
        particles.mu.Fill(mu);
        particles.mu0.Fill(mu);
        particles.lambda.Fill(lambda);
        particles.lambda0.Fill(lambda);}
}
//#####################################################################
// Function Set_Grid
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Set_Grid(const RANGE<TV>& domain,IV resolution_scale,int default_resolution)
{
    if(!user_resolution) resolution=default_resolution;
    grid.Initialize(resolution_scale*resolution,domain);
    LOG::printf("grid %P\n",grid);
    Set_Weights(order);
    poisson_disk.Set_Distance_By_Volume(grid.dX.Product()/particles_per_cell);
}
//#####################################################################
// Function Set_Grid
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Set_Grid(const RANGE<TV>& domain,IV resolution_scale,IV resolution_padding,
    int resolution_multiple,int default_resolution)
{
    if(!user_resolution) resolution=default_resolution;
    int scaled_resolution=(resolution+resolution_multiple-1)/resolution_multiple;
    grid.Initialize(resolution_scale*scaled_resolution+resolution_padding,domain);
    LOG::printf("grid %P\n",grid);
    Set_Weights(order);
    poisson_disk.Set_Distance_By_Volume(grid.dX.Product()/particles_per_cell);
}
//#####################################################################
// Function Add_Source
//#####################################################################
template<class TV> void STANDARD_TESTS_BASE<TV>::
Add_Source(const TV& source_location,const TV& source_normal,
    T source_radius,T source_speed,const TV& gravity,T density,T E,T nu,
    T start_time,T stop_time,REGION_ID r)
{
    T volume=grid.dX.Product()/particles_per_cell;
    T mass=density*volume;
    SPHERE<TV> seed_volume(source_location,source_radius);
    MPM_PARTICLE_SOURCE<TV>* source=new MPM_PARTICLE_SOURCE<TV>(poisson_disk,
        random,source_location,source_normal,source_normal*source_speed,
        gravity,Make_IO(seed_volume));
    T mu=E/(2*(1+nu));
    T lambda=E*nu/((1+nu)*(1-2*nu));
    if(input_mu!=0)mu=input_mu;
    if(input_lambda!=0)lambda=input_lambda;
    if(no_mu) mu=0;
    sources.Append(source);
    Add_Callbacks(false,"time-step",[=,this]()
        {
            TIMER_SCOPE("Add_Source callback");
            if(time<start_time || time>=stop_time) return;
            ARRAY<TV> X,V;
            ARRAY<MATRIX<T,TV::m> > dV;
            ARRAY<int> affected_particles;
            source->Seed(dt,X,V,&dV);
            for(int i=0;i<X.m;i++)
            {
                int p=Add_Particle(X(i),V(i),dV(i),mass,volume,r);
                particles.mu(p)=mu;
                particles.mu0(p)=mu;
                particles.lambda(p)=lambda;
                particles.lambda0(p)=lambda;
                affected_particles.Append(p);
            }
            LOG::printf("Source add %P particles\n",X.m);
            if(update_dp_func) update_dp_func(affected_particles);
        });
}

template class STANDARD_TESTS_BASE<VECTOR<float,2> >;
template class STANDARD_TESTS_BASE<VECTOR<float,3> >;
template class STANDARD_TESTS_BASE<VECTOR<double,2> >;
template class STANDARD_TESTS_BASE<VECTOR<double,3> >;
template MESH_OBJECT<VECTOR<double,2>,1>& STANDARD_TESTS_BASE<VECTOR<double,2> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<double,2>,1> >(MESH_OBJECT<VECTOR<double,2>,1>&,std::function<VECTOR<double,2> (VECTOR<double,2> const&)>,std::function<MATRIX<double,2,2> (VECTOR<double,2> const&)>,double,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<double,2>,2>& STANDARD_TESTS_BASE<VECTOR<double,2> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<double,2>,2> >(MESH_OBJECT<VECTOR<double,2>,2>&,std::function<VECTOR<double,2> (VECTOR<double,2> const&)>,std::function<MATRIX<double,2,2> (VECTOR<double,2> const&)>,double,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<double,3>,2>& STANDARD_TESTS_BASE<VECTOR<double,3> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<double,3>,2> >(MESH_OBJECT<VECTOR<double,3>,2>&,std::function<VECTOR<double,3> (VECTOR<double,3> const&)>,std::function<MATRIX<double,3,3> (VECTOR<double,3> const&)>,double,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<double,3>,3>& STANDARD_TESTS_BASE<VECTOR<double,3> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<double,3>,3> >(MESH_OBJECT<VECTOR<double,3>,3>&,std::function<VECTOR<double,3> (VECTOR<double,3> const&)>,std::function<MATRIX<double,3,3> (VECTOR<double,3> const&)>,double,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<float,2>,1>& STANDARD_TESTS_BASE<VECTOR<float,2> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<float,2>,1> >(MESH_OBJECT<VECTOR<float,2>,1>&,std::function<VECTOR<float,2> (VECTOR<float,2> const&)>,std::function<MATRIX<float,2,2> (VECTOR<float,2> const&)>,float,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<float,2>,2>& STANDARD_TESTS_BASE<VECTOR<float,2> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<float,2>,2> >(MESH_OBJECT<VECTOR<float,2>,2>&,std::function<VECTOR<float,2> (VECTOR<float,2> const&)>,std::function<MATRIX<float,2,2> (VECTOR<float,2> const&)>,float,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<float,3>,2>& STANDARD_TESTS_BASE<VECTOR<float,3> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<float,3>,2> >(MESH_OBJECT<VECTOR<float,3>,2>&,std::function<VECTOR<float,3> (VECTOR<float,3> const&)>,std::function<MATRIX<float,3,3> (VECTOR<float,3> const&)>,float,REGION_ID,bool,bool);
template MESH_OBJECT<VECTOR<float,3>,3>& STANDARD_TESTS_BASE<VECTOR<float,3> >::Seed_Lagrangian_Particles<MESH_OBJECT<VECTOR<float,3>,3> >(MESH_OBJECT<VECTOR<float,3>,3>&,std::function<VECTOR<float,3> (VECTOR<float,3> const&)>,std::function<MATRIX<float,3,3> (VECTOR<float,3> const&)>,float,REGION_ID,bool,bool);
}
