//#####################################################################
// Copyright 2021, Song Bai, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Data_Structures/UNION_FIND.h>
#include <Core/Random_Numbers/RANDOM_NUMBERS.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_ITERATOR.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
#include "../STANDARD_TESTS_BASE.h"
namespace PhysBAM{
// TEST-NAME 131explore
    // running command ./mpm $dim $affine -explicit -last_frame $lastframe -order $spline 131explore -test_dt -min_dt $dt -max_dt $dt -frame_dt $dt -print_stats -T $seed
template<class TV> void Function_131explore(STANDARD_TESTS_BASE<TV>& e)
{
    typedef typename TV::SCALAR T;
    typedef VECTOR<int,TV::m> IV;
    REGION_ID r0(0);
    // seed test for random position and random number from 1-8
    int seed=0;
    int max_particle_num=8;
    if(e.extra_T.m>0) seed=e.extra_T(0);
    if(e.extra_T.m>1) max_particle_num=(int)e.extra_T(1);
    LOG::printf("maxparticlenum %P \n",max_particle_num);
    bool exit_early=false;
    if(e.extra_int.m>0) exit_early=e.extra_int(0);

    e.seed=seed;
    e.random.Set_Seed(seed);
    int num_particles=e.random.Get_Uniform_Integer(1,max_particle_num);

    e.Set_Grid(RANGE<TV>::Centered_Box()*e.m);
    T density=1010*e.unit_rho*e.scale_mass;
    T volume=e.grid.dX.Product()/e.particles_per_cell;
    T mass=density*volume;
    LOG::printf("test num particles %P density %P mass %P volume %P\n",num_particles,density,mass,volume);
    for(int i=0;i<num_particles;i++)
    {
        TV pos,vel;
        e.random.Fill_Uniform(pos,e.grid.domain.min_corner+e.grid.dX*6,e.grid.domain.max_corner-e.grid.dX*6);
        e.random.Fill_Uniform(vel,(T)-1,(T)1);
        LOG::printf("test pos %P vel %P \n",pos,vel);
        e.Add_Particle(pos,[&vel](const TV&){return vel;},0,mass,volume,r0);
    }
    e.Add_Neo_Hookean(0.073e9*e.unit_p*e.scale_E,0.4);


    auto cc=[&e,num_particles]()
    {
        typename PARTICLE_GRID_WEIGHTS<TV>::SCRATCH scratch;
        HASHTABLE<IV,int> h;
        UNION_FIND<> uf(num_particles);
        for(int i=0;i<num_particles;i++)
            for(PARTICLE_GRID_ITERATOR<TV> it(e.weights,e.particles.X(i),false,false,scratch);it.Valid();it.Next())
            {
                auto pr=h.Insert(it.Index(),i);
                if(!pr.second) uf.Union(i,*pr.first);
            }
        ARRAY<ARRAY<int> > groups(num_particles);
        for(int i=0;i<num_particles;i++)
            groups(uf.Find(i)).Append(i);
        return groups;
    };
    e.reflection_bc=true;
    e.reflection_bc_friction=0;
    // e.Add_Walls(-1,MPM::BC_SEP,.3,0,false);
    e.Add_Walls(-1,MPM::BC_PERIODIC,0,0,false);
    LOG::printf("bc %P\n",e.side_bc_type);
    for(int p=0;p<e.particles.number;p++)
        if(e.particles.valid(p))
            e.simulated_particles.Append(p);
    T max_ke=e.Total_Particle_Kinetic_Energy();
    LOG::printf("%P\n",e.particles.V);
    LOG::printf("%P\n",e.particles.mass);
    LOG::printf("ke %P\n",max_ke);
    e.Add_Callbacks(false,"time-step",
        [&e,cc,max_ke,exit_early]()
        {
            T ke=e.Total_Particle_Kinetic_Energy();
            if(ke<=10*max_ke) return;
            LOG::printf("storedata seed %P dt %P particlenum %P p_interset_num %P UNSTABLE\n",e.seed,e.max_dt,e.particles.number,cc());
            LOG::printf("UNSTABLE %P vs %P    config %P\n",ke,max_ke,cc());
            if(exit_early) exit(0);
        }
    );
    e.destroy=[&e,cc]()
    {
        LOG::printf("storedata seed %P dt %P particlenum %P p_interset_num %P FINISHED\n",e.seed,e.max_dt,e.particles.number,cc());
        LOG::printf("FINISHED    config %P\n",cc());
    };
}
template void Function_131explore(STANDARD_TESTS_BASE<VECTOR<float,2> >& e);
template void Function_131explore(STANDARD_TESTS_BASE<VECTOR<double,2> >& e);
template void Function_131explore(STANDARD_TESTS_BASE<VECTOR<float,3> >& e);
template void Function_131explore(STANDARD_TESTS_BASE<VECTOR<double,3> >& e);
}
