//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __Spline_Interpolation__
#define __Spline_Interpolation__

#include <Core/Arrays/ARRAY.h>
#include <Core/Math_Tools/cube.h>
namespace PhysBAM{

// a * i + b is the x-value corresponding to y(i).
// Evaluate using a cubic B-spline.  Treat the array as though it has an
// infinite number of points beyond each end that are calculated according to:
// 0 = fill with fixed_L or fixed_R
// 1-4 = fill with interpolating polynomial passing through last 1-4 points.

template<class U,class T>
void Spline_Interpolation(const ARRAY<U>& y,T a,T b,T x,int num_L,U fixed_L,int num_R,U fixed_R,U* z,U* dz=0,U* ddz=0,U* dddz=0)
{
    x=(x-b)/a;
    int i=(int)floor(x);
    auto extrap=[](int i,U y0,U y1,U y2,U y3,U fix,int n)
    {
        if(n==0) return fix;
        if(n==1) return y0;
        if(n==2) return -(i-1)*y0+i*y1;
        if(n==3) return (T)(i-1)*(i-2)/2*y0-(T)(i-2)*i*y1+(T)(i-1)*i/2*y2;
        return (-(T)(i-1)*(i-2)*(i-3)/6*y0+(T)i*(i-2)*(i-3)/2*y1-(T)i*(i-1)*(i-3)/2*y2+(T)i*(i-1)*(i-2)/6*y3);
    };
    auto get_y=[=,&y](int i)
    {
        if(i<0) return extrap(i,y(0),y(1),y(2),y(3),fixed_L,num_L);
        if(i>=y.m) return extrap(y.m-1-i,y(y.m-1),y(y.m-2),y(y.m-3),y(y.m-4),fixed_R,num_R);
        return y(i);
    };

    U y0,y1,y2,y3;
    if(i>=1 && i<y.m-2)
    {
        y0=y(i-1);
        y1=y(i);
        y2=y(i+1);
        y3=y(i+2);
    }
    else
    {
        y0=get_y(i-1);
        y1=get_y(i);
        y2=get_y(i+1);
        y3=get_y(i+2);
    }
    T u=x-i,v=1-u,u2=u*u,v2=v*v;
    if(z) *z=(v*v2*y0+(3*u2*(u-2)+4)*y1+(3*v2*(v-2)+4)*y2+u*u2*y3)/6;
    if(dz) *dz=(-v2*y0+u*(3*u-4)*y1+(3*u+1)*v*y2+u2*y3)/(2*a);
    if(ddz) *ddz=(v*y0+(3*u-2)*y1+(-3*u+1)*y2+u*y3)/(a*a);
    if(dddz) *dddz=(-y0+3*y1-3*y2+y3)/(a*a*a);
}

// num, index, coeff-power
extern int spline_coeff_table_L[5][4][4][4];
extern int spline_coeff_table_L_fixed[4][4];
extern int spline_coeff_table_R[5][4][4][4];
extern int spline_coeff_table_R_fixed[4][4];

// returns (r,t):
//  result is y(r)...y(r+3) dotted with z,dz,ddz,dddz.
//  if r==0: add in t*fixed_L
//  if r==n-4: add in t*fixed_R
template<class T>
std::pair<int,T> Spline_Interpolation_Coeffs(int n,T a,T b,T x,int num_L,int num_R,
    VECTOR<T,4>* z,VECTOR<T,4>* dz=0,VECTOR<T,4>* ddz=0,VECTOR<T,4>* dddz=0)
{
    x=(x-b)/a;
    int i=(int)floor(x);
    T fixed=0;
    if(i<1)
    {
        i=std::max(i,-3);
        int j=-i;
        T u=x-i;
        if(z)
        {
            *z=VECTOR<T,4>();
            T pow_u[4]={1./6,u/6,u*u/6,u*u*u/6};
            auto* p=spline_coeff_table_L[num_L][j];
            for(int k=0;k<4;k++)
                for(int q=0;q<4;q++)
                    (*z)(k)+=p[k][q]*pow_u[q];
            if(num_L==0)
                for(int q=0;q<4;q++)
                    fixed+=spline_coeff_table_L_fixed[j][q]*pow_u[q];
        }
        if(dz) *dz=VECTOR<T,4>();
        if(ddz) *ddz=VECTOR<T,4>();
        if(dddz) *dddz=VECTOR<T,4>();
        return {0,fixed};
    }
    if(i>=n-2)
    {
        i=std::min(i,n+1);
        T u=x-i;
        int j=i-n+2;
        if(z)
        {
            *z=VECTOR<T,4>();
            T pow_u[4]={1./6,u/6,u*u/6,u*u*u/6};
            auto* p=spline_coeff_table_R[num_R][j];
            for(int k=0;k<4;k++)
                for(int q=0;q<4;q++)
                    (*z)(k)+=p[k][q]*pow_u[q];
            if(num_R==0)
                for(int q=0;q<4;q++)
                    fixed+=spline_coeff_table_R_fixed[j][q]*pow_u[q];
        }
        if(dz) *dz=VECTOR<T,4>();
        if(ddz) *ddz=VECTOR<T,4>();
        if(dddz) *dddz=VECTOR<T,4>();
        return {n-4,fixed};
    }
    T u=x-i;
    if(z) *z=VECTOR<T,4>(cube(1-u)/6,(3*u*u*(u-2)+4)/6,(3*(1-u)*(1-u)*(1-u-2)+4)/6,cube(u)/6);
    if(dz) *dz=VECTOR<T,4>();
    if(ddz) *ddz=VECTOR<T,4>();
    if(dddz) *dddz=VECTOR<T,4>();
    return {i-1,0};
}

}
#endif
