//#####################################################################
// Copyright 2019, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.text.
//#####################################################################
#include <Core/Matrices/LAPACK.h>
namespace PhysBAM{
// Destroys M
template<class T>
void Least_Squares_Solve_Raw(MATRIX_VIEW<T> M,
    MATRIX_VIEW<const T> rhs,
    MATRIX_VIEW<T> sol,T rcond,int * rank)
{
#ifdef USE_LAPACK
    assert(M.Rows()==rhs.Rows());
    assert(M.Columns()==sol.Rows());
    assert(rhs.Columns()==sol.Columns());

    MATRIX_MXN<T> TS;
    MATRIX_VIEW<T> VM(M.Derived()),VS(sol.Derived());

    bool b=false;
    if(sol.Rows()<rhs.Rows())
    {
        TS.Resize(rhs.Rows(),rhs.Columns());
        VS.Set(TS);
        b=true;
    }
    VS.m=rhs.Rows();
    VS=rhs;
    VS.m=sol.Rows();

    ARRAY<int> jpvt(VM.n);
    int rnk = 0;
    int ret = xgelsy( LAPACK_COL_MAJOR,VM.m,VM.n,rhs.Columns(),VM.x,VM.s,
        VS.x,VS.s,jpvt.Get_Array_Pointer(),rcond,&rnk );
    PHYSBAM_ASSERT(!ret);
    if(rank) *rank=rnk;
    if(b) sol.Derived()=VS;
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

// Preserves M
template<class T>
void Least_Squares_Solve(MATRIX_VIEW<const T> M,
    MATRIX_VIEW<const T> rhs,
    MATRIX_VIEW<T> sol,T rcond,int * rank)
{
    MATRIX_MXN<T> MM(M);
    MATRIX_VIEW<T> MV(MM);
    Least_Squares_Solve_Raw<T>(MV,rhs,sol,rcond,rank);
}

// Destroys M
template<class T>
void Least_Squares_Solve_Raw(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> rhs,
    ARRAY_VIEW<T> sol,T rcond,int * rank)
{
#ifdef USE_LAPACK
    assert(M.Rows()==rhs.Size());
    assert(M.Columns()==sol.Size());

    ARRAY<T> TS;
    MATRIX_VIEW<T> VM(M.Derived());
    ARRAY_VIEW<T> VS;

    bool b=true;
    if(sol.Size()>=rhs.Size()) b=Make_View(VS,sol,TS);
    else
    {
        TS.Resize(rhs.Size());
        VS.Set(TS);
    }
    VS.m=rhs.Size();
    VS=rhs;
    VS.m=sol.Size();

    int ldb=max(VM.m,VM.n);
    ARRAY<int> jpvt(VM.n);
    int rnk = 0;
    int ret = xgelsy( LAPACK_COL_MAJOR,VM.m,VM.n,1,VM.x,VM.s,
        VS.base_pointer,ldb,jpvt.Get_Array_Pointer(),rcond,&rnk );
    PHYSBAM_ASSERT(!ret);
    if(rank) *rank=rnk;
    if(b) sol.Derived()=VS;
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

// Preserves M
template<class T>
void Least_Squares_Solve(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<const T> rhs,ARRAY_VIEW<T> sol,T rcond,int * rank)
{
    MATRIX_MXN<T> MM(M);
    MATRIX_VIEW<T> MV(MM);
    Least_Squares_Solve_Raw<T>(MV,rhs,sol,rcond,rank);
}

// Compute (complex) eigenvalues for non-symmetric real-valued matrix.
// Destroys M
template<class T>
void Compute_Eigenvalues_Raw(MATRIX_VIEW<T> M,ARRAY_VIEW<std::complex<T> > eigenvalues)
{
#ifdef USE_LAPACK
    PHYSBAM_ASSERT(M.m==M.n);
    ARRAY<T> r(eigenvalues.m),i(eigenvalues.m);
    int ret=xgeev(LAPACK_COL_MAJOR,'N','N',M.m,M.x,M.s,
        r.Get_Array_Pointer(),i.Get_Array_Pointer(),0,1,0,1);
    PHYSBAM_ASSERT(!ret);
    for(int k=0;k<eigenvalues.m;k++)
        eigenvalues(k)=std::complex<T>(r(k),i(k));
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

// Preserves M
template<class T>
void Compute_Eigenvalues(MATRIX_VIEW<const T> M,ARRAY_VIEW<std::complex<T> > eigenvalues)
{
    MATRIX_MXN<T> m(M);
    Compute_Eigenvalues_Raw(m,eigenvalues);
}

//#####################################################################
// CBLAS wrapper definition
//#####################################################################
template<class T>
void Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,ARRAY_VIEW<const T> A,
    ARRAY_VIEW<T> sol,T alpha,T beta,bool use_transpose)
{
#ifdef USE_LAPACK
    assert(M.Rows()==sol.m);
    assert(M.Columns()==A.m);
    CBLAS_TRANSPOSE transA=(use_transpose?CblasTrans:CblasNoTrans);
    cblas_xgemv(CblasColMajor,transA,M.m,M.n,alpha,
        M.x, M.m,A.base_pointer,1,beta,sol.base_pointer,1);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

template<class T>
void Symmetric_Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<const T> A,ARRAY_VIEW<T> sol,T alpha,T beta)
{
#ifdef USE_LAPACK
    assert(M.Rows()==sol.m);
    assert(M.Columns()==A.m);
    assert(M.Columns()==M.Rows());
    cblas_xsymv(CblasColMajor,CblasUpper,M.m,alpha,M.x, M.m,
        A.base_pointer,1,beta,sol.base_pointer,1);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory A
template<class T>
void Triangular_Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<T> A,bool use_upper,bool use_transpose,
    bool unit_diagonal)
{
#ifdef USE_LAPACK
    assert(M.Columns()==A.m);
    assert(M.Columns()==M.Rows());
    CBLAS_UPLO uplo=(use_upper?CblasUpper:CblasLower);
    CBLAS_TRANSPOSE transA=(use_transpose?CblasTrans:CblasNoTrans);
    CBLAS_DIAG diag=(unit_diagonal?CblasUnit:CblasNonUnit);
    cblas_xtrmv(CblasColMajor,uplo,transA,diag,M.m,
        M.x, M.m,A.base_pointer,1);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory A
template<class T>
void Triangular_Matrix_Problem_Solver(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<T> A,bool use_upper,bool use_transpose,
    bool unit_diagonal)
{
#ifdef USE_LAPACK
    assert(M.Columns()==A.m);    
    assert(M.Columns()==M.Rows());
    CBLAS_UPLO uplo=(use_upper?CblasUpper:CblasLower);
    CBLAS_TRANSPOSE transA=(use_transpose?CblasTrans:CblasNoTrans);
    CBLAS_DIAG diag=(unit_diagonal?CblasUnit:CblasNonUnit);
    cblas_xtrsv(CblasColMajor,uplo,transA,diag,M.m,
        M.x, M.m,A.base_pointer,1);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory M
template<class T>
void Rank_One_Operation(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> A,ARRAY_VIEW<const T> B, T alpha)
{
#ifdef USE_LAPACK
    assert(M.Columns()==A.Size());    
    assert(M.Rows()==B.Size());
    cblas_xger(CblasColMajor,A.m,B.m,alpha,A.base_pointer,1,
        B.base_pointer,1,M.x,M.m);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory M
template<class T>
void Symmetric_Rank_One_Operation(MATRIX_VIEW<T> M,ARRAY_VIEW<const T> A,
     T alpha)
{
#ifdef USE_LAPACK
    assert(M.Columns()==A.Size());    
    assert(M.Rows()==M.Columns());
    cblas_xsyr(CblasColMajor,CblasLower,A.m,alpha,A.base_pointer,1,M.x,M.m);
    for(int i=0;i<M.m;i++) for(int j=0;j<i;j++) M(j,i)=M(i,j);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}
    
//Destory M
template<class T>
void Symmetric_Rank_Two_Operation(MATRIX_VIEW<T> M,ARRAY_VIEW<const T> A,
    ARRAY_VIEW<const T> B,T alpha)
{
#ifdef USE_LAPACK
    assert(M.Columns()==A.Size());    
    assert(M.Rows()==M.Columns());
    cblas_xsyr2(CblasColMajor,CblasLower,A.m,alpha,A.base_pointer,1,
        B.base_pointer,1,M.x,M.m);
    for(int i=0;i<M.m;i++) for(int j=0;j<i;j++) M(j,i)=M(i,j);

#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory C
template<class T>
void Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,MATRIX_VIEW<const T> B,
    MATRIX_VIEW<T> C,T alpha,T beta,bool A_use_transpose,
    bool B_use_transpose)
{
#ifdef USE_LAPACK
    assert(C.Rows()==(A_use_transpose?A.Columns():A.Rows()));
    assert(C.Columns()==(B_use_transpose?B.Rows():B.Columns()));
    CBLAS_TRANSPOSE transA=(A_use_transpose?CblasTrans:CblasNoTrans);
    CBLAS_TRANSPOSE transB=(B_use_transpose?CblasTrans:CblasNoTrans);
    int k=(A_use_transpose?A.m:A.n);
    cblas_xgemm(CblasColMajor,transA,transB,C.m,C.n,k,alpha,
        A.x,A.m,B.x,B.m,beta,C.x,C.m);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory C
template<class T>
void Symmetric_Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,MATRIX_VIEW<const T> B,
    MATRIX_VIEW<T> C,T alpha,T beta,bool A_on_left_side)
{
#ifdef USE_LAPACK
    assert(C.Rows()==(A_on_left_side?A.Rows():B.Rows()));
    assert(C.Columns()==(A_on_left_side?B.Columns():A.Columns()));
    CBLAS_SIDE side=(A_on_left_side?CblasLeft:CblasRight);
    cblas_xsymm(CblasColMajor,side,CblasLower,C.m,C.n,alpha,A.x,A.m,
        B.x,B.m,beta,C.x,C.m);
#else
    PHYSBAM_FATAL_ERRORc("This requies USE_LAPACK");
#endif
}

//Destory C
template<class T>
void Symmetric_Rank_K_Operation(MATRIX_VIEW<const T> A,MATRIX_VIEW<T> C,
    T alpha,T beta,bool use_transpose)
{
#ifdef USE_LAPACK
    assert(C.Rows()==(use_transpose?A.Columns():A.Rows()));
    assert(C.Rows()==C.Columns());
    CBLAS_TRANSPOSE trans=(use_transpose?CblasTrans:CblasNoTrans);
    int k=(use_transpose?A.Rows():A.Columns());
    cblas_xsyrk(CblasColMajor,CblasLower,trans,C.m,k,alpha,A.x,A.m,
        beta,C.x,C.m);
    for(int i=0;i<C.m;i++) for(int j=0;j<i;j++) C(j,i)=C(i,j);
    
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory C
template<class T>
void Symmetric_Rank_2K_Operation(MATRIX_VIEW<const T> A,MATRIX_VIEW<const T> B,
    MATRIX_VIEW<T> C,T alpha,T beta,bool use_transpose)
{
#ifdef USE_LAPACK
    assert(C.Rows()==(use_transpose?A.Columns():A.Rows()));
    assert(C.Rows()==C.Columns());
    assert(A.Rows()==B.Rows()); assert(A.Columns()==B.Columns());
    CBLAS_TRANSPOSE trans=(use_transpose?CblasTrans:CblasNoTrans);
    int k=(use_transpose?A.Rows():A.Columns());
    cblas_xsyr2k(CblasColMajor,CblasLower,trans,C.m,k,alpha,A.x,A.m,
        B.x,B.m,beta,C.x,C.m);
    for(int i=0;i<C.m;i++) for(int j=0;j<i;j++) C(j,i)=C(i,j);

    
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory B
template<class T>
void Triangular_Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<T> B,T alpha, bool A_on_left_side,bool use_upper,
    bool use_transpose,bool unit_diagonal)
{
#ifdef USE_LAPACK
    assert(A.Columns()==(A_on_left_side?B.Rows():B.Columns()));
    assert(A.Columns()==A.Rows());
    CBLAS_SIDE side=(A_on_left_side?CblasLeft:CblasRight);
    CBLAS_UPLO uplo=(use_upper?CblasUpper:CblasLower);
    CBLAS_TRANSPOSE transA=(use_transpose?CblasTrans:CblasNoTrans);
    CBLAS_DIAG diag=(unit_diagonal?CblasUnit:CblasNonUnit);
    cblas_xtrmm(CblasColMajor,side,uplo,transA,diag,B.m,B.n,alpha,
        A.x,A.m,B.x,B.m);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//Destory B
template<class T>
void Triangular_Matrix_Solver_Multiple(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<T> B,T alpha,bool A_on_left_side,bool use_upper,
    bool use_transpose,bool unit_diagonal)
{
#ifdef USE_LAPACK
    assert(A.Columns()==(A_on_left_side?B.Rows():B.Columns()));
    assert(A.Columns()==A.Rows());
    CBLAS_SIDE side=(A_on_left_side?CblasLeft:CblasRight);
    CBLAS_UPLO uplo=(use_upper?CblasUpper:CblasLower);
    CBLAS_TRANSPOSE transA=(use_transpose?CblasTrans:CblasNoTrans);
    CBLAS_DIAG diag=(unit_diagonal?CblasUnit:CblasNonUnit);
    cblas_xtrsm(CblasColMajor,side,uplo,transA,diag,B.m,B.n,alpha,
        A.x,A.m,B.x,B.m);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//#####################################################################

template void Compute_Eigenvalues_Raw<double>(MATRIX_VIEW<double>,
    ARRAY_VIEW<std::complex<double>>);
template void Compute_Eigenvalues_Raw<float>(MATRIX_VIEW<float>,
    ARRAY_VIEW<std::complex<float>>);
template void Least_Squares_Solve<double>(MATRIX_VIEW<double const>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double>,double,int*);
template void Least_Squares_Solve<float>(MATRIX_VIEW<float const>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float>,float,int*);
template void Least_Squares_Solve_Raw<double>(MATRIX_VIEW<double>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double>,double,int*);
template void Least_Squares_Solve_Raw<float>(MATRIX_VIEW<float>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float>,float,int*);
template void Matrix_Vector_Multiply<double>(MATRIX_VIEW<double const>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double>,double,double,bool);
template void Matrix_Vector_Multiply<float>(MATRIX_VIEW<float const>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float>,float,float,bool);
template void Symmetric_Matrix_Vector_Multiply<double>(MATRIX_VIEW<double const>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double>,double,double);
template void Symmetric_Matrix_Vector_Multiply<float>(MATRIX_VIEW<float const>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float>,float,float);
template void Triangular_Matrix_Vector_Multiply<double>(MATRIX_VIEW<double const>,
    ARRAY_VIEW<double>,bool,bool,bool);
template void Triangular_Matrix_Vector_Multiply<float>(MATRIX_VIEW<float const>,
    ARRAY_VIEW<float>,bool,bool,bool);
template void Triangular_Matrix_Problem_Solver<double>(MATRIX_VIEW<double const>,
    ARRAY_VIEW<double>,bool,bool,bool);
template void Triangular_Matrix_Problem_Solver<float>(MATRIX_VIEW<float const>,
    ARRAY_VIEW<float>,bool,bool,bool);
template void Rank_One_Operation<double>(MATRIX_VIEW<double>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double const>,double);
template void Rank_One_Operation<float>(MATRIX_VIEW<float>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float const>,float);
template void Symmetric_Rank_One_Operation<double>(MATRIX_VIEW<double>,
    ARRAY_VIEW<double const>,double);
template void Symmetric_Rank_One_Operation<float>(MATRIX_VIEW<float>,
    ARRAY_VIEW<float const>,float);
template void Symmetric_Rank_Two_Operation<double>(MATRIX_VIEW<double>,
    ARRAY_VIEW<double const>,ARRAY_VIEW<double const>,double);
template void Symmetric_Rank_Two_Operation<float>(MATRIX_VIEW<float>,
    ARRAY_VIEW<float const>,ARRAY_VIEW<float const>,float);
template void Matrix_Matrix_Multiply<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double const>,MATRIX_VIEW<double>,double,double,bool,bool);
template void Matrix_Matrix_Multiply<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float const>,MATRIX_VIEW<float>,float,float,bool,bool);
template void Symmetric_Matrix_Matrix_Multiply<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double const>,MATRIX_VIEW<double>,double,double,bool);
template void Symmetric_Matrix_Matrix_Multiply<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float const>,MATRIX_VIEW<float>,float,float,bool);
template void Symmetric_Rank_K_Operation<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double>,double,double,bool);
template void Symmetric_Rank_K_Operation<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float>,float,float,bool);
template void Symmetric_Rank_2K_Operation<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double const>,MATRIX_VIEW<double>,double,double,bool);
template void Symmetric_Rank_2K_Operation<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float const>,MATRIX_VIEW<float>,float,float,bool);
template void Triangular_Matrix_Matrix_Multiply<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double>,double,bool,bool,bool,bool);
template void Triangular_Matrix_Matrix_Multiply<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float>,float,bool,bool,bool,bool);
template void Triangular_Matrix_Solver_Multiple<double>(MATRIX_VIEW<double const>,
    MATRIX_VIEW<double>,double,bool,bool,bool,bool);
template void Triangular_Matrix_Solver_Multiple<float>(MATRIX_VIEW<float const>,
    MATRIX_VIEW<float>,float,bool,bool,bool,bool);
template void Compute_Eigenvalues<double>(MATRIX_VIEW<double const>,ARRAY_VIEW<std::complex<double>,int>);
template void Compute_Eigenvalues<float>(MATRIX_VIEW<float const>,ARRAY_VIEW<std::complex<float>,int>);
}







