//#####################################################################
// Copyright 2019, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class LAPACK
//#####################################################################
#ifndef __LAPACK__
#define __LAPACK__
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/ARRAY_VIEW.h>
#include <Core/Matrices/Make_View.h>
#include <Core/Matrices/MATRIX_BASE.h>
#include <Core/Matrices/MATRIX_VIEW.h>
#include <complex>
#ifdef USE_LAPACK
#define lapack_complex_double std::complex<double>
#define lapack_complex_float std::complex<float>
#include <Core/Matrices/SYMMETRIC_MATRIX_NXN.h>
extern "C"{
#include <cblas.h>
#include <lapacke.h>
}
#endif
namespace PhysBAM{

//#####################################################################
// Forward Declarations
//#####################################################################

// Least_Squares_Solve_Raw and Least_Squares_Solve work with non-square and singular M.

// Destroys M
template<class T>
void Least_Squares_Solve_Raw(MATRIX_VIEW<T> M,MATRIX_VIEW<const T> rhs,
    MATRIX_VIEW<T> sol,T rcond=(std::numeric_limits<T>::epsilon()),
    int * rank=0);

template<class T,class T_MATRIX1,class T_MATRIX2,class T_MATRIX3>
void Least_Squares_Solve_Raw(MATRIX_BASE<T,T_MATRIX1>& M,
    const MATRIX_BASE<T,T_MATRIX2>& rhs,
    MATRIX_BASE<T,T_MATRIX3>& sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0)
{
    MATRIX_VIEW<T> MM(M.Derived()),s(sol.Derived());
    MATRIX_VIEW<const T> r(rhs.Derived());
    Least_Squares_Solve_Raw(MM,r,s,rcond,rank);
}
// Preserves M
template<class T>
void Least_Squares_Solve(MATRIX_VIEW<const T> M,MATRIX_VIEW<const T> rhs,
    MATRIX_VIEW<T> sol,T rcond=(std::numeric_limits<T>::epsilon()),
    int * rank=0);

template<class T,class T_MATRIX1,class T_MATRIX2,class T_MATRIX3>
void Least_Squares_Solve(const MATRIX_BASE<T,T_MATRIX1>& M,
    const MATRIX_BASE<T,T_MATRIX2>& rhs,
    MATRIX_BASE<T,T_MATRIX3>& sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0)
{
    MATRIX_VIEW<T> s(sol.Derived());
    MATRIX_VIEW<const T> MM(M.Derived()),r(rhs.Derived());
    Least_Squares_Solve(MM,r,s,rcond,rank);
}

// Destroys M
template<class T>
void Least_Squares_Solve_Raw(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> rhs,ARRAY_VIEW<T> sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0);

// Destroys M
template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Least_Squares_Solve_Raw(MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& rhs,
    ARRAY_BASE<T,T_ARRAY2>& sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0)
{
    MATRIX_VIEW<T> MM(M.Derived());
    ARRAY_VIEW<const T> r(rhs.Derived());
    ARRAY_VIEW<T> s(sol.Derived());
    Least_Squares_Solve_Raw(MM,r,s,rcond,rank);
}

// Preserves M
template<class T>
void Least_Squares_Solve(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<const T> rhs,ARRAY_VIEW<T> sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0);

// Preserves M
template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Least_Squares_Solve(const MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& rhs,
    ARRAY_BASE<T,T_ARRAY2>& sol,
    T rcond=(std::numeric_limits<T>::epsilon()),int * rank=0)
{
    MATRIX_VIEW<const T> MM(M.Derived());
    ARRAY_VIEW<const T> r(rhs.Derived());
    ARRAY_VIEW<T> s(sol.Derived());
    Least_Squares_Solve(MM,r,s,rcond,rank);
}

// Compute (complex) eigenvalues for non-symmetric real-valued matrix.
// Destroys M
template<class T>
void Compute_Eigenvalues_Raw(MATRIX_VIEW<T> M,
    ARRAY_VIEW<std::complex<T> > eigenvalues);

template<class T,class T_MATRIX,class T_ARRAY>
void Compute_Eigenvalues_Raw(MATRIX_BASE<T,T_MATRIX>& M,
    ARRAY_BASE<std::complex<T>,T_ARRAY>& eigenvalues)
{
    MATRIX_VIEW<T> MM(M.Derived());
    ARRAY_VIEW<std::complex<T> > e(eigenvalues.Derived());
    Compute_Eigenvalues_Raw(MM,e);
}

// Preserves M
template<class T>
void Compute_Eigenvalues(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<std::complex<T> > eigenvalues);

template<class T,class T_MATRIX,class T_ARRAY>
void Compute_Eigenvalues(const MATRIX_BASE<T,T_MATRIX>& M,
    ARRAY_BASE<std::complex<T>,T_ARRAY>& eigenvalues)
{
    MATRIX_VIEW<const T> MM(M.Derived());
    ARRAY_VIEW<std::complex<T> > e(eigenvalues.Derived());
    Compute_Eigenvalues(MM,e);
}

//#####################################################################
// CBLAS forward declaration
//#####################################################################

// if use_transpose: sol=alpha*M^T*A+beta*sol
// else: sol=alpha*M*A+beta*sol
template<class T>
void Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,ARRAY_VIEW<const T> A,
    ARRAY_VIEW<T> sol,T alpha,T beta,bool use_transpose);

template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Matrix_Vector_Multiply(const MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& A,ARRAY_BASE<T,T_ARRAY2>& sol,
    T alpha,T beta,bool use_transpose)
{
    MATRIX_VIEW<const T> VM(M.Derived());
    ARRAY_VIEW<const T> VA(A.Derived());
    ARRAY_VIEW<T> VS(sol.Derived());
    Matrix_Vector_Multiply(VM,VA,VS,alpha,beta,use_transpose);
}

// Assume M is a symmetric matrix
// Computes sol=alpha*M*A+beta*sol
template<class T>
void Symmetric_Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<const T> A,ARRAY_VIEW<T> sol,T alpha,T beta);

template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Symmetric_Matrix_Vector_Multiply(const MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& A,ARRAY_BASE<T,T_ARRAY2>& sol,
    T alpha,T beta)
{
    MATRIX_VIEW<const T> VM(M.Derived());
    ARRAY_VIEW<const T> VA(A.Derived());
    ARRAY_VIEW<T> VS(sol.Derived());
    Symmetric_Matrix_Vector_Multiply(VM,VA,VS,alpha,beta);
}

// if use_upper, only look at the upper triangular port of M.
// Otherwise, only look at the lower triangular port of M.
// if use_transpose, computes A=M*A.  Otherwise, computes A=M^T*A.
// if unit_diagonal, A is a unit triangular matrix.
template<class T>
void Triangular_Matrix_Vector_Multiply(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<T> A,bool use_upper,bool use_transpose,
    bool unit_diagonal);

template<class T,class T_MATRIX,class T_ARRAY>
void Triangular_Matrix_Vector_Multiply(const MATRIX_BASE<T,T_MATRIX>& M,
    ARRAY_BASE<T,T_ARRAY>& A,bool use_upper,bool use_transpose,
    bool unit_diagonal)
{
    MATRIX_VIEW<const T> VM(M.Derived());
    ARRAY_VIEW<T> VA(A.Derived());
    Triangular_Matrix_Vector_Multiply(VM,VA,use_upper,
        use_transpose,unit_diagonal);
}

// if use_upper, only look at the upper triangular port of M.
// Otherwise, only look at the lower triangular port of M.
// if use_transpose, solves A*x=b.  Otherwise, solves A^T*x=b.
// if unit_diagonal, A is a unit triangular matrix.
template<class T>
void Triangular_Matrix_Problem_Solver(MATRIX_VIEW<const T> M,
    ARRAY_VIEW<T> A,bool use_upper,bool use_transpose,
    bool unit_diagonal);

template<class T,class T_MATRIX,class T_ARRAY>
void Triangular_Matrix_Problem_Solver(const MATRIX_BASE<T,T_MATRIX>& M,
    ARRAY_BASE<T,T_ARRAY>& A,bool use_upper,bool use_transpose,
    bool unit_diagonal)
{
    MATRIX_VIEW<const T> VM(M.Derived());
    ARRAY_VIEW<T> VA(A.Derived());
    Triangular_Matrix_Problem_Solver(VM,VA,use_upper,
        use_transpose,unit_diagonal);
}

// Computes M+=alpha*A*B^T
template<class T>
void Rank_One_Operation(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> A,ARRAY_VIEW<const T> B,T alpha);

template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Rank_One_Operation(MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& A,const ARRAY_BASE<T,T_ARRAY2>& B,
    T alpha)
{
    MATRIX_VIEW<T> VM(M.Derived());
    ARRAY_VIEW<const T> VA(A.Derived());
    ARRAY_VIEW<const T> VB(B.Derived()); 
    Rank_One_Operation(VM,VA,VB,alpha);
}

// Assume M is a symmetric matrix. Computes M+=alpha*A*A^T
template<class T>
void Symmetric_Rank_One_Operation(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> A,T alpha);

template<class T,class T_MATRIX,class T_ARRAY>
void Symmetric_Rank_One_Operation(MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY>& A,T alpha)
{
    MATRIX_VIEW<T> VM(M.Derived());
    ARRAY_VIEW<const T> VA(A.Derived());
    Symmetric_Rank_One_Operation(VM,VA,alpha);
}

// Assume M is a symmetric matrix.
// Computes M=alpha*A*B^T+alpha*B*A^T+M
template<class T>
void Symmetric_Rank_Two_Operation(MATRIX_VIEW<T> M,
    ARRAY_VIEW<const T> A,ARRAY_VIEW<const T> B,T alpha);

template<class T,class T_MATRIX,class T_ARRAY1,class T_ARRAY2>
void Symmetric_Rank_Two_Operation(MATRIX_BASE<T,T_MATRIX>& M,
    const ARRAY_BASE<T,T_ARRAY1>& A,const ARRAY_BASE<T,T_ARRAY2>& B,
    T alpha)
{
    MATRIX_VIEW<T> VM(M.Derived());
    ARRAY_VIEW<const T> VA(A.Derived());
    ARRAY_VIEW<const T> VB(B.Derived());
    Symmetric_Rank_Two_Operation(VM,VA,VB,alpha);
}

// if A_use_transpose and B_use_transpose: C=alpha*A^T*B^T+beta*C
// if A_use_transpose and not B_use_transpose: C=alpha*A^T*B+beta*C
// if not A_use_transpose and B_use_transpose: C=alpha*A*B^T+beta*C
// if not A_use_transpose and not B_use_transpose: C=alpha*A*B+beta*C
template<class T>
void Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<const T> B,MATRIX_VIEW<T> C,T alpha,T beta,
    bool A_use_transpose,bool B_use_transpose);

template<class T,class T_MATRIX1,class T_MATRIX2,class T_MATRIX3>
void Matrix_Matrix_Multiply(const MATRIX_BASE<T,T_MATRIX1>& A,
    const MATRIX_BASE<T,T_MATRIX2>& B,MATRIX_BASE<T,T_MATRIX3>& C,
    T alpha,T beta,bool A_use_transpose,bool B_use_transpose)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<const T> VB(B.Derived());
    MATRIX_VIEW<T> VC(C.Derived()); 
    Matrix_Matrix_Multiply(VA,VB,VC,alpha,beta,
        A_use_transpose,B_use_transpose);
}

// Assume A is a symmetric matrix, B and C are m by n matrices.
// If A_on_left_side, computes C=alpha*A*B+beta*C
// Else, computes C=alpha*B*A+beta*C
template<class T>
void Symmetric_Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<const T> B,MATRIX_VIEW<T> C,T alpha,T beta,
    bool A_on_left_side);

template<class T,class T_MATRIX1,class T_MATRIX2,class T_MATRIX3>
void Symmetric_Matrix_Matrix_Multiply(const MATRIX_BASE<T,T_MATRIX1>& A,
    const MATRIX_BASE<T,T_MATRIX2>& B,MATRIX_BASE<T,T_MATRIX3>& C,
    T alpha,T beta,bool A_on_left_side)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<const T> VB(B.Derived());
    MATRIX_VIEW<T> VC(C.Derived()); 
    Symmetric_Matrix_Matrix_Multiply(VA,VB,VC,alpha,beta,
        A_on_left_side);
}

// Assume C is an n by n symmetric matrix.
// If use_transpose, A is an n by k matrix,
// computes C=alpha*A^T*A+beta*C.
// Else, A is a k by n matrix, computes C=alpha*A*A^T+beta*C.
template<class T>
void Symmetric_Rank_K_Operation(MATRIX_VIEW<const T> A,MATRIX_VIEW<T> C,
    T alpha,T beta,bool use_transpose);

template<class T,class T_MATRIX1,class T_MATRIX2>
void Symmetric_Rank_K_Operation(const MATRIX_BASE<T,T_MATRIX1>& A,
    MATRIX_BASE<T,T_MATRIX2>& C,T alpha,T beta,bool use_transpose)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<T> VC(C.Derived()); 
    Symmetric_Rank_K_Operation(VA,VC,alpha,beta,use_transpose);
}

// Assume C is an n by n symmetric matrix.
// If use_transpose, A and B are k by n matrices,
// computes C=alpha*A^T*B+alpha*B^T*A+beta*C
// Else, A and B are n by k matrices,
// computes C=alpha*A*B^T+alpha*B*A^T+beta*C
template<class T>
void Symmetric_Rank_2K_Operation(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<const T> B,MATRIX_VIEW<T> C,
    T alpha,T beta,bool use_transpose);

template<class T,class T_MATRIX1,class T_MATRIX2,class T_MATRIX3>
void Symmetric_Rank_2K_Operation(const MATRIX_BASE<T,T_MATRIX1>& A,
    const MATRIX_BASE<T,T_MATRIX2>& B,MATRIX_BASE<T,T_MATRIX3>& C,
    T alpha,T beta,bool use_transpose)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<const T> VB(B.Derived());
    MATRIX_VIEW<T> VC(C.Derived()); 
    Symmetric_Rank_2K_Operation(VA,VB,VC,alpha,beta,use_transpose);
}

// if A_use_transpose and A_on_left_side: B=alpha*A^T*B.
// if not A_use_transpose and A_on_left_side: B=alpha*A*B.
// if A_use_transpose and not A_on_left_side: B=alpha*B*A^T.
// if not A_use_transpose and not A_on_left_side: B=alpha*B*A.
// if use_upper, assume A is an upper triangular matrix,
// else, assume A is an lower triangular matrix.
// if unit_diagonal, assume A is an unit triangular matrix.
template<class T>
void Triangular_Matrix_Matrix_Multiply(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<T> B,T alpha,bool A_on_left_side,bool use_upper,
    bool A_use_transpose,bool unit_diagonal);

template<class T,class T_MATRIX1,class T_MATRIX2>
void Triangular_Matrix_Matrix_Multiply(const MATRIX_BASE<T,T_MATRIX1>& A,
    MATRIX_BASE<T,T_MATRIX2>& B,T alpha,bool A_on_left_side,
    bool use_upper,bool use_transpose,bool unit_diagonal)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<T> VB(B.Derived()); 
    Triangular_Matrix_Matrix_Multiply(VA,VB,alpha,A_on_left_side,
        use_upper,use_transpose,unit_diagonal);
}

// Assume A is an nxn triangular matrix, X and B are m by n matrices.
// if A_use_transpose and A_on_left_side: A^T*X=alpha*B.
// if not A_use_transpose and A_on_left_side: A*X=alpha*B.
// if A_use_transpose and not A_on_left_side: X*A^T=alpha*B.
// if not A_use_transpose and not A_on_left_side: X*A^T=alpha*B.
// if use_upper, assume A is an upper triangular matrix,
// else, assume A is an lower triangular matrix.
// if unit_diagonal, assume A is an unit triangular.
// B is overwritten by X on exit.
template<class T>
void Triangular_Matrix_Solver_Multiple(MATRIX_VIEW<const T> A,
    MATRIX_VIEW<T> B,T alpha,bool A_on_left_side,bool use_upper,
    bool use_transpose,bool unit_diagonal);

template<class T,class T_MATRIX1,class T_MATRIX2>
void Triangular_Matrix_Solver_Multiple(const MATRIX_BASE<T,T_MATRIX1>& A,
    MATRIX_BASE<T,T_MATRIX2>& B,T alpha,bool A_on_left_side,
    bool use_upper,bool use_transpose,bool unit_diagonal)
{
    MATRIX_VIEW<const T> VA(A.Derived());
    MATRIX_VIEW<T> VB(B.Derived()); 
    Triangular_Matrix_Solver_Multiple(VA,VB,alpha,A_on_left_side,
        use_upper,use_transpose,unit_diagonal);
}

//#####################################################################
// LAPACK Templatizations Wrappers
//#####################################################################

inline int xgelsy(int matrix_layout,int m,int n,int nrhs,double* a,
    int lda,double* b,int ldb,int* jpvt,double rcond,int* rank)
{
#ifdef USE_LAPACK
    return LAPACKE_dgelsy( matrix_layout,m,n,nrhs,a,lda,b,ldb,jpvt,rcond,rank );
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

inline int xgelsy(int matrix_layout,int m,int n,int nrhs,float* a,
    int lda,float* b,int ldb,int* jpvt,float rcond,int* rank)
{
#ifdef USE_LAPACK
    return LAPACKE_sgelsy( matrix_layout,m,n,nrhs,a,lda,b,ldb,jpvt,rcond,rank );
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

inline int xgelsy(int matrix_layout,int m,int n,int nrhs,std::complex<double>* a,
    int lda,std::complex<double>* b,int ldb,int* jpvt,double rcond,int* rank)
{
#ifdef USE_LAPACK
    return LAPACKE_zgelsy( matrix_layout,m,n,nrhs,a,lda,b,ldb,jpvt,rcond,rank );
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

inline int xgelsy(int matrix_layout,int m,int n,int nrhs,std::complex<float>* a,
    int lda,std::complex<float>* b,int ldb,int* jpvt,float rcond,int* rank)
{
#ifdef USE_LAPACK
    return LAPACKE_cgelsy( matrix_layout,m,n,nrhs,a,lda,b,ldb,jpvt,rcond,rank );
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

inline int xgeev(int matrix_layout,char jobvl,char jobvr,int n,float* a,int lda,
    float* wr,float* wi,float* vl,int ldvl,float* vr,int ldvr)
{
#ifdef USE_LAPACK
    return LAPACKE_sgeev(matrix_layout,jobvl,jobvr,n,a,lda,wr,wi,vl,ldvl,vr,ldvr);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

inline int xgeev(int matrix_layout,char jobvl,char jobvr,int n,double* a,int lda,
    double* wr,double* wi,double* vl,int ldvl,double* vr,int ldvr)
{
#ifdef USE_LAPACK
    return LAPACKE_dgeev(matrix_layout,jobvl,jobvr,n,a,lda,wr,wi,vl,ldvl,vr,ldvr);
#else
    PHYSBAM_FATAL_ERROR("This requies USE_LAPACK");
#endif
}

//#####################################################################
// CBLAS Templatizations Wrappers
//#####################################################################
inline void cblas_xgemv(CBLAS_ORDER matrix_layout,CBLAS_TRANSPOSE transA,
    int m,int n,float alpha,const float* a,int lda,const float* x,int incx,
    float beta,float* y,int incy)
{
#ifdef USE_LAPACK
    cblas_sgemv(matrix_layout,transA,m,n,alpha,a,lda,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xgemv(CBLAS_ORDER matrix_layout,CBLAS_TRANSPOSE transA,
    int m,int n,double alpha,const double* a,int lda,const double* x,int incx,
    double beta,double* y,int incy)
{
#ifdef USE_LAPACK
    cblas_dgemv(matrix_layout,transA,m,n,alpha,a,lda,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsymv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    float alpha,const float *a,int lda,const float *x,int incx,float beta,
    float *y,int incy)
{
#ifdef USE_LAPACK
    cblas_ssymv(matrix_layout,uplo,n,alpha,a,lda,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsymv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    double alpha,const double *a,int lda,const double *x,int incx,double beta,
    double *y,int incy)
{
#ifdef USE_LAPACK
    cblas_dsymv(matrix_layout,uplo,n,alpha,a,lda,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xspmv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    float alpha,float *ap,float *x,int incx,float beta,float *y,int incy)
{
#ifdef USE_LAPACK
    cblas_sspmv(matrix_layout,uplo,n,alpha,ap,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xspmv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    double alpha,double *ap,double *x,int incx,double beta,double *y,int incy)
{
#ifdef USE_LAPACK
    cblas_dspmv(matrix_layout,uplo,n,alpha,ap,x,incx,beta,y,incy);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xtrmv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int n,const float *a,int lda,
    float *x,int incx)
{
#ifdef USE_LAPACK
    cblas_strmv(matrix_layout,uplo,transA,diag,n,a,lda,x,incx);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xtrmv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int n,const double *a,int lda,
    double *x,int incx)
{
#ifdef USE_LAPACK
    cblas_dtrmv(matrix_layout,uplo,transA,diag,n,a,lda,x,incx);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xtrsv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int n,const float *a,int lda,
    float *x,int incx)
{
#ifdef USE_LAPACK
    cblas_strsv(matrix_layout,uplo,transA,diag,n,a,lda,x,incx);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xtrsv(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int n,const double *a,int lda,
    double *x,int incx)
{
#ifdef USE_LAPACK
    cblas_dtrsv(matrix_layout,uplo,transA,diag,n,a,lda,x,incx);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xger(CBLAS_ORDER matrix_layout,int m,int n,float alpha,
    const float *x,int incx,const float *y,int incy,float *a,int lda)
{
#ifdef USE_LAPACK
    cblas_sger(matrix_layout,m,n,alpha,x,incx,y,incy,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xger(CBLAS_ORDER matrix_layout,int m,int n,double alpha,
    const double *x,int incx,const double *y,int incy,double *a,int lda)
{
#ifdef USE_LAPACK
    cblas_dger(matrix_layout,m,n,alpha,x,incx,y,incy,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsyr(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    float alpha,const float *x,int incx,float *a,int lda)
{
#ifdef USE_LAPACK
    cblas_ssyr(matrix_layout,uplo,n,alpha,x,incx,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsyr(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    double alpha,const double *x,int incx,double *a,int lda)
{
#ifdef USE_LAPACK
    cblas_dsyr(matrix_layout,uplo,n,alpha,x,incx,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsyr2(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    float alpha,const float *x,int incx,const float *y,int incy,float *a,
    int lda)
{
#ifdef USE_LAPACK
    cblas_ssyr2(matrix_layout,uplo,n,alpha,x,incx,y,incy,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xsyr2(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    double alpha,const double *x,int incx,const double *y,int incy,double *a,
    int lda)
{
#ifdef USE_LAPACK
    cblas_dsyr2(matrix_layout,uplo,n,alpha,x,incx,y,incy,a,lda);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xspr2(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    float alpha,const float *x,int incx,const float *y,int incy,float *a)
{
#ifdef USE_LAPACK
    cblas_sspr2(matrix_layout,uplo,n,alpha,x,incx,y,incy,a);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xspr2(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,int n,
    double alpha,const double *x,int incx,const double *y,int incy,double *a)
{
#ifdef USE_LAPACK
    cblas_dspr2(matrix_layout,uplo,n,alpha,x,incx,y,incy,a);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif
}

inline void cblas_xgemm(CBLAS_ORDER matrix_layout,CBLAS_TRANSPOSE transA,
    CBLAS_TRANSPOSE transB,int m,int n,int k,float alpha,const float *a,
    int lda,const float *b,int ldb,float beta,float *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_sgemm(matrix_layout,transA,transB,m,n,k,alpha,a,lda,
        b,ldb,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xgemm(CBLAS_ORDER matrix_layout,CBLAS_TRANSPOSE transA,
    CBLAS_TRANSPOSE transB,int m,int n,int k,double alpha,const double *a,
    int lda,const double *b,int ldb,double beta,double *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_dgemm(matrix_layout,transA,transB,m,n,k,alpha,a,lda,
        b,ldb,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsymm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,int m,int n,float alpha,const float *a,int lda,
    const float *b,int ldb,float beta,float *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_ssymm(matrix_layout,side,uplo,m,n,alpha,a,lda,b,ldb,
        beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsymm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,int m,int n,double alpha,const double *a,int lda,
    const double *b,int ldb,double beta,double *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_dsymm(matrix_layout,side,uplo,m,n,alpha,a,lda,b,ldb,
        beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsyrk(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE trans,int n,int k,float alpha,const float *a,int lda,
    float beta,float *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_ssyrk(matrix_layout,uplo,trans,n,k,alpha,a,lda,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsyrk(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE trans,int n,int k,double alpha,const double *a,int lda,
    double beta,double *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_dsyrk(matrix_layout,uplo,trans,n,k,alpha,a,lda,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsyr2k(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE trans,int n,int k,float alpha,const float *a,int lda,
    const float *b,int ldb,float beta,float *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_ssyr2k(matrix_layout,uplo,trans,n,k,alpha,a,lda,b,ldb,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xsyr2k(CBLAS_ORDER matrix_layout,CBLAS_UPLO uplo,
    CBLAS_TRANSPOSE trans,int n,int k,double alpha,const double *a,int lda,
    const double *b,int ldb,double beta,double *c,int ldc)
{
#ifdef USE_LAPACK
    cblas_dsyr2k(matrix_layout,uplo,trans,n,k,alpha,a,lda,b,ldb,beta,c,ldc);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xtrmm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int m,int n,
    float alpha,const float *a,int lda,float *b,int ldb)
{
#ifdef USE_LAPACK
    cblas_strmm(matrix_layout,side,uplo,transA,diag,m,n,alpha,a,lda,b,ldb);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xtrmm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int m,int n,
    double alpha,const double *a,int lda,double *b,int ldb)
{
#ifdef USE_LAPACK
    cblas_dtrmm(matrix_layout,side,uplo,transA,diag,m,n,alpha,a,lda,b,ldb);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xtrsm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int m,int n,
    float alpha,const float *a,int lda,float *b,int ldb)
{
#ifdef USE_LAPACK
    cblas_strsm(matrix_layout,side,uplo,transA,diag,m,n,alpha,a,lda,b,ldb);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}

inline void cblas_xtrsm(CBLAS_ORDER matrix_layout,CBLAS_SIDE side,
    CBLAS_UPLO uplo,CBLAS_TRANSPOSE transA,CBLAS_DIAG diag,int m,int n,
    double alpha,const double *a,int lda,double *b,int ldb)
{
#ifdef USE_LAPACK
    cblas_dtrsm(matrix_layout,side,uplo,transA,diag,m,n,alpha,a,lda,b,ldb);
#else
    PHYSBAM_FATAL_ERROR("This requires USE_LAPACK");
#endif    
}
//#####################################################################
}
#endif


    
