//#####################################################################
// Copyright 2002, 2003, Robert Bridson, Ronald Fedkiw, Eran Guendelman.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class PLANE
//##################################################################### 
#include <Geometry/Basic_Geometry/PLANE.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Matrices/FRAME.h>
#include <Geometry/Basic_Geometry/RAY.h>
#include <Geometry/Basic_Geometry/RAY_HIT.h>
#include <Geometry/Implicit_Objects/INTERSECTION_HELPERS.h>
#include <Geometry/Meshes/SEGMENTED_CURVE.h>
#include <Geometry/Meshes/TRIANGULATED_SURFACE.h>
namespace PhysBAM{
//#####################################################################
// Segment_Plane_Intersection
//#####################################################################
template<class T> bool PLANE<T>::
Segment_Plane_Intersection(const TV& endpoint1,const TV& endpoint2,T& interpolation_fraction) const
{
    T denominator=TV::Dot_Product(endpoint2-endpoint1,normal);
    if(!denominator){interpolation_fraction=FLT_MAX;return false;} // parallel
    interpolation_fraction=TV::Dot_Product(x0-endpoint1,normal)/denominator;
    return (interpolation_fraction>=0 && interpolation_fraction<=1);
}
//#####################################################################
// Function Scaled
//#####################################################################
template<class T,class TV> PLANE<T>
Scaled(const PLANE<T>& b,const TV& center,T scale)
{
    return PLANE<T>(b.normal,(b.x0-center)*scale+center);
}
//#####################################################################
// Function Transformed
//#####################################################################
template<class T,class TV> PLANE<T>
Transformed(const PLANE<T>& b,const FRAME<TV>& f)
{
    return PLANE<T>(f.r.Rotate(b.normal),f*b.x0);
}
//#####################################################################
// Function Generate_Triangles
//#####################################################################
template<class T> TRIANGULATED_SURFACE<T>*
Generate_Triangles(const PLANE<T>& plane,T tesselated_size,int elements_per_side)
{
    typedef VECTOR<T,3> TV;
    TRIANGULATED_SURFACE<T>* surface=TRIANGULATED_SURFACE<T>::Create();
    int m=elements_per_side+1,n=elements_per_side;
    surface->particles.Add_Elements(m*m);
    surface->mesh.Initialize_Herring_Bone_Mesh(m,m);
    TV u=tesselated_size*plane.normal.Unit_Orthogonal_Vector(),v=TV::Cross_Product(plane.normal,u);
    for(int j=0,p=0;j<=n;j++)
        for(int i=0;i<=n;i++)
            surface->particles.X(p++)=plane.x0+u*((T)2*j/n-1)-v*((T)2*i/n-1);
    surface->Update_Number_Nodes();
    return surface;
}
//#####################################################################
// Function Intersects
//#####################################################################
template<class T> RAY_HIT<T> Intersects(const RAY<VECTOR<T,3> >& ray,const PLANE<T>& plane,T isocontour)
{
    T distance=plane.normal.Dot(ray.endpoint-plane.x0)-isocontour;
    T rate_of_approach=-plane.normal.Dot(ray.direction);
    if(rate_of_approach*distance<=0) return {}; // no intersection
    return RAY_HIT<T>{distance/rate_of_approach};
}
//#####################################################################
// Function Intersection_Interval
//#####################################################################
template<class T> VECTOR<RAY_HIT<T>,2>
Intersection_Interval(const RAY<VECTOR<T,3> >& ray,const PLANE<T>& plane,T isocontour)
{
    T distance=plane.normal.Dot(ray.endpoint-plane.x0)-isocontour;
    T rate_of_approach=-plane.normal.Dot(ray.direction);
    VECTOR<RAY_HIT<T>,2> v;
    if(distance<0)
    {
        v(0).Set(0);
        if(rate_of_approach<0) v(1).Set(distance/rate_of_approach);
    }
    else if(rate_of_approach>0) v(0).Set(distance/rate_of_approach);
    return v;
}
//#####################################################################
// Function Intersects
//#####################################################################
template<class T> RAY_HIT<T> Intersects(const RAY<VECTOR<T,3> >& ray,
    const PLANE<T>& plane, const T thickness_over_two,const T distance,
    const T rate_of_approach)
{
    RAY_HIT<T> hit;
    if(distance>-thickness_over_two && distance<thickness_over_two)
        return RAY_HIT<T>{0}; // within the boundary
    if(rate_of_approach*distance<=0) return {}; // no intersection
    if(rate_of_approach>0) hit.Set((distance-thickness_over_two)/rate_of_approach);
    else hit.Set((distance+thickness_over_two)/rate_of_approach);
    if(hit.dist>FLT_MAX || hit.dist<-FLT_MAX) return {}; // guard against overflow
    return hit;
}
//#####################################################################
// Function Intersects
//#####################################################################
template<class T> RAY_HIT<T> Rectangle_Intersects(const RAY<VECTOR<T,3> >& ray,
    const PLANE<T>& plane,const PLANE<T>& bounding_plane1,
    const PLANE<T>& bounding_plane2,const PLANE<T>& bounding_plane3,
    const PLANE<T>& bounding_plane4,const T thickness_over_two)
{
    auto hit=Intersects(ray,plane,thickness_over_two);
    if(!hit.Hit()) return hit;
    VECTOR<T,3> point=ray.Point(hit.dist);
    if(bounding_plane1.Outside(point,thickness_over_two)) return {};
    if(bounding_plane2.Outside(point,thickness_over_two)) return {};
    if(bounding_plane3.Outside(point,thickness_over_two)) return {};
    if(bounding_plane4.Outside(point,thickness_over_two)) return {};
    return hit;
}
//#####################################################################
// Function All_Intersections
//#####################################################################
template<class T> void
All_Intersections(ARRAY<RAY_HIT<T> >& hits,const RAY<VECTOR<T,3> >& ray,const PLANE<T>& plane, T isocontour)
{
    auto iv=Intersection_Interval(ray,plane,isocontour);
    hits=iv.Array_View(0,Actual_Length(iv));
}
//#####################################################################
template class PLANE<float>;
template class PLANE<double>;
template PLANE<double> Scaled<double,VECTOR<double,3> >(PLANE<double> const&,VECTOR<double,3> const&,double);
template PLANE<double> Transformed<double,VECTOR<double,3> >(PLANE<double> const&,FRAME<VECTOR<double,3> > const&);
template PLANE<float> Scaled<float,VECTOR<float,3> >(PLANE<float> const&,VECTOR<float,3> const&,float);
template PLANE<float> Transformed<float,VECTOR<float,3> >(PLANE<float> const&,FRAME<VECTOR<float,3> > const&);
template TRIANGULATED_SURFACE<double>* Generate_Triangles<double>(PLANE<double> const&,double,int);
template TRIANGULATED_SURFACE<float>* Generate_Triangles<float>(PLANE<float> const&,float,int);
template RAY_HIT<double> Intersects<double>(RAY<VECTOR<double,3> > const&,PLANE<double> const&,double);
template RAY_HIT<float> Intersects<float>(RAY<VECTOR<float,3> > const&,PLANE<float> const&,float);
template VECTOR<RAY_HIT<double>,2> Intersection_Interval<double>(RAY<VECTOR<double,3> > const&,PLANE<double> const&,double);
template VECTOR<RAY_HIT<float>,2> Intersection_Interval<float>(RAY<VECTOR<float,3> > const&,PLANE<float> const&,float);
template RAY_HIT<float> Intersects(const RAY<VECTOR<float,3> >&,const PLANE<float>&,const float,const float,const float);
template RAY_HIT<float> Rectangle_Intersects(const RAY<VECTOR<float,3> >&,const PLANE<float>&,const PLANE<float>&,const PLANE<float>&,const PLANE<float>&,const PLANE<float>&,const float);
template RAY_HIT<double> Intersects(const RAY<VECTOR<double,3> >&,const PLANE<double>&,const double,const double,const double);
template RAY_HIT<double> Rectangle_Intersects(const RAY<VECTOR<double,3> >&,const PLANE<double>&,const PLANE<double>&,const PLANE<double>&,const PLANE<double>&,const PLANE<double>&,const double);
template void All_Intersections<double>(ARRAY<RAY_HIT<double>,int>&,RAY<VECTOR<double,3> > const&,PLANE<double> const&,double);
template void All_Intersections<float>(ARRAY<RAY_HIT<float>,int>&,RAY<VECTOR<float,3> > const&,PLANE<float> const&,float);
}
