//#####################################################################
// Copyright 2007, Michael Lentine, Andrew Selle.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class IMPLICIT_OBJECT_COMBINED
//#####################################################################
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_COMBINED.h>
#include <Core/Matrices/MATRIX_3X3.h>
#include <Grid_Tools/Grids/GRID.h>
#include <Geometry/Basic_Geometry/BASIC_GEOMETRY_POLICY.h>
#include <Geometry/Basic_Geometry/ORIENTED_BOX.h>
#include <Geometry/Basic_Geometry/RAY.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT.h>

namespace PhysBAM{
//#####################################################################
// Destructor
//#####################################################################
template<class TV> IMPLICIT_OBJECT_COMBINED<TV>::
~IMPLICIT_OBJECT_COMBINED()
{
    if(owns_implicit_object1) delete implicit_object1;
    if(owns_implicit_object2) delete implicit_object2;
}
//#####################################################################
// Function Update_Box
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Update_Box()
{
    implicit_object1->Update_Box();
    implicit_object2->Update_Box();
    box=implicit_object1->box;
    box.Enlarge_Nonempty_Box_To_Include_Points(implicit_object2->box.min_corner,implicit_object2->box.max_corner);
}
//#####################################################################
// Function Update_Minimum_Cell_Size
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Update_Minimum_Cell_Size(const int maximum_depth)
{
    implicit_object1->Update_Minimum_Cell_Size(maximum_depth);
    implicit_object2->Update_Minimum_Cell_Size(maximum_depth);
}
//#####################################################################
// Function Minimum_Cell_Size_Within_Box
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Minimum_Cell_Size_Within_Box(const RANGE<TV>& box) const -> T
{
    return min(implicit_object1->Minimum_Cell_Size_Within_Box(box),implicit_object2->Minimum_Cell_Size_Within_Box(box));
} 
//#####################################################################
// Function operator()
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
operator()(const TV& location) const -> T
{
    return (1-alpha)*(*implicit_object1)(location)+alpha*(*implicit_object2)(location);
}
//#####################################################################
// Function Signed_Distance
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Signed_Distance(const TV& location) const -> T
{
    return (*this)(location);
} // to make this class compatible with the geometry classes
//#####################################################################
// Function Extended_Phi
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Extended_Phi(const TV& location) const -> T
{
    return Extended_Value(location);
}
//#####################################################################
// Function Phi_Secondary
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Phi_Secondary(const TV& location) const -> T
{
    return (1-alpha)*implicit_object1->Phi_Secondary(location)+alpha*implicit_object2->Phi_Secondary(location);
}
//#####################################################################
// Function Normal
//#####################################################################
template<class TV> TV IMPLICIT_OBJECT_COMBINED<TV>::
Normal(const TV& location,const int aggregate) const
{
    return (1-alpha)*implicit_object1->Normal(location,aggregate)+alpha*implicit_object2->Normal(location,aggregate);
}
//#####################################################################
// Function Extended_Normal
//#####################################################################
template<class TV> TV IMPLICIT_OBJECT_COMBINED<TV>::
Extended_Normal(const TV& location,const int aggregate) const
{
    return (1-alpha)*implicit_object1->Extended_Normal(location,aggregate)+alpha*implicit_object2->Extended_Normal(location,aggregate);
}
//#####################################################################
// Function Compute_Normals
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Compute_Normals()
{
    implicit_object1->Compute_Normals();
    implicit_object2->Compute_Normals();
}
//#####################################################################
// Function Compute_Cell_Minimum_And_Maximum
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Compute_Cell_Minimum_And_Maximum(const bool recompute_if_exists)
{
    implicit_object1->Compute_Cell_Minimum_And_Maximum(recompute_if_exists);
    implicit_object2->Compute_Cell_Minimum_And_Maximum(recompute_if_exists);
}
//#####################################################################
// Function Rescale
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Rescale(const T scaling_factor)
{
    implicit_object1->Rescale(scaling_factor);
    implicit_object2->Rescale(scaling_factor);
}
//#####################################################################
// Function Inflate
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Inflate(const T inflation_distance)
{
    implicit_object1->Inflate(inflation_distance);
    implicit_object2->Inflate(inflation_distance);
}
//#####################################################################
// Function Lazy_Inside
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Inside(const TV& location,const T contour_value) const
{
    return box.Inside(location,-contour_value) && (*this)(location)<=contour_value;
}
//#####################################################################
// Function Lazy_Inside_And_Value
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Inside_And_Value(const TV& location,T& phi,const T contour_value) const
{
    if(box.Inside(location,-contour_value)){
        phi=(*this)(location);
        if(phi<=contour_value) return true;
    } return false;
}
//#####################################################################
// Function Lazy_Inside_Extended_Levelset
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Inside_Extended_Levelset(const TV& location,const T contour_value) const
{
    return box.Inside(location,-contour_value) && Extended_Value(location)<=contour_value;
}
//#####################################################################
// Function Lazy_Inside_Extended_Levelset_And_Value
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Inside_Extended_Levelset_And_Value(const TV& location,T& phi_value,const T contour_value) const
{
    if(box.Inside(location,-contour_value)){
        phi_value=Extended_Phi(location);
        if(phi_value<=contour_value) return true;
    } return false;
}
//#####################################################################
// Function Lazy_Outside
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Outside(const TV& location,const T contour_value) const
{
    return !Lazy_Inside(location,contour_value);
}
//#####################################################################
// Function Lazy_Outside_Extended_Levelset
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Outside_Extended_Levelset(const TV& location,const T contour_value) const
{
    return !Lazy_Inside_Extended_Levelset(location,contour_value);
}
//#####################################################################
// Function Lazy_Outside_Extended_Levelset_And_Value
//#####################################################################
template<class TV> bool IMPLICIT_OBJECT_COMBINED<TV>::
Lazy_Outside_Extended_Levelset_And_Value(const TV& location,T& phi_value,const T contour_value) const
{
    return !Lazy_Inside_Extended_Levelset_And_Value(location,phi_value,contour_value);
}
//#####################################################################
// Function Hessian
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Hessian(const TV& X) const -> SYMMETRIC_MATRIX<T,TV::m>
{
    return (1-alpha)*implicit_object1->Hessian(X)+alpha*implicit_object2->Hessian(X);
}
//#####################################################################
// Function Integration_Step
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Integration_Step(const T phi) const -> T
{
    return (1-alpha)*implicit_object1->Integration_Step(phi)+alpha*implicit_object2->Integration_Step(phi);
}
//#####################################################################
// Function Minimum_Cell_Size
//#####################################################################
template<class TV> auto IMPLICIT_OBJECT_COMBINED<TV>::
Minimum_Cell_Size() const -> T
{
    return min(implicit_object1->Minimum_Cell_Size(),implicit_object2->Minimum_Cell_Size());
}
//#####################################################################
// Function Intersects
//#####################################################################
template<class TV> RAY_HIT<typename TV::SCALAR> IMPLICIT_OBJECT_COMBINED<TV>::
Intersects(const RAY<TV>& ray,const T isocontour) const
{
    PHYSBAM_NOT_IMPLEMENTED();
}
//#####################################################################
// Function All_Intersections
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
All_Intersections(ARRAY<RAY_HIT<T> >& hits,const RAY<TV>& ray,const T isocontour) const
{
    PHYSBAM_NOT_IMPLEMENTED();
}
//#####################################################################
// Function Read
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Read(TYPED_ISTREAM input)
{
    PHYSBAM_FATAL_ERROR();
}
//#####################################################################
// Function Write
//#####################################################################
template<class TV> void IMPLICIT_OBJECT_COMBINED<TV>::
Write(TYPED_OSTREAM output) const
{
    PHYSBAM_FATAL_ERROR();
}
//#####################################################################
}

