//#####################################################################
// Class SMOOTHING_KERNEL
//#####################################################################
#ifndef __SMOOTHING_KERNEL__
#define __SMOOTHING_KERNEL__
//Includes
#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/pow.h>
#include <Core/Vectors/VECTOR.h>
#include <Core/Vectors/VECTOR_2D.h>
namespace PhysBAM{

template<class TV>
class SMOOTHING_KERNEL
{
    typedef typename TV::SCALAR T;
//variables
public:
    T radius;
    T radius_sqr;
    T poly6_coefficient;
    T poly6_zero_weight;
    T poly6_gradient_coefficient;
    T poly6_laplacian_coefficient;
    T spiky_coefficient;
    T spiky_gradient_coefficient;
    T viscosity_coefficient;
    T viscosity_laplacian_coefficient;

    static const constexpr T coeff_a=TV::m==2?4/(T)pi:315/(64*(T)pi);
    
public:
    SMOOTHING_KERNEL(T radius_input=(T)1.0):radius(radius_input){
        Initialize();
    }
    ~SMOOTHING_KERNEL(){}

    //#####################################################################
    // Function Initialize
    //#####################################################################
    void Initialize(){
        radius_sqr=sqr(radius);
        poly6_coefficient=coeff_a/pow<6+TV::m>(radius);
        poly6_gradient_coefficient=coeff_a/pow<6+TV::m>(radius);
        poly6_laplacian_coefficient=coeff_a/pow<6+TV::m>(radius);
        spiky_coefficient=5*TV::m/(pi*pow<3+TV::m>(radius));
        spiky_gradient_coefficient=5*TV::m/(pi*pow<3+TV::m>(radius));
        viscosity_coefficient=5*TV::m/((5-TV::m)*pi*pow<TV::m>(radius));
        viscosity_laplacian_coefficient=(T)45.0/(pi*pow<6>(radius));
        poly6_zero_weight=Poly6_Smoothing_Kernel(0);
    }

    //#####################################################################
    // Function Reinitialize
    //#####################################################################
    void Reinitialize(const T radius_input){
        radius=radius_input;
        Initialize();
    }

    //#####################################################################
    // Function Poly6_Smoothing_Kernel
    //#####################################################################
    T Poly6_Smoothing_Kernel(const T distance)
    {
        //if(distance>radius) return 0;
        assert(distance<=radius);
        return poly6_coefficient*cube(sqr(radius)-sqr(distance));
    }
    
    //#####################################################################
    // Function Poly6_Smoothing_Kernel
    //#####################################################################
    TV Poly6_Smoothing_Kernel(const TV displacement)
    {
        T distance_sqr=displacement.Magnitude_Squared();
        //if(distance_sqr>radius_sqr || distance_sqr==0) return TV();
        assert(distance_sqr<=radius_sqr && distance_sqr!=0);
        return poly6_coefficient*cube(radius_sqr-distance_sqr)*displacement/sqrt(distance_sqr);
    }
    //#####################################################################
    // Function Inv_Poly6_Smoothing_Kernel
    //#####################################################################
    T Inv_Poly6_Smoothing_Kernel(T value)
    {
        PHYSBAM_ASSERT(TV::m==2);
        if(value<=0) return radius;
        if(value>poly6_coefficient*pow<6>(radius)) return 0;
        return sqrt(sqr(radius)-cbrt(value/poly6_coefficient));
    }

    //#####################################################################
    // Function Poly6_Smoothing_Kernel_Gradient
    //#####################################################################
    TV Poly6_Smoothing_Kernel_Gradient(const TV displacement)
    {
        T distance_squared=displacement.Magnitude_Squared();
        return (-6*poly6_gradient_coefficient*sqr(sqr(radius)-distance_squared))*displacement;
    }
    //#####################################################################
    // Function Poly6_Smoothing_Kernel_Laplacian
    //#####################################################################
    T Poly6_Smoothing_Kernel_Laplacian(const TV displacement)
    {
        T distance_squared=displacement.Magnitude_Squared();
        T tmp=sqr(radius)-distance_squared;
        return -6*poly6_laplacian_coefficient*tmp*(TV::m*tmp-4*distance_squared);
    }
    //#####################################################################
    // Function Spiky_Smoothing_Kernel
    //#####################################################################
    T Spiky_Smoothing_Kernel(const T distance)
    {
        // spiky
        return spiky_coefficient*cube(radius-distance);
    }
    //#####################################################################
    // Function Spiky_Smoothing_Kernel_Gradient
    //#####################################################################
    TV Spiky_Smoothing_Kernel_Gradient(const TV displacement)
    {
        T distance=displacement.Magnitude();
        if(distance<0.01*radius) distance=0.01*radius;
        else assert(distance<=radius);
        return -3*spiky_gradient_coefficient*sqr(radius-distance)*displacement/distance;
    }
    //#####################################################################
    // Function Viscosity_Smoothing_Kernel
    //#####################################################################
    T Viscosity_Smoothing_Kernel(const T distance)
    {
        assert(distance != 0);
        T distance_over_interaction_radius=distance/radius;
        return viscosity_coefficient*(-0.5*cube(distance_over_interaction_radius) +
            sqr(distance_over_interaction_radius) +
            0.5/distance_over_interaction_radius - 1);
    }
    //#####################################################################
    // Function Viscosity_Smoothing_Kernel_Laplacian
    //#####################################################################
    T Viscosity_Smoothing_Kernel_Laplacian(const T distance)
    {
        assert(distance != 0);
        return viscosity_laplacian_coefficient*(radius-distance);
    }

//#####################################################################
//other functions

private:
    
//#####################################################################
};
}
#endif 
