//#####################################################################
// Copyright 2021.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __REFLECTION_HELPER__
#define __REFLECTION_HELPER__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays/ARRAY_VIEW.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Math_Tools/RANGE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Core/Vectors/VECTOR.h>
#include <Grid_Tools/Grids/CELL_ITERATOR.h>
#include <Grid_Tools/Grids/FACE_INDEX.h>
#include <Grid_Tools/Grids/FACE_ITERATOR.h>
#include <Grid_Tools/Grids/FACE_RANGE_ITERATOR.h>
#include <Grid_Tools/Grids/GRID.h>
#include <Grid_Tools/Grids/NODE_ITERATOR.h>

namespace PhysBAM{

template<int d> VECTOR<int,d> Face_To_Fine_Node(const FACE_INDEX<d>& face)
{
    return (2*face.index+1).Add_Axis(face.axis,-1);
}

template<int d> VECTOR<int,d> Cell_To_Fine_Node(const VECTOR<int,d>& cell)
{
    return 2*cell+1;
}

template<int d> VECTOR<int,d> Node_To_Fine_Node(const VECTOR<int,d>& node)
{
    return 2*node;
}

template<int d> FACE_INDEX<d> Fine_Node_To_Face(const VECTOR<int,d>& node,int axis)
{
    return FACE_INDEX<d>(axis,(node.Add_Axis(axis,1)-1)/2);
}

template<int d> VECTOR<int,d> Fine_Node_To_Cell(const VECTOR<int,d>& node)
{
    return (node-1)/2;
}

template<int d> VECTOR<int,d> Fine_Node_To_Node(const VECTOR<int,d>& node)
{
    return node/2;
}

//#####################################################################
// Function Reflection_Cell_Neumann
// Mirrors the cell values in the interior of u_ghost to the ghost region.
//
// Assumes that the boundary values are all 0.
//#####################################################################
template<class TV,typename T> void Reflection_Cell_Neumann(const GRID<TV>& grid,ARRAY<T,VECTOR<int,TV::m> >& u_ghost,int ghost)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);
    
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        IV cell_df=Cell_To_Fine_Node(it.index);
        IV boundary_df=cell_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=cell_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Cell(reflected_df);
        u_ghost(it.index)=u_ghost(reflected_index);
    }
}
//#####################################################################
// Uses boundary to determine the derivative (normal to the boundary) at a boundary location
//#####################################################################
template<class TV,typename T> void Reflection_Cell_Neumann(const GRID<TV>& grid,ARRAY<T,VECTOR<int,TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int)> boundary)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);
    
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        IV cell_df=Cell_To_Fine_Node(it.index);
        IV boundary_df=cell_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=cell_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Cell(reflected_df);
        u_ghost(it.index)=u_ghost(reflected_index)+boundary(boundary_df,it.side)*grid.dX(axis)*abs(reflected_index(axis)-it.index(axis));
    }
}
//#####################################################################
// Function Reflection_Cell_NoSlip
// Mirrors and reflects the cell values in the interior of u_ghost to the ghost region around the boundary value.
//
// The boundary values are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//#####################################################################
template<class TV,typename T> void Reflection_Cell_NoSlip(const GRID<TV>& grid,ARRAY<T,VECTOR<int,TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int)> boundary)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);
    
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        IV cell_df=Cell_To_Fine_Node(it.index);
        IV boundary_df=cell_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=cell_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Cell(reflected_df);
        T boundary_value=boundary?boundary(grid_df.Node_Indices().Clamp(boundary_df),it.side):(T)0;
        u_ghost(it.index)=u_ghost(reflected_index)-2*(u_ghost(reflected_index)-boundary_value);
    }
}
//#####################################################################
// Function Reflection_Cell_Slip
// Mirrors and reflects the cell values in the interior of u_ghost to the ghost region around the boundary value.
// Assumes that the contents of the cells are vectors of type T.
//    Components of the interior vector that are normal to the boundary are reflected around the boundary value.
//    Components of the interior vector that are tangential to the boundary are mirrored.
//
// The boundary value components are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//    int: the component axis of the boundary value to return
//#####################################################################
template<class TV,typename T> void Reflection_Cell_Slip(const GRID<TV>& grid,ARRAY<VECTOR<T,TV::m>,VECTOR<int,TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int,int)> boundary)
{
    typedef VECTOR<int,TV::m> IV;
    typedef VECTOR<T,TV::m> TV2;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);
     
    for(RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        int cell_axis=axis;
        IV cell_df=Cell_To_Fine_Node(it.index);
        IV boundary_df=cell_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=cell_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Cell(reflected_df);
        T boundary_value=boundary?boundary(grid_df.Node_Indices().Clamp(boundary_df),it.side,cell_axis):(T)0;
        TV2 value=u_ghost(reflected_index);
        value(cell_axis)-=2*(u_ghost(reflected_index)(cell_axis)-boundary_value);
        u_ghost(it.index)=value;
    }
}
//#####################################################################
// Function Reflection_Node_Neumann
// Mirrors the node values in the interior of u_ghost to the ghost region.
//
// Assumes that the boundary values are all 0.
//
// Optionally enforces this boundary condition on all interior boundary nodes.
//#####################################################################
template<class TV,typename T> void Reflection_Node_Neumann(const GRID<TV>& grid,ARRAY<T,VECTOR<int,TV::m> >& u_ghost,int ghost,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(NODE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
            u_ghost(it.index)=0;
    
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices(ghost),grid.Node_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
    
        IV node_df=Node_To_Fine_Node(it.index);
        IV boundary_df=node_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=node_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Node(reflected_df);
        u_ghost(it.index)=u_ghost(reflected_index);
    }
}
//#####################################################################
// Function Reflection_Node_NoSlip
// Mirrors and reflects the node values in the interior of u_ghost to the ghost region around the boundary value.
//
// The boundary values are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//
// Optionally enforces the boundary values provided on the interior boundary nodes.
//#####################################################################
template<class TV,typename T> void Reflection_Node_NoSlip(const GRID<TV>& grid,ARRAY<T,VECTOR<int,TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int)> boundary,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(NODE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
            u_ghost(it.index)=boundary?boundary(Node_To_Fine_Node(it.index),it.side):(T)0;
    
    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices(ghost),grid.Node_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        IV node_df=Node_To_Fine_Node(it.index);
        IV boundary_df=node_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=node_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Node(reflected_df);
        T boundary_value=boundary?boundary(grid_df.Node_Indices().Clamp(boundary_df),it.side):(T)0;
        u_ghost(it.index)=u_ghost(reflected_index)-2*(u_ghost(reflected_index)-boundary_value);
    }
}
//#####################################################################
// Function Reflection_Node_Slip
// Mirrors and reflects the node values in the interior of u_ghost to the ghost region around the boundary value.
// Assumes that the contents of the nodes are vectors of type T.
//    Components of the interior vector that are normal to the boundary are reflected around the boundary value.
//    Components of the interior vector that are tangential to the boundary are mirrored.
//
// The boundary value components are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//    int: the component axis of the boundary value to return
//
// Optionally enforces the boundary values provided on the interior boundary nodes.
//#####################################################################
template<class TV,typename T> void Reflection_Node_Slip(const GRID<TV>& grid,ARRAY<VECTOR<T,TV::m>,VECTOR<int,TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int,int)> boundary,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    typedef VECTOR<T,TV::m> TV2;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(NODE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
        {
            TV2 value;
            if(boundary)
                for(int i=0;i<TV2::m;i++)
                    value(i)=boundary(Node_To_Fine_Node(it.index),it.side,i);
            u_ghost(it.index)=value;
        }

    for(RANGE_ITERATOR<TV::m> it(grid.Node_Indices(ghost),grid.Node_Indices(),RI::ghost|RI::delay_corners);it.Valid();it.Next())
    {
        int axis=it.side/2;
        int node_axis=axis;

        IV node_df=Node_To_Fine_Node(it.index);
        IV boundary_df=node_df;
        boundary_df(axis)=(it.side&1)?grid_df.Node_Indices().max_corner(axis)-1:grid_df.Node_Indices().min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=node_df(axis)-boundary_df(axis);
        IV reflected_index=Fine_Node_To_Node(reflected_df);
        T boundary_value=boundary?boundary(grid_df.Node_Indices().Clamp(boundary_df),it.side,node_axis):(T)0;
        TV2 value=u_ghost(reflected_index);
        value(node_axis)-=2*(u_ghost(reflected_index)(node_axis)-boundary_value);
        u_ghost(it.index)=value;
    }
}
//#####################################################################
// Function Reflection_Face_Neumann
// Mirrors the face values in the interior of u_ghost to the ghost region.
//
// Assumes that the boundary values are all 0.
//
// Optionally enforces this boundary condition on all interior boundary faces.
//#####################################################################
template<class TV,typename T> void Reflection_Face_Neumann(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<TV::m> >& u_ghost,int ghost,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
            u_ghost(it.face)=0;
    
    for(FACE_RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RF::ghost|RF::delay_corners|RF::skip_inner);it.Valid();it.Next())
    {
        int axis=it.side/2;
        int face_axis=it.face.axis;
        IV face_df=Face_To_Fine_Node(it.face);
        IV boundary_df=face_df;
        if(axis!=face_axis) boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis):grid_df.Face_Indices(face_axis).min_corner(axis);
        else boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis)-1:grid_df.Face_Indices(face_axis).min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=face_df(axis)-boundary_df(axis);
        FACE_INDEX<TV::m> reflected_index=Fine_Node_To_Face(reflected_df,face_axis);
        u_ghost(it.face)=u_ghost(reflected_index);
    }
}
//#####################################################################
// Function Reflection_Face_NoSlip
// Mirrors and reflects the face values in the interior of u_ghost to the ghost region around the boundary value.
//
// The boundary values are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//    int: the axis of the face index 
//
// Optionally enforces the boundary values provided on the interior boundary faces.
//#####################################################################
template<class TV,typename T> void Reflection_Face_NoSlip(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int,int)> boundary,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
            u_ghost(it.face)=boundary?boundary(Face_To_Fine_Node(it.face),it.side,it.face.axis):(T)0;

    for(FACE_RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RF::ghost|RF::delay_corners|RF::skip_inner);it.Valid();it.Next())
    {
        int axis=it.side/2;
        int face_axis=it.face.axis;

        IV face_df=Face_To_Fine_Node(it.face);
        IV boundary_df=face_df;
        if(axis!=face_axis) boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis):grid_df.Face_Indices(face_axis).min_corner(axis);
        else boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis)-1:grid_df.Face_Indices(face_axis).min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=face_df(axis)-boundary_df(axis);
        FACE_INDEX<TV::m> reflected_index=Fine_Node_To_Face(reflected_df,face_axis);
        T boundary_value=boundary?boundary(grid_df.Face_Indices(face_axis).Clamp(boundary_df),it.side,face_axis):(T)0;

        u_ghost(it.face)=u_ghost(reflected_index)-2*(u_ghost(reflected_index)-boundary_value);
    }
}
//#####################################################################
// Function Reflection_Face_Slip
// Mirrors and reflects the face values in the interior of u_ghost to the ghost region around the boundary value.
//    Values of the interior faces that are normal to the boundary are reflected around the boundary value.
//    Values of the interior faces that are tangential to the boundary are mirrored.
//
// The boundary values are provided by the user-provided function 'boundary'
// boundary parameters:
//    VECTOR<int,m>: boundary location index (into a double-fine version of the provided grid)
//    int: the side of the grid that the boundary location is on
//    int: the axis of the face index 
//
// Optionally enforces the boundary values provided on the interior boundary faces.
//#####################################################################
template<class TV,typename T> void Reflection_Face_Slip(const GRID<TV>& grid,ARRAY<T,FACE_INDEX<TV::m> >& u_ghost,int ghost,std::function<T(const VECTOR<int,TV::m>&,int,int)> boundary,bool enforce_boundary=false)
{
    typedef VECTOR<int,TV::m> IV;
    GRID<TV> grid_df(grid.numbers_of_cells*2,grid.domain);

    if(enforce_boundary)
        for(FACE_ITERATOR<TV> it(grid,0,BOUNDARY_REGION);it.Valid();it.Next())
            u_ghost(it.face)=boundary?boundary(Face_To_Fine_Node(it.face),it.side,it.face.axis):(T)0;

    for(FACE_RANGE_ITERATOR<TV::m> it(grid.Cell_Indices(ghost),grid.Cell_Indices(),RF::ghost|RF::delay_corners|RF::skip_inner);it.Valid();it.Next())
    {
        int axis=it.side/2;
        int face_axis=it.face.axis;
        IV face_df=Face_To_Fine_Node(it.face);
        IV boundary_df=face_df;
        if(axis!=face_axis) boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis):grid_df.Face_Indices(face_axis).min_corner(axis);
        else boundary_df(axis)=(it.side&1)?grid_df.Face_Indices(face_axis).max_corner(axis)-1:grid_df.Face_Indices(face_axis).min_corner(axis);
        IV reflected_df=boundary_df;
        reflected_df(axis)-=face_df(axis)-boundary_df(axis);
        FACE_INDEX<TV::m> reflected_index=Fine_Node_To_Face(reflected_df,face_axis);
        T boundary_value=boundary?boundary(grid_df.Face_Indices(face_axis).Clamp(boundary_df),it.side,face_axis):(T)0;

        T value=u_ghost(reflected_index);
        if(axis==face_axis) value-=2*(u_ghost(reflected_index)-boundary_value);
        u_ghost(it.face)=value;
    }
}
}
//#####################################################################
#endif
