//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __GENERAL_INTERPOLATION_AXIS_FACE__
#define __GENERAL_INTERPOLATION_AXIS_FACE__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Log/DEBUG_UTILITIES.h>
#include <Tools/Interpolation/INTERPOLATION_FORWARD.h>
#include <Grid_Tools/Arrays/FACE_ARRAYS.h>
#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_HELPERS.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_FACE.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_HELPERS.h>
namespace PhysBAM{

template<class TV,class T2,class STENCIL,bool const_extrap=false,class T_FACE_ARRAY=ARRAY<T2,FACE_INDEX<TV::m> > >
class GENERAL_INTERPOLATION_AXIS_FACE:public INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY>
{
    static_assert(TV::m>=1 && TV::m<=3);
    typedef typename TV::SCALAR T;typedef VECTOR<int,TV::m> IV;
    typedef VECTOR<T2,TV::m> TV2;typedef SYMMETRIC_MATRIX<T2,TV::m> SM2;
    typedef INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY> BASE;
    constexpr static int ss=STENCIL::stencil_size;
    using typename INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY>::REF;

  public:
    typedef T_FACE_ARRAY FACE_LOOKUP;
    STENCIL stencil;
    
    GENERAL_INTERPOLATION_AXIS_FACE() = default;
    GENERAL_INTERPOLATION_AXIS_FACE(const GENERAL_INTERPOLATION_AXIS_FACE&) = delete;
    void operator=(const GENERAL_INTERPOLATION_AXIS_FACE&) = delete;
    virtual ~GENERAL_INTERPOLATION_AXIS_FACE() = default;

    TV2 Interpolate_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        TV2 r;
        if constexpr(TV::m>0) r(0)=Interpolate<0>(stencil,grid,u.Component(0),p);
        if constexpr(TV::m>1) r(1)=Interpolate<1>(stencil,grid,u.Component(1),p);
        if constexpr(TV::m>2) r(2)=Interpolate<2>(stencil,grid,u.Component(2),p);
        return r;
    }
    T2 Interpolate_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        if constexpr(TV::m>0) if(TV::m==1 || axis==0) return Interpolate<0>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>1) if(TV::m==2 || axis==1) return Interpolate<1>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>2) return Interpolate<2>(stencil,grid,u.Component(axis),p);
        return T2();
    }

    VECTOR<VECTOR<T2,2>,TV::m> Extrema_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X) const
    {
        REF ref0(u0,X),ref1(u1,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u0.Domain_Indices(),X);
        VECTOR<VECTOR<T2,2>,TV::m> r;
        if constexpr(TV::m>0) r(0)=Extrema<ss,0>(grid,u0.Component(0),u1.Component(0),p);
        if constexpr(TV::m>1) r(1)=Extrema<ss,1>(grid,u0.Component(1),u1.Component(1),p);
        if constexpr(TV::m>2) r(2)=Extrema<ss,2>(grid,u0.Component(2),u1.Component(2),p);
        return r;
    }
    VECTOR<T2,2> Extrema_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X,int axis) const
    {
        REF ref0(u0,X),ref1(u1,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u0.Domain_Indices(),X);
        if constexpr(TV::m>0) if(TV::m==1 || axis==0) return Extrema<ss,0>(grid,u0.Component(axis),u1.Component(axis),p);
        if constexpr(TV::m>1) if(TV::m==2 || axis==1) return Extrema<ss,1>(grid,u0.Component(axis),u1.Component(axis),p);
        if constexpr(TV::m>2) return Extrema<ss,2>(grid,u0.Component(axis),u1.Component(axis),p);
        return VECTOR<T2,2>();
    }

    VECTOR<TV2,TV::m> Gradient_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        VECTOR<TV2,TV::m> r;
        if constexpr(TV::m>0) r(0)=Gradient<0>(stencil,grid,u.Component(0),p);
        if constexpr(TV::m>1) r(1)=Gradient<1>(stencil,grid,u.Component(1),p);
        if constexpr(TV::m>2) r(2)=Gradient<2>(stencil,grid,u.Component(2),p);
        return r;
    }
    TV2 Gradient_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        if constexpr(TV::m>0) if(TV::m==1 || axis==0) return Gradient<0>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>1) if(TV::m==2 || axis==1) return Gradient<1>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>2) return Gradient<2>(stencil,grid,u.Component(axis),p);
        return TV2();
    }

    VECTOR<SM2,TV::m> Hessian_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        VECTOR<SM2,TV::m> r;
        if constexpr(TV::m>0) r(0)=Hessian<0>(stencil,grid,u.Component(0),p);
        if constexpr(TV::m>1) r(1)=Hessian<1>(stencil,grid,u.Component(1),p);
        if constexpr(TV::m>2) r(2)=Hessian<2>(stencil,grid,u.Component(2),p);
        return r;
    }
    SM2 Hessian_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index<ss,const_extrap>(grid,at_cells,u.Domain_Indices(),X);
        if constexpr(TV::m>0) if(TV::m==1 || axis==0) return Hessian<0>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>1) if(TV::m==2 || axis==1) return Hessian<1>(stencil,grid,u.Component(axis),p);
        if constexpr(TV::m>2) return Hessian<2>(stencil,grid,u.Component(axis),p);
        return SM2();
    }

protected:
    virtual TV2 Interpolate_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Interpolate_Face(grid,u,X);}
    virtual T2 Interpolate_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Interpolate_Face(grid,u,X,axis);}
    virtual VECTOR<VECTOR<T2,2>,TV::m> Extrema_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X) const
    {return Extrema_Face(grid,u0,u1,X);}
    virtual VECTOR<T2,2> Extrema_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X,int axis) const
    {return Extrema_Face(grid,u0,u1,X,axis);}
    virtual VECTOR<TV2,TV::m> Gradient_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Gradient_Face(grid,u,X);}
    virtual TV2 Gradient_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Gradient_Face(grid,u,X,axis);}
    virtual VECTOR<SM2,TV::m> Hessian_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Hessian_Face(grid,u,X);}
    virtual SM2 Hessian_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Hessian_Face(grid,u,X,axis);}
//#####################################################################
};
}
#endif
