//#####################################################################
// Copyright 2021, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __GENERAL_INTERPOLATION_FACE__
#define __GENERAL_INTERPOLATION_FACE__

#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Log/DEBUG_UTILITIES.h>
#include <Tools/Interpolation/INTERPOLATION_FORWARD.h>
#include <Grid_Tools/Arrays/FACE_ARRAYS.h>
#include <Grid_PDE/Interpolation/GENERAL_INTERPOLATION_HELPERS.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_FACE.h>
#include <Grid_PDE/Interpolation/INTERPOLATION_HELPERS.h>
namespace PhysBAM{

template<class TV,class T2,class STENCIL,bool const_extrap=false,class T_FACE_ARRAY=ARRAY<T2,FACE_INDEX<TV::m> > >
class GENERAL_INTERPOLATION_FACE:public INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY>
{
    static_assert(TV::m>=1 && TV::m<=3);
    typedef typename TV::SCALAR T;typedef VECTOR<int,TV::m> IV;
    typedef VECTOR<T2,TV::m> TV2;typedef SYMMETRIC_MATRIX<T2,TV::m> SM2;
    typedef INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY> BASE;
    constexpr static int ss=STENCIL::stencil_size;
    using typename INTERPOLATION_FACE<TV,T2,T_FACE_ARRAY>::REF;

  public:
    typedef T_FACE_ARRAY FACE_LOOKUP;
    STENCIL stencil;
    
    GENERAL_INTERPOLATION_FACE() = default;
    GENERAL_INTERPOLATION_FACE(const GENERAL_INTERPOLATION_FACE&) = delete;
    void operator=(const GENERAL_INTERPOLATION_FACE&) = delete;
    virtual ~GENERAL_INTERPOLATION_FACE() = default;

    TV2 Interpolate_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X);
        TV2 r;
        for(int i=0;i<TV::m;i++) r(i)=Interpolate<-1>(stencil,grid,u.Component(i),p(i));
        return r;
    }
    T2 Interpolate_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X,axis);
        return Interpolate<-1>(stencil,grid,u.Component(axis),p);
    }

    VECTOR<VECTOR<T2,2>,TV::m> Extrema_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X) const
    {
        REF ref0(u0,X),ref1(u1,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u0.Domain_Indices(),X);
        VECTOR<VECTOR<T2,2>,TV::m> r;
        for(int i=0;i<TV::m;i++) r(i)=Extrema<ss,-1>(grid,u0.Component(i),u1.Component(i),p(i));
        return r;
    }
    VECTOR<T2,2> Extrema_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X,int axis) const
    {
        REF ref0(u0,X),ref1(u1,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u0.Domain_Indices(),X,axis);
        return Extrema<ss,-1>(grid,u0.Component(axis),u1.Component(axis),p);
    }

    VECTOR<TV2,TV::m> Gradient_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X);
        VECTOR<TV2,TV::m> r;
        for(int i=0;i<TV::m;i++) r(i)=Gradient<-1>(stencil,grid,u.Component(i),p(i));
        return r;
    }
    TV2 Gradient_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X,axis);
        return Gradient<-1>(stencil,grid,u.Component(axis),p);
    }

    VECTOR<SM2,TV::m> Hessian_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X);
        VECTOR<SM2,TV::m> r;
        for(int i=0;i<TV::m;i++) r(i)=Hessian<-1>(stencil,grid,u.Component(i),p(i));
        return r;
    }
    SM2 Hessian_Face(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {
        REF ref(u,X);
        auto p=INTERPOLATION_HELPERS::Clamped_Index_Face<ss,const_extrap>(grid,u.Domain_Indices(),X,axis);
        return Hessian<-1>(stencil,grid,u.Component(axis),p);
    }

protected:
    virtual TV2 Interpolate_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Interpolate_Face(grid,u,X);}
    virtual T2 Interpolate_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Interpolate_Face(grid,u,X,axis);}
    virtual VECTOR<VECTOR<T2,2>,TV::m> Extrema_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X) const
    {return Extrema_Face(grid,u0,u1,X);}
    virtual VECTOR<T2,2> Extrema_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u0,const T_FACE_ARRAY& u1,const TV& X,int axis) const
    {return Extrema_Face(grid,u0,u1,X,axis);}
    virtual VECTOR<TV2,TV::m> Gradient_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Gradient_Face(grid,u,X);}
    virtual TV2 Gradient_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Gradient_Face(grid,u,X,axis);}
    virtual VECTOR<SM2,TV::m> Hessian_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X) const
    {return Hessian_Face(grid,u,X);}
    virtual SM2 Hessian_Face_V(const GRID<TV>& grid,const T_FACE_ARRAY& u,const TV& X,int axis) const
    {return Hessian_Face(grid,u,X,axis);}
//#####################################################################
};
}
#endif
