//#####################################################################
// Copyright 2005-2007, Eran Guendelman, Geoffrey Irving, Nipun Kwatra, Avi Robinson-Mosher, Andrew Selle, Tamar Shinar.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
// Class FACE_ARRAYS
//#####################################################################
#ifndef __FACE_ARRAYS__
#define __FACE_ARRAYS__

#include <Core/Arrays/ARRAYS_FORWARD.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Arrays_Nd/ARRAYS_ND_VIEW.h>
#include <Core/Vectors/SCALAR_POLICY.h>
#include <Grid_Tools/Grids/FACE_INDEX.h>
namespace PhysBAM{

template<class T,int d>
class ARRAY<T,FACE_INDEX<d> >
{
    typedef VECTOR<T,d> TV;typedef VECTOR<int,d> IV;
    struct UNUSABLE{};
public:
    ARRAY_VIEW<T> array; // one-dimensional data storage

    enum {dimension=TV::m};
    template<class T2> struct REBIND{typedef ARRAY<T2,FACE_INDEX<dimension> > TYPE;};
    typedef T ELEMENT;typedef FACE_INDEX<d> INDEX;typedef T& RESULT_TYPE;typedef const T& CONST_RESULT_TYPE;
    typedef SCALAR_TYPE<T> SCALAR;
    typedef int HAS_UNTYPED_READ_WRITE;

    RANGE<IV> domain;
    VECTOR<ARRAY_VIEW<T,IV>,dimension> data;

    ARRAY()
    {}

    ARRAY(const RANGE<IV>& domain_input)
        :ARRAY(domain_input,init_new,T())
    {}

    ARRAY(const RANGE<IV>& domain_input,NO_INIT)
    {Initialize(domain_input);}

    ARRAY(const RANGE<IV>& domain_input,INIT_NEW,const T& initialization_value=T())
    {Initialize(domain_input);array.Fill(initialization_value);}
    
    ARRAY(const ARRAY& old_array)
    {Initialize(old_array.domain);array=old_array.array;}

    ARRAY(ARRAY&& old_array)
    {Exchange(old_array);}

    virtual ~ARRAY()
    {delete[] array.base_pointer;}

    template<class T2>
    ARRAY& operator*=(const T2 a)
    {for(int i=0;i<dimension;i++) data(i)*=a;
    return *this;}

    ARRAY& operator+=(const ARRAY& a)
    {data+=a.data;return *this;}

    ARRAY& operator-=(const ARRAY& a)
    {data-=a.data;return *this;}

    const RANGE<IV>& Domain_Indices() const
    {return domain;}

    ARRAY& operator=(const ARRAY& source)
    {if(source.Domain_Indices()!=Domain_Indices()) Resize(source.Domain_Indices(),no_init);Copy(source);return *this;}

    ARRAY& operator=(ARRAY&&) = default;

    void Resize(const RANGE<IV>& domain_input)
    {Resize(domain_input,true,true,T());}

    void Resize(const RANGE<IV>& domain_input,NO_INIT)
    {Resize(domain_input,false,false,T());}

    void Resize(const RANGE<IV>& domain_input,INIT_NEW,const T& initialization_value=T())
    {Resize(domain_input,true,true,initialization_value);}

    void Resize(const RANGE<IV>& domain_input,INIT_ALL,const T& initialization_value=T())
    {Resize(domain_input,true,false,initialization_value);}

    static int Calculate_Raw_Size(const RANGE<IV>& domain_input)
    {
        IV sizes=domain_input.Edge_Lengths();
        int total=0;
        for(int i=0;i<dimension;i++)
            total+=sizes.Add_Axis(i,1).Product();
        return total;
    }

    void Initialize_Using_Array(const RANGE<IV>& domain_input)
    {
        domain=domain_input;
        T* p=array.Get_Array_Pointer();
        for(int i=0;i<dimension;i++)
        {
            RANGE<IV> new_domain(domain_input);
            new_domain.max_corner(i)++;
            data(i).Set(p,new_domain);
            p+=data(i).array.m;
        }
    }

    void Initialize(const RANGE<IV>& domain_input)
    {
        int new_length=Calculate_Raw_Size(domain_input);
        assert(!array.Get_Array_Pointer());
        array.Set(new T[new_length],new_length);
        Initialize_Using_Array(domain_input);
    }

  protected:

    virtual void Resize(const RANGE<IV>& domain_input,const bool initialize_new_elements,const bool copy_existing_elements,const T& initialization_value)
    {
        if(domain_input==domain)
        {
            if(initialize_new_elements && !copy_existing_elements)
                array.Fill(initialization_value);
            return;
        }
        ARRAY new_array(domain_input);
        if(initialize_new_elements)
            new_array.Fill(initialization_value);
        
        if(copy_existing_elements)
            for(int i=0;i<dimension;i++)
                ARRAY_VIEW<T,IV>::Limited_Shifted_Get(new_array.data(i),data(i),IV());

        Exchange(new_array);
    }
  public:

    void Clean_Memory()
    {
        array.Set(0,0);
        domain=RANGE<IV>::Empty_Box();
        for(auto& a:data) a.Set({});
    }

    void Delete_Pointers_And_Clean_Memory()
    {for(int i=0;i<array.m;i++) delete array.base_pointer[i];Clean_Memory();}

    int Number_Of_Ghost_Cells() const
    {return -domain.min_corner.x;}

    T& operator()(const int axis,const IV& face)
    {return Component(axis)(face);}

    const T& operator()(const int axis,const IV& face) const
    {return Component(axis)(face);}

    T& operator()(const FACE_INDEX<dimension>& index)
    {return Component(index.axis)(index.index);}

    const T& operator()(const FACE_INDEX<dimension>& index) const
    {return Component(index.axis)(index.index);}

    bool Valid_Index(const FACE_INDEX<dimension>& index) const
    {return ((unsigned)index.axis<(unsigned)dimension) && Component(index.axis).Valid_Index(index.index);}

    int Standard_Index(const FACE_INDEX<TV::m>& index) const
    {assert((unsigned)index.axis<(unsigned)dimension);
    return data(index.axis).array.base_pointer-array.base_pointer+data(index.axis).Standard_Index(index.index);}

    ARRAY_VIEW<T,IV>& Component(const int axis)
    {assert((unsigned)axis<dimension);return data(axis);}

    const ARRAY_VIEW<T,IV>& Component(const int axis) const
    {assert((unsigned)axis<dimension);return data(axis);}

    T* begin(){return array.begin();}
    const T* begin() const {return array.begin();}
    T* end(){return array.end();}
    const T* end() const {return array.end();}

    TV Cell_Centered_Average(const IV& cell_index) const
    {TV average;
    for(int i=0;i<dimension;i++) average(i)=(T).5*(data(i)(cell_index)+data(i)(cell_index+IV::Axis_Vector(i)));
    return average;}

    void Set_All_Faces(const T& value,const IV& cell_index)
    {for(int i=0;i<dimension;i++) data(i)(cell_index)=data(i)(cell_index+IV::Axis_Vector(i))=value;}

    static void Extract_Dimension(const ARRAY& old_array,ARRAY<typename ELEMENT_OF_VECTOR<T>::TYPE,FACE_INDEX<dimension> >& extracted_array,const IV& dimensions_to_extract)
    {for(int i=0;i<dimension;i++) ARRAY_VIEW<T,IV>::Extract_Dimension(old_array.data(i),extracted_array.data(i),dimensions_to_extract(i));}

    void Fill(const T& constant)
    {for(int i=0;i<dimension;i++) data(i).Fill(constant);}

    void Copy(const ARRAY& old_copy)
    {for(int i=0;i<dimension;i++) data(i).Copy(old_copy.data(i));}

    template<class T2>
    void Copy(const T2 c,const ARRAY& old)
    {for(int i=0;i<dimension;i++) data(i).Copy(c,old.data(i));}

    void Fill(const TV& value)
    {for(int i=0;i<dimension;i++) data(i).Fill(value.x);}

    template<class T2>
    void Copy(const T2 c1,const ARRAY& v1,const T2 c2,const ARRAY& v2)
    {for(int i=0;i<dimension;i++) data(i).Copy(c1,v1.data(i),c2,v2.data(i));}

    static void Put(const ARRAY& old_copy,ARRAY& new_copy)
    {for(int i=0;i<dimension;i++) ARRAY_VIEW<T,IV>::Put(old_copy.data(i),new_copy.data(i));}

    TV Max_Abs() const
    {TV maxabs_vector;
    for(int i=0;i<dimension;i++) maxabs_vector(i)=data(i).Max_Abs();
    return maxabs_vector;}

    static void Exchange(ARRAY& a,ARRAY& b)
    {a.Exchange(b);}

    void Exchange(ARRAY& a)
    {exchange(domain,a.domain);array.Exchange(a.array);
    for(int i=0;i<dimension;i++) ARRAY_VIEW<T,IV>::Exchange(data(i),a.data(i));}

    template<class T_INDICES>
    INDIRECT_ARRAY<ARRAY,T_INDICES&> Subset(const T_INDICES& indices)
    {return INDIRECT_ARRAY<ARRAY,T_INDICES&>(*this,indices);}

    template<class T_INDICES>
    INDIRECT_ARRAY<const ARRAY,T_INDICES&> Subset(const T_INDICES& indices) const
    {return INDIRECT_ARRAY<const ARRAY,T_INDICES&>(*this,indices);}

    template<class RW> void Read(std::istream& input)
    {Clean_Memory();Read_Binary<RW>(input,domain);Read_Binary<RW>(input,array.m);
    if(array.m<0){
        char buff[100];
        sprintf(buff,"Invalid negative array size %d",array.m);
        throw READ_ERROR(buff);}
    if(!array.m) return;
    array.base_pointer=new T[array.m];
    Read_Binary_Array<RW>(input,array.base_pointer,array.m);
    T* p_start=array.base_pointer;
    for(int i=0;i<d;i++){
        RANGE<IV> domain_input;
        Read_Binary<RW>(input,domain_input);
        ARRAY_VIEW<T,IV> array_new(p_start,domain_input);
        ARRAY_VIEW<T,IV>::Exchange(array_new,data(i));
        p_start+=(domain_input.Edge_Lengths()).Product();}}

    template<class RW> void Write(std::ostream& output) const
    {Write_Binary<RW>(output,domain);
    Write_Binary<RW>(output,array.m);
    Write_Binary_Array<RW>(output,array.base_pointer,array.m);
    for(int i=0;i<d;i++) Write_Binary<RW>(output,data(i).domain);}
//#####################################################################
};

template<class T> inline std::ostream& operator<<(std::ostream& output_stream,const ARRAY<T,FACE_INDEX<1> >& a)
{
    for(int i=a.domain.min_corner.x;i<a.domain.max_corner.x+1;i++)
        output_stream<<a.Component(0)(i)<<" ";
    output_stream<<std::endl;
    return output_stream;
}

template<class T,int d> inline std::ostream& operator<<(std::ostream& output_stream,const ARRAY<T,FACE_INDEX<d> >& a)
{
    for(int i=0;i<d;i++) output_stream<<a.data(i)<<std::endl;
    return output_stream;
}

extern template ARRAY<bool,FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<bool,FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<bool,FACE_INDEX<3> >::~ARRAY();
extern template ARRAY<int,FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<int,FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<int,FACE_INDEX<3> >::~ARRAY();
extern template ARRAY<float,FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<float,FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<float,FACE_INDEX<3> >::~ARRAY();
extern template ARRAY<double,FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<double,FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<double,FACE_INDEX<3> >::~ARRAY();

extern template ARRAY<VECTOR<double,1>, FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<VECTOR<double,2>, FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<VECTOR<double,3>, FACE_INDEX<3> >::~ARRAY();
extern template ARRAY<VECTOR<float,1>, FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<VECTOR<float,2>, FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<VECTOR<float,3>, FACE_INDEX<3> >::~ARRAY();

extern template ARRAY<VECTOR<double,3>, FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<VECTOR<double,4>, FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<VECTOR<double,5>, FACE_INDEX<3> >::~ARRAY();
extern template ARRAY<VECTOR<float,3>, FACE_INDEX<1> >::~ARRAY();
extern template ARRAY<VECTOR<float,4>, FACE_INDEX<2> >::~ARRAY();
extern template ARRAY<VECTOR<float,5>, FACE_INDEX<3> >::~ARRAY();

extern template void ARRAY<bool, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const bool&);
extern template void ARRAY<bool, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const bool&);
extern template void ARRAY<bool, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const bool&);
extern template void ARRAY<int, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const int&);
extern template void ARRAY<int, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const int&);
extern template void ARRAY<int, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const int&);
extern template void ARRAY<float, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const float&);
extern template void ARRAY<float, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const float&);
extern template void ARRAY<float, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const float&);
extern template void ARRAY<double, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const double&);
extern template void ARRAY<double, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const double&);
extern template void ARRAY<double, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const double&);

extern template void ARRAY<VECTOR<double,1>, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const VECTOR<double,1>&);
extern template void ARRAY<VECTOR<double,2>, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const VECTOR<double,2>&);
extern template void ARRAY<VECTOR<double,3>, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const VECTOR<double,3>&);
extern template void ARRAY<VECTOR<float,1>, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const VECTOR<float,1>&);
extern template void ARRAY<VECTOR<float,2>, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const VECTOR<float,2>&);
extern template void ARRAY<VECTOR<float,3>, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const VECTOR<float,3>&);

extern template void ARRAY<VECTOR<double,3>, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const VECTOR<double,3>&);
extern template void ARRAY<VECTOR<double,4>, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const VECTOR<double,4>&);
extern template void ARRAY<VECTOR<double,5>, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const VECTOR<double,5>&);
extern template void ARRAY<VECTOR<float,3>, FACE_INDEX<1> >::Resize(const RANGE<VECTOR<int, 1> >&, bool, bool, const VECTOR<float,3>&);
extern template void ARRAY<VECTOR<float,4>, FACE_INDEX<2> >::Resize(const RANGE<VECTOR<int, 2> >&, bool, bool, const VECTOR<float,4>&);
extern template void ARRAY<VECTOR<float,5>, FACE_INDEX<3> >::Resize(const RANGE<VECTOR<int, 3> >&, bool, bool, const VECTOR<float,5>&);

extern template void ARRAY<int, FACE_INDEX<1> >::Read<float>(std::istream&);
extern template void ARRAY<int, FACE_INDEX<2> >::Read<float>(std::istream&);
extern template void ARRAY<int, FACE_INDEX<3> >::Read<float>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<1> >::Read<float>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<2> >::Read<float>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<3> >::Read<float>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<1> >::Read<float>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<2> >::Read<float>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<3> >::Read<float>(std::istream&);
extern template void ARRAY<int, FACE_INDEX<1> >::Read<double>(std::istream&);
extern template void ARRAY<int, FACE_INDEX<2> >::Read<double>(std::istream&);
extern template void ARRAY<int, FACE_INDEX<3> >::Read<double>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<1> >::Read<double>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<2> >::Read<double>(std::istream&);
extern template void ARRAY<float, FACE_INDEX<3> >::Read<double>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<1> >::Read<double>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<2> >::Read<double>(std::istream&);
extern template void ARRAY<double, FACE_INDEX<3> >::Read<double>(std::istream&);

extern template void ARRAY<int, FACE_INDEX<1> >::Write<float>(std::ostream&) const;
extern template void ARRAY<int, FACE_INDEX<2> >::Write<float>(std::ostream&) const;
extern template void ARRAY<int, FACE_INDEX<3> >::Write<float>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<1> >::Write<float>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<2> >::Write<float>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<3> >::Write<float>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<1> >::Write<float>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<2> >::Write<float>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<3> >::Write<float>(std::ostream&) const;
extern template void ARRAY<int, FACE_INDEX<1> >::Write<double>(std::ostream&) const;
extern template void ARRAY<int, FACE_INDEX<2> >::Write<double>(std::ostream&) const;
extern template void ARRAY<int, FACE_INDEX<3> >::Write<double>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<1> >::Write<double>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<2> >::Write<double>(std::ostream&) const;
extern template void ARRAY<float, FACE_INDEX<3> >::Write<double>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<1> >::Write<double>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<2> >::Write<double>(std::ostream&) const;
extern template void ARRAY<double, FACE_INDEX<3> >::Write<double>(std::ostream&) const;
}
#endif
