//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Hybrid_Methods/Examples_And_Drivers/MPM_DRIVER.h>
#include <Core/Log/DEBUG_SUBSTEPS.h>
#include <Core/Log/FINE_TIMER.h>
#include <Core/Log/LOG.h>
#include <Core/Log/SCOPE.h>
#include <Core/Math_Tools/RANGE_ITERATOR.h>
#include <Tools/Parallel_Computation/APPEND_HOLDER.h>
#include <Grid_Tools/Grids/NODE_ITERATOR_THREADED.h>
#include <Grid_PDE/Boundaries/PERIODIC_BOUNDARY_HELPER.h>
#include <Geometry/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Geometry/Level_Sets/SMOOTHING_KERNEL.h>
#include <Deformables/Constitutive_Models/ST_VENANT_KIRCHHOFF_HENCKY_STRAIN.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
#include <Solids/Solids/SOLID_BODY_COLLECTION.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_EXAMPLE.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_TRANSFERS.h>
#include <Hybrid_Methods/Forces/MPM_FINITE_ELEMENTS.h>
#include <Hybrid_Methods/Forces/MPM_FORCE_HELPER.h>
#include <Hybrid_Methods/Forces/MPM_PLASTIC_FINITE_ELEMENTS.h>
#include <Hybrid_Methods/Forces/MPM_PLASTICITY_MODEL.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_ITERATOR.h>
#include <Hybrid_Methods/Regions/REGION_ARRAY.h>
#include <fstream>
#ifdef USE_OPENMP
#include <omp.h>
#endif
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> MPM_DRIVER<TV>::
MPM_DRIVER(MPM_EXAMPLE<TV>& example)
    :example(example)
{
    DEBUG_SUBSTEPS::write_substeps_level=example.write_substeps_level;
    DEBUG_SUBSTEPS::writer=[this](const std::string& title){Write_Substep(title);};
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> MPM_DRIVER<TV>::
~MPM_DRIVER()
{
    DEBUG_SUBSTEPS::writer=0;
}
//#####################################################################
// Execute_Main_Program
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Execute_Main_Program()
{
    Step([this](){Initialize();},"initialize");
    Simulate_To_Frame(example.last_frame);
    FINE_TIMER::Dump_Timing_Info();
}
//#####################################################################
// Initialize
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Initialize()
{
    TIMER_SCOPE_FUNC;
    LOG::cout<<std::setprecision(16)<<std::endl;
    DEBUG_SUBSTEPS::write_substeps_level=example.substeps_delay_frame<0?example.write_substeps_level:-1;

    if(example.auto_restart){
        example.viewer_dir.Read_Last_Frame(0);
        example.restart=example.viewer_dir.frame_stack(0);}
    else if(example.restart)
        example.viewer_dir.Set(example.restart);
    if(example.restart) example.viewer_dir.Make_Common_Directory(true);
    current_frame=example.restart;
    example.time=current_frame*example.frame_dt;

    example.Initialize();
    for(int i=0;i<TV::m;i++){
        bool a=example.side_bc_type(i*2)==MPM::BC_PERIODIC;
        bool b=example.side_bc_type(i*2+1)==MPM::BC_PERIODIC;
        PHYSBAM_ASSERT(a==b);
        example.is_periodic(i)=a;}

    if(example.restart)
        example.Read_Output_Files();

    example.velocity.u.Resize(REGION_ID(1));
    example.velocity_save.u.Resize(REGION_ID(1));
    example.particles.Store_C(example.opt.use_affine);
    example.particles.Store_S(example.opt.use_S);

    // TODO
    // for(CELL_ITERATOR<TV> iterator(example.grid,example.ghost,GHOST_REGION);iterator.Valid();iterator.Next()){
    //     for(int k=0;k<fluid_walls.m;k++){
    //         if(fluid_walls(k)->Lazy_Inside(iterator.Location())){
    //             cell_soli....
    Update_Simulated_Particles();

    if(!example.restart && !example.suppress_output) Write_Output_Files();
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after init",1);
}
//#####################################################################
// Function Advance_One_Time_Step
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Advance_One_Time_Step()
{
    TIMER_SCOPE_FUNC;
    Step([this](){Update_Simulated_Particles();},"simulated-particles");
    Print_Particle_Stats("particle state",example.dt);
    Step([this](){Update_Particle_Weights();},"update-weights",false);
    Step([this](){example.gather_scatter.Prepare_Scatter(example.particles);},"prepare-scatter",false);
    Step([this](){Compute_Regions();},"compute-regions",false);
    Step([this](){Particle_To_Grid();},"p2g");
    Print_Grid_Stats("after particle to grid",example.dt,example.velocity,0);
    Print_Energy_Stats("after particle to grid",example.velocity);
    Step([this](){Apply_Forces();},"forces");
    Print_Grid_Stats("after forces",example.dt,example.velocity,&example.velocity_save);
    Step([this](){Collide_Regions();},"collide-regions");
    Step([this](){Reduce_Dt();},"reduce-dt",false);
    Step([this](){Grid_To_Particle();},"g2p",true);
    Step([this](){Update_Plasticity_And_Hardening();},"plasticity-hardening",false);
}
//#####################################################################
// Function Simulate_To_Frame
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Simulate_To_Frame(const int frame)
{
    TIMER_SCOPE_FUNC;
    for(;current_frame<frame;current_frame++){
        LOG::SCOPE scope("FRAME","frame %d",current_frame+1);
        for(int i=0;i<example.begin_frame.m;i++)
            example.begin_frame(i)(current_frame);
        if(example.substeps_delay_frame==current_frame)
            DEBUG_SUBSTEPS::write_substeps_level=example.write_substeps_level;
        T time_at_frame=example.time+example.frame_dt;
        bool done=false;
        for(int substep=0;!done;substep++){
            LOG::SCOPE scope("SUBSTEP","substep %d",substep+1);
            example.dt=Compute_Dt();
            example.dt=clamp(example.dt,example.min_dt,example.max_dt);
            T next_time=example.time+example.dt;
            if(next_time+(T).0001*example.dt>time_at_frame){ // Account for roundoff errors
                next_time=time_at_frame;
                done=true;}
            else if(next_time+example.dt>time_at_frame) next_time=(example.time+time_at_frame)/2;
            example.dt=next_time-example.time;
            T original_dt=example.dt;
            LOG::cout<<"substep dt: "<<example.dt<<std::endl;
            PHYSBAM_ASSERT(example.dt>0);

            Step([this](){Advance_One_Time_Step();},"time-step");
            LOG::cout<<"actual dt: "<<example.dt<<std::endl;

            // Time step was reduced
            if(example.dt<original_dt){
                LOG::printf("dt reduced: %g to %g  (%g)\n",original_dt,example.dt,original_dt-example.dt);
                next_time=example.time+example.dt;
                done=false;}

            PHYSBAM_DEBUG_WRITE_SUBSTEP("end substep %i",0,substep);
            example.time=next_time;}
        for(int i=0;i<example.end_frame.m;i++)
            example.end_frame(i)(current_frame);
        if(!example.suppress_output)Write_Output_Files();}
}
//#####################################################################
// Function Write_Substep
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Write_Substep(const std::string& title)
{
    TIMER_SCOPE_FUNC;
    example.frame_title=title;
    LOG::printf("Writing substep [%s]: output_number=%P, time=%g, frame=%i\n",
        example.frame_title,example.viewer_dir.frame_stack,example.time,current_frame);
    Write_Output_Files();
    example.frame_title="";
}
//#####################################################################
// Write_Output_Files
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Write_Output_Files()
{
    TIMER_SCOPE_FUNC;
    LOG::SCOPE scope("Write_Output_Files");
    example.viewer_dir.Start_Directory(0,example.frame_title);
    example.frame_title="";
    example.Write_Output_Files();
    std::ofstream fout(example.viewer_dir.Frame_File("timing_info.txt"));
    FINE_TIMER::Dump_Timing_Info(fout);
    example.viewer_dir.Finish_Directory();
}
//#####################################################################
// Function Update_Particle_Weights
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Update_Particle_Weights()
{
    TIMER_SCOPE_FUNC;
    example.weights->Update(example.particles.X);
}
//#####################################################################
// Function Particle_To_Grid
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Particle_To_Grid()
{
    TIMER_SCOPE_FUNC;
    MPM_PARTICLES<TV>& particles=example.particles;
    bool use_normals=example.mass.u.m>REGION_ID(1);
    
#pragma omp parallel for
    for(REGION_FLAT_ID i=REGION_FLAT_ID(0);i<example.mass.Flat_Size();i++)
    {
        example.mass(i)=0;
        example.velocity(i)=TV();
        example.velocity_save(i)=TV();
        example.velocity_friction_save(i)=TV();
        if(use_normals) example.normal(i)=TV();
    }

    auto depth_field=example.particles.template Get_Array<T>("depth");
    ARRAY_VIEW<T> normal_field=depth_field?*depth_field:particles.mass;

    auto [ng,np]=Transfer_P2G(example.opt);
    example.gather_scatter.Scatter(use_normals||ng,np,0,0,
        [this,&particles,normal_field,use_normals](int p,const PARTICLE_GRID_ITERATOR<TV>& it_,int data)
        {
            auto r=example.particles.Region(p);
            Transfer_P2G(example.grid,example.opt,it_,
                example.mass.u(r),example.velocity.u(r),
                particles.mass(p),particles.V(p),particles.X(p),
                example.opt.use_affine?&particles.C(p):nullptr);
            if(use_normals)
                for(auto it(it_);it.Valid();it.Next())
                    example.normal({r,it.Index()})+=it.Gradient()*normal_field(p);
        });
    for(auto& a:example.mass.u) Fix_Periodic_Accum(a,PBH::ADD());
    if(use_normals)
        for(auto& a:example.normal.u)
            Fix_Periodic_Accum(a,PBH::ADD());
    for(auto& a:example.velocity.u) Fix_Periodic_Accum(a,PBH::ADD());
    
    example.valid_grid_indices.Remove_All();
    example.valid_grid_node_indices.Remove_All();
    PHYSBAM_DEBUG_WRITE_SUBSTEP("momentum",1);
    if(example.reflection_bc && !example.use_full_reflection)
    {
        Reflect_Boundary_Mass_Momentum();
        PHYSBAM_DEBUG_WRITE_SUBSTEP("after reflect m mv",1);
    }

    APPEND_HOLDER<REGION_FLAT_ID> flat_h(example.valid_grid_indices);
    APPEND_HOLDER<REGION_INDEX<TV::m> > indices_h(example.valid_grid_node_indices);
#pragma omp parallel
    {
#pragma omp single
        {
            flat_h.Init();
            indices_h.Init();
        }
#pragma omp barrier
        ARRAY<REGION_FLAT_ID>& flat_t=flat_h.Array();
        ARRAY<REGION_INDEX<TV::m> >& indices_t=indices_h.Array();


        for(REGION_ID r(0);r<example.location.u.m;r++)
        {
            auto& a=example.location.u(r);
            for(NODE_ITERATOR_THREADED<TV> it(example.grid,a.domain);it.Valid();it.Next()){
                REGION_INDEX<TV::m> ri={r,it.index};
                REGION_FLAT_ID fi=example.mass.Flat_Index(ri);
            if(example.mass(fi)){
                flat_t.Append(fi);
                indices_t.Append(ri);
                TV v=example.velocity(fi)/example.mass(fi);
                example.velocity(fi)=v;
                example.velocity_save(fi)=v;
                if(use_normals) example.normal(fi).Normalize();}
            else example.velocity(fi)=TV();}
        }
    }
    flat_h.Combine();
    indices_h.Combine();
}

template<class U,class F,class BF,class IV,class EX>
void Visit_Nodes(EX& example,RI flags,const REGION_ARRAY<U,IV>& a,F f,BF bf)
{
    IV ranges[2]={IV(),example.grid.numbers_of_cells};
    int side_mask=0;
    for(int i=0;i<IV::m;i++)
        if(example.side_bc_type(2*i)!=MPM::BC_PERIODIC)
            side_mask|=3<<(2*i);
    for(RANGE_ITERATOR<IV::m> it(example.grid.Node_Indices(),example.ghost,0,
            RI::ghost|RI::side_mask|flags,side_mask);it.Valid();it.Next_Side())
    {
        for(REGION_ID r(0);r<a.u.m;r++)
        {
            RANGE<IV> range=a.u(r).domain.Intersect(it.Current_Region());
            if(range.Empty()) continue;
            for(RANGE_ITERATOR<IV::m> ri(range);ri.Valid();ri.Next())
            {
                REGION_INDEX<IV::m> in{r,ri.index},out=in;
                int axis=it.side/2,side=it.side%2;
                out.i(axis)=2*ranges[side](axis)-in.i(axis);
                if(!example.mass.Valid_Index(out)) continue;
                f(it.side,in,out);
            }
        }
    }
    for(RANGE_ITERATOR<IV::m> it(example.grid.Node_Indices(),0,-1,
            RI::ghost|RI::side_mask|flags,side_mask);it.Valid();it.Next_Side())
    {
        for(REGION_ID r(0);r<a.u.m;r++)
        {
            RANGE<IV> range=a.u(r).domain.Intersect(it.Current_Region());
            if(range.Empty()) continue;
            for(RANGE_ITERATOR<IV::m> ri(range);ri.Valid();ri.Next())
                bf(it.side,REGION_INDEX<IV::m>{r,ri.index});
        }
    }
}

//#####################################################################
// Function Reflection_Boundary_Mass_Momentum
//#####################################################################
template <class TV> void MPM_DRIVER<TV>::
Reflect_Boundary_Mass_Momentum()
{
    TIMER_SCOPE_FUNC;
    Visit_Nodes(example,RI::duplicate_corners,example.mass,
        [this](int side,const auto& out,const auto& in)
        {
            auto bc_type=example.side_bc_type(side);
            T& m_i=example.mass(in),&m_o=example.mass(out),m_t=m_i+m_o;
            TV& mv_i=example.velocity(in),&mv_o=example.velocity(out);
            TV mv_id=mv_i-mv_o,mv_od=-mv_id,mv_in=mv_i+mv_o,mv_on=mv_in;
            if(bc_type!=MPM::BC_FREE && example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Node_X(out.i));
                TV bc=example.bc_velocity(nearest_point,example.time);
                mv_id+=2*m_o*bc;
                mv_od+=2*m_i*bc;}
            if(bc_type==MPM::BC_SLIP){
                int axis=side/2;
                mv_in(axis)=mv_id(axis);
                mv_on(axis)=mv_od(axis);}
            if(bc_type==MPM::BC_NOSLIP){
                mv_i=mv_id;
                mv_o=mv_od;}
            else{
                mv_i=mv_in;
                mv_o=mv_on;}
            m_o=m_i=m_t;
        },
        [this](int side,const auto& out)
        {
            auto bc_type=example.side_bc_type(side);
            if(bc_type==MPM::BC_FREE) return;
            TV& mv_o=example.velocity(out);
            TV mv_od;
            if(example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Node_X(out.i));
                TV bc=example.bc_velocity(nearest_point,example.time);
                mv_od=example.mass(out)*bc;}
            if(bc_type==MPM::BC_SLIP)
            {
                int axis=side/2;
                mv_o(axis)=mv_od(axis);
            }
            else mv_o=mv_od;
        });
}
template<class TV,class T,class TYPE>
pair<TV,TV> Reflect_Helper(T m_i,T m_o,TV v_i,TV v_o,TYPE type,T mu,int a,int s,TV bc)
{
    T m_t=m_i+m_o,a_i=m_i/m_t,a_o=1-a_i;
    TV v_in=a_i*v_i+a_o*v_o,v_on=v_in;
    if(type==MPM::BC_FREE) return {v_in,v_on};

    TV v_id=a_i*v_i+a_o*(2*bc-v_o),v_od=a_o*v_o+a_i*(2*bc-v_i);
    if(type==MPM::BC_NOSLIP) return {v_id,v_od};

    if(type==MPM::BC_SLIP){
        v_in(a)=v_id(a);
        v_on(a)=v_od(a);
        return {v_in,v_on};}

    v_in-=bc;
    v_id-=bc;
    T d_v=2*a_o*a_i*(2*bc(a)-v_i(a)-v_o(a));
    T d_v_n=s?-d_v:d_v;
    if(d_v_n<0) return {v_i,v_o};

    TV v_id_n;
    v_id_n(a)=v_id(a);
    v_in(a)=0;

    T d_v_t_mag=v_in.Normalize();
    T r=d_v_t_mag-mu*d_v_n;
    if(r<0) return {v_id_n+bc,bc-v_id_n};
    return {v_id_n+r*v_in+bc,bc-v_id_n+r*v_in};
}

template<class TV,class T,class TYPE>
TV Reflect_Helper(TV v,TYPE type,T mu,int a,int s,TV bc)
{
    if(type==MPM::BC_FREE) return v;
    if(type==MPM::BC_NOSLIP) return bc;
    if(type==MPM::BC_SLIP){
        v(a)=bc(a);
        return v;}

    T d_v=bc(a)-v(a);
    T d_v_n=s?-d_v:d_v;
    if(d_v_n<0) return v;

    v-=bc;
    v(a)=0;

    T d_v_t_mag=v.Normalize();
    T r=d_v_t_mag-mu*d_v_n;
    if(r<0) return bc;
    return bc+r*v;
}

//#####################################################################
// Function Reflect_Boundary_Friction
//#####################################################################
template <class TV> void MPM_DRIVER<TV>::
Reflect_Boundary_Friction(const REGION_ARRAY<T,IV>& mass,REGION_ARRAY<TV,IV>& u) const
{
    TIMER_SCOPE_FUNC;
    Visit_Nodes(example,RI::duplicate_corners,example.mass,
        [this](int side,const auto& out,const auto& in)
        {
            auto bc_type=example.side_bc_type(side);
            T m_i=example.mass(in),m_o=example.mass(out);
            if(!m_i || !m_o) return;
            TV& v_i=example.velocity(in),&v_o=example.velocity(out);
            TV bc;
            if(bc_type!=MPM::BC_FREE && example.bc_velocity)
                bc=example.bc_velocity(example.grid.domain.Clamp(example.grid.Node_X(out.i)),example.time);
            auto pr=Reflect_Helper(m_i,m_o,v_i,v_o,bc_type,example.reflection_bc_friction,side/2,side%2,bc);
            v_i=pr.first;
            v_o=pr.second;
        },
        [this](int side,const auto& out)
        {
            auto bc_type=example.side_bc_type(side);
            T m=example.mass(out);
            if(!m) return;
            TV &v_o=example.velocity(out);
            TV bc;
            if(bc_type!=MPM::BC_FREE && example.bc_velocity)
                bc=example.bc_velocity(example.grid.domain.Clamp(example.grid.Node_X(out.i)),example.time);
            v_o=Reflect_Helper(v_o,bc_type,example.reflection_bc_friction,side/2,side%2,bc);
        });
    PHYSBAM_DEBUG_WRITE_SUBSTEP("velocities after boundary collisions",1);
}
//#####################################################################
// Function Reflect_Boundary_Velocity
//#####################################################################
template <class TV> void MPM_DRIVER<TV>::
Reflect_Boundary_Velocity(REGION_ARRAY<TV,IV>& u)
{
    TIMER_SCOPE_FUNC;
    if(!example.reflection_bc) return;
    Visit_Nodes(example,RI::duplicate_corners,example.mass,
        [this,&u](int side,const auto& out,const auto& in)
        {
            auto bc_type=example.side_bc_type(side);
            if(!example.mass(out)) return;
            TV v_i=u(in),&v_o=u(out);
            TV v_od=-v_i,v_on=v_i;
            if(bc_type!=MPM::BC_FREE && example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Node_X(in.i));
                TV bc=example.bc_velocity(nearest_point,example.time);
                v_od+=bc*2;}
            int axis=side/2;
            if(bc_type==MPM::BC_SLIP) v_on(axis)=v_od(axis);
            if(bc_type==MPM::BC_NOSLIP) v_o=v_od;
            else v_o=v_on;
        },
        [this,&u](int side,const auto& out)
        {
            auto bc_type=example.side_bc_type(side);
            if(bc_type==MPM::BC_FREE) return;
            if(!example.mass(out)) return;
            TV &v_o=u(out);
            TV bc;
            if(example.bc_velocity){
                TV nearest_point=example.grid.domain.Clamp(example.grid.Node_X(out.i));
                bc=example.bc_velocity(nearest_point,example.time);}
            int axis=side/2;
            if(bc_type==MPM::BC_SLIP) v_o(axis)=bc(axis);
            else v_o=bc;
        });
}
//#####################################################################
// Function Reflect_Boundary_Force
//#####################################################################
template <class TV> void MPM_DRIVER<TV>::
Reflect_Boundary_Force(REGION_ARRAY<TV,IV>& force)
{
    TIMER_SCOPE_FUNC;
    if(!example.reflection_bc) return;
    Visit_Nodes(example,RI::duplicate_corners,example.mass,
        [this,&force](int side,const auto& out,const auto& in)
        {
            auto bc_type=example.side_bc_type(side);
            TV& f_i=force(in),&f_o=force(out);
            TV f_id=f_i-f_o,f_od=-f_id,f_in=f_i+f_o,f_on=f_in;
            if(bc_type==MPM::BC_SLIP){
                int axis=side/2;
                f_in(axis)=f_id(axis);
                f_on(axis)=f_od(axis);}
            if(bc_type==MPM::BC_NOSLIP){
                f_i=f_id;
                f_o=f_od;}
            else{
                f_i=f_in;
                f_o=f_on;}
        },
        [this,&force](int side,const auto& out)
        {
            auto bc_type=example.side_bc_type(side);
            if(bc_type==MPM::BC_FREE) return;
            TV &f=force(out);
            if(bc_type==MPM::BC_SLIP) f(side/2)=0;
            else f=TV();
        });
}
//#####################################################################
// Function Grid_To_Particle
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Grid_To_Particle()
{
    TIMER_SCOPE_FUNC;
    MPM_PARTICLES<TV>& particles=example.particles;
    T dt=example.dt;

    ARRAY<TV> buoyancy_force,surface_tension_force;
    PARTICLE_BOUNDARY_CONDITIONS<TV> bc_helper(example.particles,example.grid,example.side_bc_type);
    auto [ng,np]=Transfer_G2P(example.opt);
    example.gather_scatter.template Gather(ng,np,0,0,
        [this,dt,&particles,&buoyancy_force,&surface_tension_force,&bc_helper](int p,const PARTICLE_GRID_ITERATOR<TV>& it_,int)
        {
            auto r=example.particles.Region(p);
            TV xp_new;
            Transfer_G2P<TV,T>(example.grid,example.opt,it_,
                example.velocity_save.u(r),example.velocity_friction_save.u(r),
                example.velocity.u(r),example.weights->Dp_inv,particles.X(p),
                particles.V(p),dt,xp_new,particles.V(p),
                example.opt.use_affine?&particles.C(p):nullptr,particles.F(p),
                example.opt.use_S?&particles.S(p):nullptr,nullptr);

            // I will need to update the position
            if(!example.fix_particles_location) particles.X(p)=xp_new;

            // I will need to update the velocity
            if(example.set_damp_particle_velocity) particles.V(p)/=(T)2; // Damping velocity?

            if(example.use_particle_collisions) Apply_Particle_Collisions(p);

            bc_helper.Enforce(p);
        });
}
//#####################################################################
// Function Reduce_Dt
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Reduce_Dt()
{
    TIMER_SCOPE_FUNC;
    T dt=example.dt;
    if(example.use_strong_cfl)
    {
        T new_dt=Grid_To_Particle_Limit_Dt();
        if(new_dt<dt)
        {
            LOG::printf("STRONG CFL %g %g (%g)\n",new_dt,dt,new_dt/dt);
            dt=new_dt;
        }
    }
//    dt=std::min(dt,Limit_Dt_Sound_Speed()); // This was done before forces.
    dt=std::max(dt,example.min_dt);
    if(dt>=example.dt*(T)0.99) return;

    T s=dt/example.dt;
    LOG::printf("dt scale: %g -> %g (%g)\n",example.dt,dt,s);
#pragma omp parallel for
    for(int i=0;i<example.valid_grid_indices.m;i++){
        auto j=example.valid_grid_indices(i);
        TV vs=example.velocity_save(j);
        TV& v=example.velocity(j);
        v=(v-vs)*s+vs;
        TV& vf=example.velocity_friction_save(j);
        vf=(vf-vs)*s+vs;}
    example.dt=dt;
}
// Lower s if necessary so that |t*(a+t*b)|<=bound for all 0<=t<=s
template<class T>
void Enforce_Limit_Max(T& s,T bound,T a,T b)
{
    if(a<0){a=-a;b=-b;}
    if(a*a+4*b*bound>=0){
        if(s*(a+s*b)>bound)
            s=(2*bound)/(a+sqrt(a*a+4*b*bound));}
    else{
        if(s*(a+s*b)<-bound)
            s=-(a+sqrt(a*a-4*b*bound))/(2*b);}
}
// Lower s if necessary so that t*(a+t*b)<=bound for all 0<=t<=s
template<class T,int d>
void Enforce_Limit_Max(T& s,T bound,const VECTOR<T,d>& a,const VECTOR<T,d>& b)
{
    for(int i=0;i<d;i++) Enforce_Limit_Max(s,bound,a(i),b(i));
}
// Lower s if necessary so that t*(a+t*b)<=bound for all 0<=t<=s
template<class T,int d>
void Enforce_Limit_Max(T& s,T bound,const MATRIX<T,d>& a,const MATRIX<T,d>& b)
{
    for(int i=0;i<d;i++)
        for(int j=0;j<d;j++)
            Enforce_Limit_Max(s,bound,a(i,j),b(i,j));
}
//#####################################################################
// Function Grid_To_Particle_Limit_Dt
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Grid_To_Particle_Limit_Dt() -> T
{
    TIMER_SCOPE_FUNC;
    T s=1;
    auto [ng,np]=Transfer_G2P(example.opt);
    example.gather_scatter.template Gather(ng,np,s,[&s](T t){s=std::min(s,t);},
        [this](int p,const PARTICLE_GRID_ITERATOR<TV>& it_,T& s)
        {
            auto r=example.particles.Region(p);
            s=Transfer_G2P_Limit(example.grid,example.opt,it_,
                example.velocity_save.u(r),example.velocity.u(r),example.dt,s,
                example.cfl,example.cfl_F);
        });
    if(example.verbose_cfl) LOG::printf("F CFL %P\n",example.dt*s);
    return example.dt*s;
}
//#####################################################################
// Function Limit_Dt_Sound_Speed
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Limit_Dt_Sound_Speed() -> T
{
    TIMER_SCOPE_FUNC;
    T dt=example.dt;
    T ratio_sound_single=0;
    if(example.use_sound_speed_cfl)
    {
        T max_speed=Compute_Max_Sound_Speed();
        T max_speed_async=Compute_Max_Sound_Speed_Async();

        LOG::printf("max sound speed: %.16P  (async %.16P)\n",max_speed,max_speed_async);
        if(example.use_async_sound_speed_cfl) max_speed=max_speed_async;
        LOG::printf("dx: %.16P\n",example.grid.dX.Min());
        LOG::printf("dx/soundspeed: %.16P\n",Robust_Divide(example.grid.dX.Min(),max_speed));
        T new_dt=Robust_Divide(example.grid.dX.Min(),max_speed)*example.cfl_sound;
        LOG::printf("SOUND CFL %g %g (%g)\n",example.dt,new_dt,Robust_Divide(new_dt,example.dt));
        dt=std::min(dt,new_dt);
        ratio_sound_single=Robust_Divide(example.grid.dX.Min(),max_speed);
        new_dt=Max_Dt_Single_Particle();
        ratio_sound_single=new_dt/ratio_sound_single;
        LOG::printf("Ratio of single_particle_dt/sound_speed_dt %P single_particle_dt %P\n",ratio_sound_single,new_dt);
    }
    if(example.use_single_particle_cfl && example.opt.dilation_only)
    {
        T new_dt=Max_Dt_Single_Particle_Pressure()*example.cfl_single_particle;
        if(new_dt<dt)
        {
            LOG::printf("SINGLE PARTICLE (J) %g %g (%g)\n",dt,new_dt,new_dt/dt);
            dt=new_dt;
        }
    }
    if(example.use_single_particle_cfl && !example.opt.dilation_only)
    {
        T new_dt=Max_Dt_Single_Particle()*example.cfl_single_particle;
        if(new_dt<dt)
        {
            LOG::printf("SINGLE PARTICLE (F) %g %g (%g)\n",dt,new_dt,new_dt/dt);
            dt=new_dt;
        }
    }
    return dt;
}
//#####################################################################
// Function Max_Dt_Single_Particle_Pressure
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Max_Dt_Single_Particle_Pressure() const -> T
{
    TIMER_SCOPE_FUNC;
    T dt=example.dt;
    T K=TV::m==2?6:(T)3.14;
    for(int f=0;f<example.forces.m;f++){
        if(const MPM_FINITE_ELEMENTS<TV>* force=dynamic_cast<MPM_FINITE_ELEMENTS<TV>*>(example.forces(f))){
#pragma omp parallel for reduction(min:dt)
            for(int k=0;k<example.simulated_particles.m;k++){
                int p=example.simulated_particles(k);
                T density=example.particles.mass(p)/example.particles.volume(p);
                T J=force->sigma(p).Determinant();
                T num=sqr(example.cfl_single_particle*example.grid.dX.Min())*density;
                T den=K*TV::m;
#if 0
                T dp_over_Jm1=force->constitutive_model.Robust_Divided_Pressure(J,p);
                den*=sqr(J)*dp_over_Jm1*J/(J+1);
#else
                T lambda=force->constitutive_model.Pressure_Bound(J,p);
                if(J>=1)
                {
                    num*=J+1;
                    den*=J*J*J*lambda;
                }
                else
                {
                    num*=2;
                    den*=sqr(2-J)*lambda;
                }
#endif
                if(num<sqr(dt)*den) dt=sqrt(num/den);}}}
    return dt;
}
//#####################################################################
// Function Max_Dt_Single_Particle
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Max_Dt_Single_Particle() const -> T
{
    T k=0;
    if(example.weights->order==2)
    {
        if(!example.opt.use_affine)
            k=1.0/6;
        if(example.opt.use_affine && !example.opt.use_cpic)
            k=1.0/4;
        if(example.opt.use_affine && example.opt.use_cpic)
            k=1.0/2;
    }
    else if(example.weights->order==3)
    {
        if(!example.opt.use_affine)
            k=0.3187942869;
        if(example.opt.use_affine && !example.opt.use_cpic)
            k=0.6109410720;
        if(example.opt.use_affine && example.opt.use_cpic)
            k=2.0/3;
    }
    T result_dt=(T)FLT_MAX;
    if(example.particles.mu.m==example.particles.number && example.particles.lambda.m==example.particles.number)
        for(int p=0;p<example.particles.number;p++)
        {
            T density=example.particles.mass(p)/example.particles.volume(p);
            T p_dt;
            if(example.particles.mu.m>p && example.particles.lambda.m>p)
                p_dt=example.grid.dX.Min()*sqrt(density*k/(example.particles.mu(p)+(T).5*TV::m*example.particles.lambda(p)));
            else
                p_dt=9999;
            if(p_dt<result_dt)
                result_dt=p_dt;
        }
    return result_dt;
}
//#####################################################################
// Function Compute_Max_Sound_Speed
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Compute_Max_Sound_Speed() const -> T
{
    TIMER_SCOPE_FUNC;
    T max_speed=0;
    for(int f=0;f<example.forces.m;f++){
        if(const MPM_FINITE_ELEMENTS<TV>* force=dynamic_cast<MPM_FINITE_ELEMENTS<TV>*>(example.forces(f))){
#pragma omp parallel for reduction(max:max_speed)
            for(int k=0;k<example.simulated_particles.m;k++){
                int p=example.simulated_particles(k);
                T density=example.particles.mass(p)/example.particles.volume(p);
                T speed=ISOTROPIC_CONSTITUTIVE_MODEL<T,TV::m>::Sound_Speed(
                    force->sigma(p),force->dPi_dF(p),density);
                max_speed=max(max_speed,speed);}}}
    return max_speed;
}
//#####################################################################
// Function Compute_Max_Sound_Speed_Async
//#####################################################################
template<class TV> auto MPM_DRIVER<TV>::
Compute_Max_Sound_Speed_Async() const -> T
{
    TIMER_SCOPE_FUNC;
    T max_speed=0;
    for(int f=0;f<example.forces.m;f++){
        if(const MPM_FINITE_ELEMENTS<TV>* force=dynamic_cast<MPM_FINITE_ELEMENTS<TV>*>(example.forces(f))){
            
            auto* cm=dynamic_cast<ST_VENANT_KIRCHHOFF_HENCKY_STRAIN<T,TV::m>*>(&force->constitutive_model);
            if(!example.use_async_sound_speed_cfl && !cm) continue;
#pragma omp parallel for reduction(max:max_speed)
            for(int k=0;k<example.simulated_particles.m;k++){
                int p=example.simulated_particles(k);
                T rho0=example.particles.mass(p)/example.particles.volume(p);
                T mu=cm->Mu(p);
                T lambda=cm->Lambda(p);
                T J=force->sigma(p).Determinant();
//                LOG::printf("BLAH %P %P\n",force->sigma(p),J);
                T K=2*mu/3+lambda;
                T dp_drho=K/rho0*(1-log(J));
                T speed=sqrt(4*mu*J/(3*rho0)+dp_drho);

                max_speed=max(max_speed,speed);}}}
    return max_speed;
}
//#####################################################################
// Function Update_Plasticity_And_Hardening
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Update_Plasticity_And_Hardening()
{
    TIMER_SCOPE_FUNC;
    for(int i=0;i<example.plasticity_models.m;i++)
        example.plasticity_models(i)->Update_Particles();
}
//#####################################################################
// Function Apply_Grid_Forces
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Apply_Grid_Forces(REGION_ARRAY<TV,IV>& F)
{
    TIMER_SCOPE_FUNC;
    for(int i=0;i<example.forces.m;i++)
        example.forces(i)->Add_Forces(F,example.time);
}
//#####################################################################
// Function Apply_Forces
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Apply_Forces()
{
    LOG::printf("max velocity: %.16P\n",Max_Particle_Speed());
    PHYSBAM_ASSERT(example.use_symplectic_euler);
    Apply_Forces_Explicit();
}
//#####################################################################
// Function Apply_Body_Collisions
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Apply_Body_Collisions()
{
    LOG::printf("Apply_Body_Collisions\n");
    T t=example.time+example.dt;
#pragma omp parallel for
    for(int k=0;k<example.valid_grid_indices.m;k++)
    {
        auto i=example.valid_grid_indices(k);
        TV V=example.velocity(i);
        TV X=example.location(i)+example.dt*V;
        V=Examine_Collisions(X,V,t);
        example.velocity(i)=V;
    }
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after body collisions",1);
}
//#####################################################################
// Function Apply_Particle_Collisions
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Apply_Particle_Collisions(int p)
{
    T t=example.time+example.dt;
    TV V=example.particles.V(p);
    TV X=example.particles.X(p)+example.dt*V;
    V=Examine_Collisions(X,V,t);
    example.particles.V(p)=V;
}
//#####################################################################
// Function Examine_Collisions
//#####################################################################
template<class TV> TV MPM_DRIVER<TV>::
Examine_Collisions(TV X,TV V,T t)
{
    for(int j=0;j<example.collision_objects.m;j++)
    {
        MPM_COLLISION_OBJECT<TV>* io=example.collision_objects(j);
        T phi=io->Phi(X,t);
        if(phi>0) continue;
        TV Vc=io->Velocity(X,t);
        if(io->type==MPM::BC_NOSLIP){V=Vc;break;}
        TV N=io->Normal(X,t),V_rel=V-Vc;
        T Vn_rel=V_rel.Dot(N);
        if(io->type==MPM::BC_SEP && Vn_rel>0) continue;
        TV Vt_rel=V_rel-Vn_rel*N,Vt_rel_unit=Vt_rel;
        T t_mag=Vt_rel_unit.Normalize();
        T coefficient_of_friction=io->friction;
        T friction_magnitude=-coefficient_of_friction*Vn_rel;
        if(t_mag<=friction_magnitude) V=Vc;
        else V=Vt_rel-friction_magnitude*Vt_rel_unit+Vc;
    }
    return V;
}

//#####################################################################
// Function Apply_Forces_Explicit
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Apply_Forces_Explicit()
{
    TIMER_SCOPE_FUNC;
    example.Capture_Stress();
    example.force_helper.B.Resize(example.particles.number);
    example.force_helper.B.Fill(MATRIX<T,TV::m>());
    example.Precompute_Forces(example.time,example.dt,0);

    T cfl_dt=max(Limit_Dt_Sound_Speed(),example.min_dt);
    if(cfl_dt<example.dt*(T)0.99)
        example.dt=cfl_dt;
    
    REGION_ARRAY<TV,IV> &F=example.velocity_friction_save;
    F.raw.Fill(TV());

    TIMER_SCOPE("Apply_Forces B");
    //Add Particle Forces
    Apply_Grid_Forces(F);

    F.Exchange(example.velocity);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("raw forces",1);
    F.Exchange(example.velocity);
    for(auto&u:F.u) Fix_Periodic_Accum(u,PBH::ADD());
    
    F.Exchange(example.velocity);
    PHYSBAM_DEBUG_WRITE_SUBSTEP("raw forces after periodic",1);
    F.Exchange(example.velocity);
    if(example.reflection_bc && !example.use_full_reflection)
        Reflect_Boundary_Force(F);

    TIMER_SCOPE("Apply_Forces C");
#pragma omp parallel for
    for(int i=0;i<example.valid_grid_indices.m;i++){
        auto p=example.valid_grid_indices(i);
        example.velocity(p)+=example.dt/example.mass(p)*F(p);}
    PHYSBAM_DEBUG_WRITE_SUBSTEP("after apply forces",1);

    if(example.collision_objects.m)
        Apply_Body_Collisions();
    if(example.reflection_bc && example.use_full_reflection)
        Reflect_Boundary_Friction(example.mass,example.velocity);
    if(example.reflection_bc && !example.use_full_reflection)
        Reflect_Boundary_Velocity(example.velocity);

    for(auto&u:example.velocity.u) Fix_Periodic(u);
    example.velocity_friction_save=example.velocity;
}
//#####################################################################
// Function Compute_Dt
//#####################################################################
template<class TV> typename TV::SCALAR MPM_DRIVER<TV>::
Compute_Dt() const
{
    T critical_speed=example.cfl*example.grid.dX.Min()/example.max_dt;
    T v=Grid_V_Upper_Bound();
    LOG::printf("v %P\n",v);
    return (v>critical_speed)?(example.cfl*example.grid.dX.Min()/v):example.max_dt;
}
//#####################################################################
// Function Max_Particle_Speed
//#####################################################################
template<class TV> typename TV::SCALAR MPM_DRIVER<TV>::
Max_Particle_Speed() const
{
    T v2=0;
    T max_F=0;
#pragma omp parallel for reduction(max:v2)
    for(int k=0;k<example.simulated_particles.m;k++){
        int p=example.simulated_particles(k);
        v2=max(v2,example.particles.V(p).Magnitude_Squared());
        auto F=example.particles.F(p);
        max_F=max(max_F,(F.Transpose_Times(F)-1).Frobenius_Norm_Squared());}
    LOG::printf("maximum F magnitude: %P\n",sqrt(max_F));
    return sqrt(v2);
}
//#####################################################################
// Function Grid_V_Upper_Bound
//#####################################################################
template<class TV> typename TV::SCALAR MPM_DRIVER<TV>::
Grid_V_Upper_Bound() const
{
    if(!example.opt.use_affine || !example.use_affine_cfl) return Max_Particle_Speed();
    T result=0,max_C=0;
    T xi=(T)(example.weights->order+1)/2*sqrt((T)TV::m)*example.grid.dX.Max();
#pragma omp parallel for reduction(max:result,max_C)
    for(int k=0;k<example.simulated_particles.m;k++){
        int p=example.simulated_particles(k);
        T v=example.particles.V(p).Magnitude();
        if(example.particles.store_C)
        {
            T C=example.particles.C(p).Frobenius_Norm();
            if(example.C_bound)v+=C*xi;
            max_C=max(max_C,C);
        }
        result=max(result,v);}
    LOG::printf("maximum C magnitude %P\n",max_C);
    return result;
}
//#####################################################################
// Function Update_Simulated_Particles
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Update_Simulated_Particles()
{
    TIMER_SCOPE_FUNC;
    example.simulated_particles.Remove_All();
    for(int p=0;p<example.particles.number;p++)
        if(example.particles.valid(p))
            example.simulated_particles.Append(p);
    // example.simulated_particles=IDENTITY_ARRAY<>(example.particles.X.m);
    example.particle_is_simulated.Remove_All();
    example.particle_is_simulated.Resize(example.particles.X.m);
    example.particle_is_simulated.Subset(example.simulated_particles).Fill(true);
}
//#####################################################################
// Function Print_Grid_Stats
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Print_Grid_Stats(const char* str,T dt,const REGION_ARRAY<TV,IV>& u,const REGION_ARRAY<TV,IV>* u0)
{
    TIMER_SCOPE_FUNC;
    if(!example.print_stats) return;
    typename TV::SPIN am=example.Total_Grid_Angular_Momentum(dt,u,u0);
    TV lm=example.Total_Grid_Linear_Momentum(u);
    T ke=example.Total_Grid_Kinetic_Energy(u);
    LOG::cout<<str<<" linear  "<<"time " <<example.time<<" value "<<lm<<"  diff "<<(lm-example.last_linear_momentum)<<std::endl;
    LOG::cout<<str<<" angular "<<"time " <<example.time<<" value "<<am<<"  diff "<<(am-example.last_angular_momentum)<<std::endl;
    LOG::cout<<str<<" ke "<<"time " <<example.time<<" value "<<ke<<"  diff "<<(ke-example.last_grid_ke)<<std::endl;
    example.last_linear_momentum=lm;
    example.last_angular_momentum=am;
    example.last_grid_ke=ke;
}
//#####################################################################
// Function Print_Grid_Stats
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Print_Particle_Stats(const char* str,T dt)
{
    TIMER_SCOPE_FUNC;
    if(!example.print_stats) return;
    typename TV::SPIN am=example.Total_Particle_Angular_Momentum();
    TV lm=example.Total_Particle_Linear_Momentum();
    // T ke=example.Total_Particle_Kinetic_Energy();
    LOG::cout<<str<<" linear  "<<"time " <<example.time<<" value "<<lm<<"  diff "<<(lm-example.last_linear_momentum)<<std::endl;
    LOG::cout<<str<<" angular "<<"time " <<example.time<<" value "<<am<<"  diff "<<(am-example.last_angular_momentum)<<std::endl;
    // LOG::cout<<str<<" ke "<<ke<<"  diff "<<(ke-example.last_grid_ke)<<std::endl;
    example.last_linear_momentum=lm;
    example.last_angular_momentum=am;
    // example.last_grid_ke=ke;
}
//#####################################################################
// Function Print_Energy_Stats
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Print_Energy_Stats(const char* str,const REGION_ARRAY<TV,IV>& u)
{
    TIMER_SCOPE_FUNC;
    if(!example.print_stats) return;
    example.Capture_Stress();
    example.Precompute_Forces(example.time,example.dt,false);
    T ke=example.Total_Grid_Kinetic_Energy(u);
    T ke2=example.Total_Particle_Kinetic_Energy();
    T pe=example.Potential_Energy(example.time);
    T te=ke+pe;
    LOG::cout<<str<<" kinetic  "<<"time " <<example.time<<" value "<<ke<<std::endl;
    LOG::cout<<str<<" potential "<<"time " <<example.time<<" value "<<pe<<std::endl;
    LOG::cout<<str<<" total energy "<<"time " <<example.time<<" value "<<te<<" diff "<<(te-example.last_te)<<std::endl;
    LOG::cout<<str<<" particle total energy "<<"time " <<example.time<<" value "<<(ke2+pe)<<std::endl;
    example.last_te=te;
}
template<class T,int d> void flip(VECTOR<T,d>& u,int a){u(a)=-u(a);}
template<class T> void flip(T& u,int a){}
//#####################################################################
// Function Step
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Step(std::function<void()> func,const char* name,bool dump_substep,bool do_step)
{
    TIMER_SCOPE_FUNC;
    auto& p=example.time_step_callbacks.Get_Or_Insert(name);
    p.first=true; // Flag the callback name as recognized, for sanity checking later
    if(!do_step) return;
    for(int i=0;i<p.second(1).m;i++) p.second(1)(i)();
    func();
    // Note: p may be invalidated by func(), so we must re-access it here.
    const auto& q=example.time_step_callbacks.Get_Or_Insert(name);
    for(int i=0;i<q.second(0).m;i++) q.second(0)(i)();
    if(dump_substep) PHYSBAM_DEBUG_WRITE_SUBSTEP(name,1);
}
//#####################################################################
// Function Compute_Regions
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Compute_Regions()
{
    ARRAY<RANGE<IV> > particle_ranges;
    ARRAY<RANGE<IV>,REGION_ID> region_ranges;
    example.weights->Particle_Ranges(particle_ranges);
    for(int p=0;p<example.particles.X.m;p++)
        if(example.particles.valid(p))
        {
            REGION_ID i=example.particles.Region(p);
            if(i>=region_ranges.m)
                region_ranges.Resize(i+1,init_new,RANGE<IV>::Empty_Box());
            region_ranges(i).Enlarge_To_Include_Box(particle_ranges(p));
        }
    example.mass.Resize(region_ranges);
    example.volume.Resize(example.mass);
    example.location.Resize(example.mass);
    example.velocity.Resize(example.mass);
    example.velocity_save.Resize(example.mass);
    example.velocity_friction_save.Resize(example.mass);
    if(example.mass.u.m>REGION_ID(1))
        example.normal.Resize(example.mass);
    LOG::printf("regions %P\n",example.mass.u.m);

    for(REGION_ID r(0);r<example.location.u.m;r++)
    {
        auto& a=example.location.u(r);
#pragma omp parallel
        for(NODE_ITERATOR_THREADED<TV> it(example.grid,a.domain);it.Valid();it.Next())
            a(it.index)=example.grid.Node_X(it.index);
    }
}
//#####################################################################
// Function Collide_Regions
//#####################################################################
template<class TV> void MPM_DRIVER<TV>::
Collide_Regions()
{
    if(example.mass.u.m<=REGION_ID(1)) return;
    ARRAY<VECTOR<T,3>,REGION_ID> colors={{1,1,1},{1,0,0}};

    if(example.write_substeps_level>=1)
    {
        for(REGION_ID r(0);r<example.mass.u.m;r++)
            for(RANGE_ITERATOR<TV::m> it(example.mass.u(r).domain);it.Valid();it.Next())
            {
                Add_Point(example.grid.Node_X(it.index),colors(r));
                Add_Arrow(example.grid.Node_X(it.index),example.normal.u(r)(it.index),colors(r));
            }
        PHYSBAM_DEBUG_WRITE_SUBSTEP("normals",1);
    }
    
    if(example.write_substeps_level>=1)
    {
        for(REGION_ID r(0);r<example.mass.u.m;r++)
            for(RANGE_ITERATOR<TV::m> it(example.mass.u(r).domain);it.Valid();it.Next())
            {
                Add_Point(example.grid.Node_X(it.index),colors(r));
                Add_Arrow(example.grid.Node_X(it.index),example.velocity.u(r)(it.index),colors(r));
            }
        PHYSBAM_DEBUG_WRITE_SUBSTEP("before collide",1);
    }    
    
    for(REGION_ID r(0);r<example.mass.u.m;r++)
        for(REGION_ID s=r+1;s<example.mass.u.m;s++)
        {
            //LOG::printf("COLLIDE TRY %P %P\n",r,s);
            auto cp=example.pairwise_collision_parameters.Get_Default(
                {r,s},example.default_collision_parameters);
            if(cp.first==MPM::BC_FREE) continue;
            
            auto range=example.mass.u(r).domain.Intersect(example.mass.u(s).domain);
            if(range.Empty()) continue;
            for(RANGE_ITERATOR<TV::m> it(range);it.Valid();it.Next())
            {
                REGION_INDEX<TV::m> r0{r,it.index};
                REGION_INDEX<TV::m> r1{s,it.index};
                T m0=example.mass(r0),m1=example.mass(r1);
                if(!m0 || !m1) continue;
                TV n0=example.normal(r0),n1=example.normal(r1);
                TV& v0=example.velocity(r0),&v1=example.velocity(r1);
                TV n=(m0*n0-m1*n1).Normalized();
                TV dv=v0-v1,j;
                T dv_n=dv.Dot(n),a0=m1/(m0+m1),a1=1-a0;
                //LOG::printf("COLLIDE PAIR %P %P   %P %P    %P %P %P    %P %P\n",m0,m1,v0,v1,n0,n1,n,dv,dv_n);
                if(cp.first==MPM::BC_NOSLIP) j=dv;
                else if(cp.first==MPM::BC_SLIP) j=dv_n*n;
                else if(dv_n<=0) continue;
                else
                {
                    TV dv_t=dv-dv_n*n;
                    if(dv_t.Normalize()<cp.second*dv_n) j=dv;
                    else j=dv_n*n+cp.second*dv_n*dv_t;
                }
                v0-=a0*j;
                v1+=a1*j;
                example.velocity_friction_save(r0)=v0;
                example.velocity_friction_save(r1)=v1;
                //LOG::printf("AFTER %P\n",v0-v1);
            }
        }
    for(auto&u:example.velocity.u) Fix_Periodic(u);
    for(auto&u:example.velocity_friction_save.u) Fix_Periodic(u);

    if(example.write_substeps_level>=1)
    {
        for(REGION_ID r(0);r<example.mass.u.m;r++)
            for(RANGE_ITERATOR<TV::m> it(example.mass.u(r).domain);it.Valid();it.Next())
            {
                Add_Point(example.grid.Node_X(it.index),colors(r));
                Add_Arrow(example.grid.Node_X(it.index),example.velocity.u(r)(it.index),colors(r));
            }
        PHYSBAM_DEBUG_WRITE_SUBSTEP("after collide",1);
    }
}
//#####################################################################
// Function Fix_Periodic
//#####################################################################
template<class TV> template<class T2> void MPM_DRIVER<TV>::
Fix_Periodic(ARRAY_VIEW<T2,IV> u) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Nodes_Periodic(example.grid,u,example.is_periodic);
}
//#####################################################################
// Function Fix_Periodic_Accum
//#####################################################################
template<class TV> template<class T2,class OP> void MPM_DRIVER<TV>::
Fix_Periodic_Accum(ARRAY_VIEW<T2,IV> u,const OP& op) const
{
    if(!example.side_bc_type.Contains(MPM::BC_PERIODIC)) return;
    Fill_Ghost_Nodes_Periodic_Accum(example.grid,u,example.is_periodic,op);
}
//#####################################################################
namespace PhysBAM{
template class MPM_DRIVER<VECTOR<float,2> >;
template class MPM_DRIVER<VECTOR<float,3> >;
template class MPM_DRIVER<VECTOR<double,2> >;
template class MPM_DRIVER<VECTOR<double,3> >;
}
