//#####################################################################
// Copyright 2015, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Hybrid_Methods/Examples_And_Drivers/MPM_EXAMPLE.h>
#include <Tools/Read_Write/OCTAVE_OUTPUT.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
#include <Solids/Solids/SOLID_BODY_COLLECTION.h>
#include <Hybrid_Methods/Collisions/MPM_COLLISION_OBJECT.h>
#include <Hybrid_Methods/Examples_And_Drivers/MPM_PARTICLES.h>
#include <Hybrid_Methods/Forces/MPM_FORCE_HELPER.h>
#include <Hybrid_Methods/Forces/MPM_PLASTICITY_MODEL.h>
#include <Hybrid_Methods/Iterators/GATHER_SCATTER.h>
#include <Hybrid_Methods/Iterators/PARTICLE_GRID_WEIGHTS_SPLINE.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class TV> MPM_EXAMPLE<TV>::
MPM_EXAMPLE(const STREAM_TYPE stream_type)
    :stream_type(stream_type),particles(*new MPM_PARTICLES<TV>),
    solid_body_collection(*new SOLID_BODY_COLLECTION<TV>(&particles)),
    debug_particles(DEBUG_PARTICLES<TV>::Get()),
    gather_scatter(*new GATHER_SCATTER<TV>(grid,simulated_particles)),
    force_helper(*new MPM_FORCE_HELPER<TV>(particles,opt))
{
    particles.Store_Region(true);
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> MPM_EXAMPLE<TV>::
~MPM_EXAMPLE()
{
    delete &solid_body_collection;
    delete &particles;
    delete weights;
    delete &gather_scatter;
    delete &force_helper;
    plasticity_models.Delete_Pointers_And_Clean_Memory();
    collision_objects.Delete_Pointers_And_Clean_Memory();
    forces.Delete_Pointers_And_Clean_Memory();
}
//#####################################################################
// Function Write_Output_Files
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Write_Output_Files()
{
    for(int i=0;i<write_output_files.m;i++) write_output_files(i)();
#pragma omp parallel
#pragma omp single
    {
#pragma omp task
        Write_To_File(stream_type,viewer_dir.Common_File("grid"),grid);
#pragma omp task
        if(!system(LOG::sprintf("rm -f %s ;  ln -s ./deformable_object_particles.gz %s",
                    viewer_dir.Frame_File("mpm_particles.gz"),
                    viewer_dir.Frame_File("mpm_particles.gz")).c_str())){}
#pragma omp task
        Write_To_File(stream_type,viewer_dir.Frame_File("restart_data"),time);
#pragma omp task
        {
            solid_body_collection.Write(stream_type,viewer_dir,write_structures_every_frame);
        }

        if(!only_write_particles){
            if(velocity.u.m)
#pragma omp task
                Write_To_File(stream_type,viewer_dir.Frame_File("node_velocities"),velocity.u(REGION_ID()));
            if(velocity_save.u.m && write_save_velocity)
#pragma omp task
                Write_To_File(stream_type,viewer_dir.Frame_File("save_velocities"),velocity_save.u(REGION_ID()));
#pragma omp task
            {
                GRID<TV> ghost_grid=grid.Get_Ghost_Grid(ghost);
                for(int i=0;i<collision_objects.m;i++)
                    if(IMPLICIT_OBJECT<TV>* io=collision_objects(i)->Get_Implicit_Object(time))
                        Dump_Levelset(ghost_grid,at_nodes,*io,VECTOR<T,3>(0.7,0.3,0.3));
                debug_particles.Write_Debug_Particles(stream_type,viewer_dir);
            }
        }
    }
}
//#####################################################################
// Function Read_Output_Files
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Read_Output_Files()
{
    Read_From_File(viewer_dir.Frame_File("mpm_particles"),particles);
    Read_From_File(viewer_dir.Frame_File("restart_data"),time);
    for(int i=0;i<read_output_files.m;i++) read_output_files(i)();
}
//#####################################################################
// Function Capture_Stress
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Capture_Stress()
{
    force_helper.Fn=particles.F;
    if(particles.store_S) force_helper.Sn=particles.S;
}
//#####################################################################
// Function Precompute_Forces
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Precompute_Forces(const T time,const T dt,const bool update_hessian)
{
    for(int i=0;i<forces.m;i++)
        forces(i)->Precompute(time,dt,true,update_hessian);
}
//#####################################################################
// Function Potential_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_EXAMPLE<TV>::
Potential_Energy(const T time) const
{
    typename TV::SCALAR pe=0;
    for(int i=0;i<forces.m;i++)
        pe+=forces(i)->Potential_Energy(time);
    return pe;
}
//#####################################################################
// Function Add_Forces
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Add_Forces(REGION_ARRAY<TV,IV>& F,const T time) const
{
    for(int i=0;i<forces.m;i++)
        forces(i)->Add_Forces(F,time);
}
//#####################################################################
// Function Add_Hessian_Times
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Add_Hessian_Times(REGION_ARRAY<TV,IV>& F,const REGION_ARRAY<TV,IV>& V,const T time) const
{
    for(int i=0;i<forces.m;i++)
        forces(i)->Add_Hessian_Times(F,V,time);
}
//#####################################################################
// Function Add_Force
//#####################################################################
template<class TV> int MPM_EXAMPLE<TV>::
Add_Force(PARTICLE_GRID_FORCES<TV>& force)
{
    return forces.Append(&force);
}
//#####################################################################
// Function Set_Weights
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Set_Weights(int order)
{
    delete weights;
    if(order==2) gather_scatter.weights=weights=new PARTICLE_GRID_WEIGHTS_SPLINE<TV,2>(grid,threads);
    else if(order==3) gather_scatter.weights=weights=new PARTICLE_GRID_WEIGHTS_SPLINE<TV,3>(grid,threads);
    else PHYSBAM_FATAL_ERROR("Unrecognized interpolation order");
    weights->order=order;
}
//#####################################################################
// Function Total_Particle_Linear_Momentum
//#####################################################################
template<class TV> TV MPM_EXAMPLE<TV>::
Total_Particle_Linear_Momentum() const
{
    TV result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*simulated_particles.m/threads;
        int b=(t+1)*simulated_particles.m/threads;
        TV result_local;
        for(int k=a;k<b;k++){
            int p=simulated_particles(k);
            result_local+=particles.mass(p)*particles.V(p);}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Linear_Momentum
//#####################################################################
template<class TV> TV MPM_EXAMPLE<TV>::
Total_Grid_Linear_Momentum(const REGION_ARRAY<TV,IV>& u) const
{
    TV result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*valid_grid_indices.m/threads;
        int b=(t+1)*valid_grid_indices.m/threads;
        TV result_local;
        for(int k=a;k<b;k++){
            auto j=valid_grid_indices(k);
            result_local+=mass(j)*u(j);}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Angular_Momentum
//#####################################################################
template<class TV> typename TV::SPIN MPM_EXAMPLE<TV>::
Total_Particle_Angular_Momentum() const
{
    typename TV::SPIN result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*simulated_particles.m/threads;
        int b=(t+1)*simulated_particles.m/threads;
        typename TV::SPIN result_local;
        for(int k=a;k<b;k++){
            int p=simulated_particles(k);
            result_local+=particles.mass(p)*particles.X(p).Cross(particles.V(p));
            if(particles.store_C) result_local-=particles.mass(p)*weights->Dp*particles.C(p).Contract_Permutation_Tensor();}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Angular_Momentum
//#####################################################################
template<class TV> typename TV::SPIN MPM_EXAMPLE<TV>::
Total_Grid_Angular_Momentum(T dt,const REGION_ARRAY<TV,IV>& u,const REGION_ARRAY<TV,IV>* u0) const
{
    typename TV::SPIN result;
#pragma omp parallel for
    for(int t=0;t<threads;t++){
        int a=t*valid_grid_indices.m/threads;
        int b=(t+1)*valid_grid_indices.m/threads;
        typename TV::SPIN result_local;
        for(int k=a;k<b;k++){
            auto i=valid_grid_indices(k);
            TV X=location(i);
            if(opt.use_midpoint && u0) X+=dt/2*(*u0)(i);
            result_local+=mass(i)*TV::Cross_Product(X,u(i));}
#pragma omp critical
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Total_Grid_Kinetic_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_EXAMPLE<TV>::
Total_Grid_Kinetic_Energy(const REGION_ARRAY<TV,IV>& u) const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int i=0;i<valid_grid_indices.m;i++){
        auto j=valid_grid_indices(i);
        result+=(T).5*mass(j)*u(j).Magnitude_Squared();}
    return result;
}
//#####################################################################
// Function Total_Particle_Kinetic_Energy
//#####################################################################
template<class TV> typename TV::SCALAR MPM_EXAMPLE<TV>::
Total_Particle_Kinetic_Energy() const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int k=0;k<simulated_particles.m;k++){
        int p=simulated_particles(k);
        T result_local=particles.mass(p)/2*particles.V(p).Magnitude_Squared();
        if(particles.store_C){
            T a=particles.C(p).Frobenius_Norm_Squared();
            result_local+=particles.mass(p)/2*weights->Dp*a;}
        result+=result_local;}
    return result;
}
//#####################################################################
// Function Average_Particle_Mass
//#####################################################################
template<class TV> typename TV::SCALAR MPM_EXAMPLE<TV>::
Average_Particle_Mass() const
{
    T result=0;
#pragma omp parallel for reduction(+:result)
    for(int k=0;k<simulated_particles.m;k++)
        result+=particles.mass(simulated_particles(k));
    return result/(T)particles.number;
}
//#####################################################################
// Function Add_Collision_Object
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Add_Collision_Object(IMPLICIT_OBJECT<TV>* io,MPM::BC_TYPE type,T friction,std::function<FRAME<TV>(T)> func_frame,std::function<TWIST<TV>(T)> func_twist)
{
    auto* co=new MPM_COLLISION_OBJECT<TV>(io,type,friction);
    co->func_frame=func_frame;
    co->func_twist=func_twist;
    collision_objects.Append(co);
}
//#####################################################################
// Function Add_Callbacks
//#####################################################################
template<class TV> void MPM_EXAMPLE<TV>::
Add_Callbacks(bool is_begin,const char* func_name,std::function<void()> func)
{
    time_step_callbacks.Get_Or_Insert(func_name).second(is_begin).Append(func);
}
namespace PhysBAM{
template class MPM_EXAMPLE<VECTOR<float,2> >;
template class MPM_EXAMPLE<VECTOR<float,3> >;
template class MPM_EXAMPLE<VECTOR<double,2> >;
template class MPM_EXAMPLE<VECTOR<double,3> >;
}
