//#####################################################################
// Copyright 2004-2009, Eran Guendelman, Andrew Selle.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <OpenGL/OpenGL/OPENGL_MAC_VELOCITY_FIELD_2D.h>
#include <Core/Arrays/ARRAY.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Math_Tools/RANGE.h>
#include <Grid_PDE/Interpolation/AVERAGING_UNIFORM.h>
#include <Grid_Tools/Grids/FACE_RANGE_ITERATOR.h>
#include <Grid_Tools/Grids/GRID.h>
#include <Grid_PDE/Interpolation/AVERAGING_UNIFORM.h>
#include <OpenGL/OpenGL/OPENGL_GRID_2D.h>
#include <OpenGL/OpenGL_Components/OPENGL_COMPONENT_PARTICLES_2D.h>
using namespace PhysBAM;
using namespace std;
//#####################################################################
// Constructor
//#####################################################################
template<class T> OPENGL_MAC_VELOCITY_FIELD_2D<T>::
OPENGL_MAC_VELOCITY_FIELD_2D(const GRID<TV> &grid_input)
    :OPENGL_VECTOR_FIELD_2D<ARRAY<TV> >(vector_field,vector_locations),
    grid(grid_input),active_cells(0),active_faces(0)
{
    Set_Velocity_Mode(CELL_CENTERED);
}
//#####################################################################
// Destructor
//#####################################################################
template<class T> OPENGL_MAC_VELOCITY_FIELD_2D<T>::
~OPENGL_MAC_VELOCITY_FIELD_2D()
{
    delete active_cells;
    delete active_faces;
}
//#####################################################################
// Function Set_Velocity_Mode
//#####################################################################
template<class T> void OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Set_Velocity_Mode(VELOCITY_MODE velocity_mode_input)
{
    velocity_mode = velocity_mode_input;
    Update();
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Bounding_Box() const
{
    return World_Space_Box(grid.domain);
}
//#####################################################################
// Function Update
//#####################################################################
template<class T> void OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Update()
{
    vector_field.Remove_All();
    vector_locations.Remove_All();

    if(velocity_mode==FACE_CENTERED)
        for(FACE_RANGE_ITERATOR<TV::m> it(face_velocities.domain);it.Valid();it.Next()){
            vector_locations.Append(grid.Face_X(it.face));
            vector_field.Append(face_velocities(it.face)*TV::Axis_Vector(it.face.axis));}
    else{
        AVERAGING_UNIFORM<TV> au;
        for(RANGE_ITERATOR<TV::m> it(face_velocities.domain);it.Valid();it.Next()){
            vector_locations.Append(grid.Cell_X(it.index));
            vector_field.Append(au.Face_To_Cell_Vector(grid,it.index,face_velocities));}}
}
//#####################################################################
// Function Print_Cell_Selection_Info
//#####################################################################
template<class T> void OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Print_Cell_Selection_Info(std::ostream& stream,const IV& cell) const
{
    if(!face_velocities.domain.Lazy_Inside_Half_Open(cell)) return;
    VECTOR<FACE_INDEX<TV::m>,4> faces;
    grid.Neighboring_Faces(faces,cell);
    T u_left=face_velocities(faces(0));
    T u_right=face_velocities(faces(1));
    T v_bottom=face_velocities(faces(2));
    T v_top=face_velocities(faces(3));
    TV center_velocity(0.5*(u_left+u_right),0.5*(v_bottom+v_top));
    stream<<"    u left = "<<u_left<<",right = "<<u_right<<" avg="<<center_velocity.x<<std::endl;
    stream<<"    v bottom = "<<v_bottom<<",top = "<<v_top<<" avg="<<center_velocity.y<<std::endl;
    T ux=(u_right-u_left)*grid.one_over_dX.x,vy=(v_top-v_bottom)*grid.one_over_dX.y;
    stream<<"    divergence = "<<ux+vy<<" (ux="<<ux<<", vy="<<vy<<")"<<std::endl;
}
//#####################################################################
// Function Print_Node_Selection_Info
//#####################################################################
template<class T> void OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Print_Node_Selection_Info(std::ostream& stream,const IV& node) const
{
}
//#####################################################################
// Convenience functions
//#####################################################################
template<class T> void OPENGL_MAC_VELOCITY_FIELD_2D<T>::
Toggle_Velocity_Mode()
{
    VELOCITY_MODE new_velocity_mode = (VELOCITY_MODE)(((int)velocity_mode+1)%2);
    Set_Velocity_Mode(new_velocity_mode);
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_MAC_VELOCITY_FIELD_2D<float>;
template class OPENGL_MAC_VELOCITY_FIELD_2D<double>;
}
