//#####################################################################
// Copyright 2011-2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <OpenGL/OpenGL_Components/OPENGL_COMPONENT_DEBUG_PARTICLES_2D.h>
#include <Core/Arrays_Nd/ARRAYS_ND.h>
#include <Core/Log/LOG.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Geometry/Basic_Geometry/TRIANGLE_2D.h>
#include <Geometry/Geometry_Particles/DEBUG_PARTICLES.h>
#include <Geometry/Geometry_Particles/GEOMETRY_PARTICLES.h>
#include <OpenGL/OpenGL/OPENGL_INDEXED_COLOR_MAP.h>
#include <OpenGL/OpenGL/OPENGL_MATERIAL.h>
#include <OpenGL/OpenGL/OPENGL_SELECTION.h>
#include <OpenGL/OpenGL/OPENGL_SHAPES.h>
#include <OpenGL/OpenGL/OPENGL_WORLD.h>
#include <sstream>
#include <stdexcept>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class T> OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
OPENGL_COMPONENT_DEBUG_PARTICLES_2D(const VIEWER_DIR& viewer_dir,const std::string& filename)
    :OPENGL_COMPONENT<T>(viewer_dir,"Particles 2D"),
    debug_particles(*new DEBUG_PARTICLES<TV>),filename(filename)
{
    viewer_callbacks.Set("increase_vector_size",{[this](){scale_factor*=(T)1.1;},"Increase vector size"});
    viewer_callbacks.Set("decrease_vector_size",{[this](){scale_factor/=(T)1.1;},"Decrease vector size"});
    viewer_callbacks.Set("reset_vector_size",{[this](){scale_factor=1;},"Reset vector size"});
    viewer_callbacks.Set("command_prompt",{[this](){Command_Prompt();},"Command prompt"});
}
//#####################################################################
// Destructor
//#####################################################################
template<class T> OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
~OPENGL_COMPONENT_DEBUG_PARTICLES_2D()
{
    delete &debug_particles;
}
//#####################################################################
// Function Set_Frame
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Set_Frame()
{
    Reinitialize();
}
//#####################################################################
// Function Set_Draw
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Set_Draw(bool draw_input)
{
    OPENGL_COMPONENT<T>::Set_Draw(draw_input);
    Reinitialize();
}
//#####################################################################
// Function Display
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Display() const
{
    if(!valid || !draw) return;
    if(slice && slice->Is_Slice_Mode()){
        glPushAttrib(GL_ENABLE_BIT);
        slice->Enable_Clip_Planes();}

    glPushAttrib(GL_LIGHTING_BIT | GL_CURRENT_BIT);
    glDisable(GL_LIGHTING);
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE,1);
    glDisable(GL_CULL_FACE);
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    Send_Transform_To_GL_Pipeline();

    glPushAttrib(GL_POINT_BIT);
    glPointSize(5);
    glPushName(0);
    for(int i=0;i<debug_particles.points.X(0).m;i++)
    {
        glPushName(i);
        auto& c=debug_particles.points;
        OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
        OpenGL_Begin(GL_POINTS);
        OpenGL_Vertex(c.X(0)(i));
        OpenGL_End();
        glPopName();
    };
    glPopAttrib();
    glLoadName(1);
    for(int i=0;i<debug_particles.segments.X(0).m;i++)
    {
        glPushName(i);
        auto& c=debug_particles.segments;
        OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
        OpenGL_Begin(GL_LINE_STRIP);
        OpenGL_Vertex(c.X(0)(i));
        OpenGL_Vertex(c.X(1)(i));
        OpenGL_End();
        glPopName();
    };
    glLoadName(2);
    for(int f=0;f<2;f++)
    {
        glPushName(f);
        for(int i=0;i<debug_particles.tris(f).X(0).m;i++)
        {
            glPushName(i);
            auto& c=debug_particles.tris(f);
            OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
            OpenGL_Begin(f?GL_POLYGON:GL_LINE_LOOP);
            OpenGL_Vertex(c.X(0)(i));
            OpenGL_Vertex(c.X(1)(i));
            OpenGL_Vertex(c.X(2)(i));
            OpenGL_End();
            glPopName();
        }
        glPopName();
    }
    glLoadName(3);
    for(int f=0;f<2;f++)
    {
        glPushName(f);
        for(int s=0;s<2;s++)
        {
            T sc=s?scale_factor:1;
            glPushName(s);
            for(int i=0;i<debug_particles.balls(f)(s).X(0).m;i++)
            {
                glPushName(i);
                auto& c=debug_particles.balls(f)(s);
                OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
                OPENGL_SHAPES::Draw_Circle(c.X(0)(i),c.t(0)(i)*sc,20,f);
                glPopName();
            }
            glPopName();
        }
        glPopName();
    }
    glLoadName(4);
    for(int a=0;a<2;a++)
    {
        glPushName(a);
        for(int s=0;s<2;s++)
        {
            T sc=s?scale_factor:1;
            glPushName(s);
            for(int i=0;i<debug_particles.arrows(a)(s).X(0).m;i++)
            {
                glPushName(i);
                auto& c=debug_particles.arrows(a)(s);
                OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
                OpenGL_Begin(GL_LINES);
                if(a) OPENGL_SHAPES::Draw_Arrow(c.X(0)(i),c.X(0)(i)+c.X(1)(i)*sc);
                else
                {
                    OpenGL_Vertex(c.X(0)(i));
                    OpenGL_Vertex(c.X(0)(i)+c.X(1)(i)*sc);
                }
                OpenGL_End();
                glPopName();
            }
            glPopName();
        }
        glPopName();
    }
    glLoadName(5);
    for(int f=0;f<2;f++)
    {
        glPushName(f);
        for(int s=0;s<2;s++)
        {
            T sc=s?scale_factor:1;
            glPushName(s);
            for(int i=0;i<debug_particles.regular_polys(f)(s).X(0).m;i++)
            {
                glPushName(i);
                auto& c=debug_particles.regular_polys(f)(s);
                OPENGL_COLOR(c.color(i)).Send_To_GL_Pipeline();
                OPENGL_SHAPES::Draw_Regular_Polygon(c.X(0)(i),c.t(0)(i)*sc,
                    c.i(0)(i),(T)pi/4*c.i(1)(i),f);
                glPopName();
            }
            glPopName();
        }
        glPopName();
    }
    glLoadName(6);
    for(int i=0;i<debug_particles.text.m;i++)
    {
        glPushName(i);
        auto& c=debug_particles.text(i);
        OPENGL_COLOR(c.color).Send_To_GL_Pipeline();
        glPushAttrib(GL_ENABLE_BIT);
        glDisable(GL_LIGHTING);
        OpenGL_String(c.X,c.str,GLUT_BITMAP_HELVETICA_12);
        glPopAttrib();
        glPopName();
    }
    glPopName();

    glPopMatrix();
    glEnable(GL_CULL_FACE);
    glPopAttrib();
    if(slice && slice->Is_Slice_Mode()) glPopAttrib();
}
//#####################################################################
// Function Use_Bounding_Box
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Use_Bounding_Box() const
{
    return draw && valid;
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Bounding_Box() const
{
    if(!valid || !draw) return RANGE<VECTOR<T,3> >::Centered_Box();
    RANGE<TV> box=debug_particles.Bounding_Box();
    return World_Space_Box(box);
}
//#####################################################################
// Function Get_Selection_Priority
//#####################################################################
template<class T> int OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Get_Selection_Priority(ARRAY_VIEW<GLuint> indices)
{
    if(indices.m<1) return -1;
    return 110;
}
//#####################################################################
// Function Get_Selection
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Set_Selection(ARRAY_VIEW<GLuint> indices,int modifiers)
{
    selected_index.Fill(-1);
    for(int i=0;i<std::min(indices.m,4);i++)
        selected_index(i)=indices(i);
    return true;
}
//#####################################################################
// Function Clear_Selection
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Clear_Selection()
{
    selected_index.Fill(-1);
}
//#####################################################################
// Function Print_Selection_Info
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Print_Selection_Info(std::ostream &output_stream) const
{
    int i,j,k,l;
    selected_index.Get(i,j,k,l);
    if(i<0) return;
    auto pr=[&output_stream](const auto& c,int i)
    {
        if(c.info(i).name.size()) output_stream<<"name: "<<c.info(i).name<<"\n";
        for(const auto& a:c.X) output_stream<<"X: "<<a(i)<<"\n";
        for(const auto& a:c.t) output_stream<<"r: "<<a(i)<<"\n";
        for(const auto& a:c.i) output_stream<<"i: "<<a(i)<<"\n";
        output_stream<<"color: "<<c.color(i)<<"\n";
        if(c.info(i).info.size()) output_stream<<"info:\n"<<c.info(i).info<<"\n";
    };
    switch(i)
    {
        case 0: pr(debug_particles.points,j); break;
        case 1: pr(debug_particles.segments,j); break;
        case 2: pr(debug_particles.tris(j),k); break;
        case 3: pr(debug_particles.balls(j)(k),l); break;
        case 4: pr(debug_particles.arrows(j)(k),l); break;
        case 5: pr(debug_particles.regular_polys(j)(k),l); break;
        case 6:
            {
                auto& c=debug_particles.text(i);
                if(c.info.name.size()) output_stream<<"name: "<<c.info.name<<"\n";
                output_stream<<"X: "<<c.color<<"\n";
                output_stream<<"label: "<<c.str<<"\n";
                output_stream<<"color: "<<c.color<<"\n";
                if(c.info.info.size()) output_stream<<"info:\n"<<c.info.info<<"\n";
                break;
            }
        default:break;
    }
}
//#####################################################################
// Function Destroy_Selection_After_Frame_Change
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Destroy_Selection_After_Frame_Change()
{
    return true;
}
//#####################################################################
// Function Selection_Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Selection_Bounding_Box() const
{
    RANGE<TV> box;
    int i,j,k,l;
    selected_index.Get(i,j,k,l);
    if(i<0) return RANGE<VECTOR<T,3> >::Centered_Box();
    auto pr=[&box](const auto& c,int i,bool all)
    {
        if(all) for(const auto& a:c.X) box.Enlarge_To_Include_Point(a(i));
        else box.Enlarge_To_Include_Point(c.X(0)(i));
        if(c.X.m==1 && c.i.m==1) box=box.Thickened(c.i(0)(i));
    };
    switch(i)
    {
        case 0: pr(debug_particles.points,j,true); break;
        case 1: pr(debug_particles.segments,j,true); break;
        case 2: pr(debug_particles.tris(j),k,true); break;
        case 3: pr(debug_particles.balls(j)(k),l,true); break;
        case 4: pr(debug_particles.arrows(j)(k),l,false); break;
        case 5: pr(debug_particles.regular_polys(j)(k),l,true); break;
        default:break;
    }
    if(box.Empty()) return RANGE<VECTOR<T,3> >::Centered_Box();
    return World_Space_Box(box);
}
//#####################################################################
// Function Reinitialize
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Reinitialize()
{
    if(!draw) return;
    valid=true;

    std::string frame_filename=viewer_dir.Frame_File(filename);
        
    try{Read_From_File(frame_filename,debug_particles);}
    catch(FILESYSTEM_ERROR&){valid=false;}
}
//#####################################################################
// Function Command_Prompt_Response
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Command_Prompt_Response()
{
    if(!OPENGL_WORLD<T>::Singleton()->prompt_response.empty()){
        std::string command;
        std::istringstream sstream(OPENGL_WORLD<T>::Singleton()->prompt_response);
        sstream>>command;
        if(command=="s"){
            ARRAY<int> indices;
            int index;
            while(sstream>>index) indices.Append(index);
            //Store_Point_Colors(false);
            //Set_Point_Colors(indices,OPENGL_COLOR::Yellow());
        }}
}
//#####################################################################
// Function Set_Slice
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Set_Slice(OPENGL_SLICE *slice_input)
{
    slice=slice_input;
    Slice_Has_Changed();
}
//#####################################################################
// Function Command_Prompt
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEBUG_PARTICLES_2D<T>::
Command_Prompt()
{
    OPENGL_WORLD<T>::Singleton()->Prompt_User("Command: ",{[this](){Command_Prompt_Response();},""});
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_COMPONENT_DEBUG_PARTICLES_2D<float>;
template class OPENGL_COMPONENT_DEBUG_PARTICLES_2D<double>;
}
