//#####################################################################
// Copyright 2004-2009, Eran Guendelman, Geoffrey Irving, Frank Losasso, Tamar Shinar, Eftychios Sifakis, Jonathan Su, Wen Zheng.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <OpenGL/OpenGL_Components/OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D.h>
#include <Core/Log/LOG.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Geometry/Meshes/TRIANGULATED_AREA.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_BODY_COLLECTION.h>
#include <Deformables/Deformable_Objects/DEFORMABLE_PARTICLES.h>
#include <OpenGL/OpenGL/OPENGL_LEVELSET_COLOR_MAP.h>
#include <OpenGL/OpenGL/OPENGL_SEGMENTED_CURVE_2D.h>
#include <OpenGL/OpenGL/OPENGL_TRIANGULATED_AREA.h>
#include <OpenGL/OpenGL/OPENGL_WORLD.h>
using namespace PhysBAM;
//#####################################################################
// Function OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D
//#####################################################################
template<class T> OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D(const VIEWER_DIR& viewer_dir)
    :OPENGL_COMPONENT<T>(viewer_dir,"Deformable Object List"),
    valid(false),display_mode(0),draw_velocities(false),velocity_scale(0.025),
    deformable_body_collection(*new DEFORMABLE_BODY_COLLECTION<TV>(0)),
    velocity_field(deformable_body_collection.particles.V,deformable_body_collection.particles.X),color_map(OPENGL_INDEXED_COLOR_MAP::Basic_16_Color_Map()),
    selected_segmented_curve(-1),selected_triangulated_area(-1),selected_triangles_of_material(-1),
    selected_embedded_curve(-1),selected_free_particles(-1),
    selected_index(-1),selected_object(0)
{
    viewer_callbacks.Set("increase_vector_size",{[this](){Increase_Vector_Size();},"Increase vector size"});
    viewer_callbacks.Set("decrease_vector_size",{[this](){Decrease_Vector_Size();},"Decrease vector size"});
    viewer_callbacks.Set("reset_vector_size",{[this](){Set_Vector_Size(1);},"Reset vector size"});
    viewer_callbacks.Set("toggle_draw_velocities",{[this](){Toggle_Draw_Velocities();},"Toggle draw velocities"});
    viewer_callbacks.Set("cycle_display_mode",{[this](){Cycle_Display_Mode();},"Cycle embedded display mode"});
    // check for per frame structures
    if(File_Exists(viewer_dir.Frame_File("deformable_object_structures")))
        invalidate_deformable_objects_selection_each_frame=true;
    else invalidate_deformable_objects_selection_each_frame=false;

    Initialize();
}
//#####################################################################
// Function ~OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D
//#####################################################################
template<class T> OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
~OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D()
{
    delete color_map;
    delete &deformable_body_collection;
}
//#####################################################################
// Function Initialize
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Initialize()
{}
//#####################################################################
// Function Reinitialize
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Reinitialize()
{
    if(!draw) return;
    static bool first_time=true;
    bool frame_has_static=File_Exists(viewer_dir.Frame_File("deformable_object_structures"));
    bool read_static_variables=frame_has_static || first_time || !deformable_body_collection.structures.m;
    deformable_body_collection.Read(viewer_dir,read_static_variables);

    if(read_static_variables){
        int m=deformable_body_collection.structures.m;
        active_list.Clean_Memory();active_list.Resize(m);active_list.Fill(true);
        embedded_curve_objects.Delete_Pointers_And_Clean_Memory();embedded_curve_objects.Resize(m);
        segmented_curve_objects.Delete_Pointers_And_Clean_Memory();segmented_curve_objects.Resize(m);
        triangulated_area_objects.Delete_Pointers_And_Clean_Memory();triangulated_area_objects.Resize(m);
        triangles_of_material_objects.Delete_Pointers_And_Clean_Memory();triangles_of_material_objects.Resize(m);
        free_particles_objects.Delete_Pointers_And_Clean_Memory();free_particles_objects.Resize(m);
        free_particles_indirect_arrays.Delete_Pointers_And_Clean_Memory();free_particles_indirect_arrays.Resize(m);
        phi_list.Delete_Pointers_And_Clean_Memory();phi_list.Resize(m);
        int color_map_index=14;
        for(int i=0;i<m;i++){
            STRUCTURE<TV>* structure=deformable_body_collection.structures(i);
            if(SEGMENTED_CURVE<TV>* segmented_curve=dynamic_cast<SEGMENTED_CURVE<TV>*>(structure)){
                if(first_time) LOG::cout<<"object "<<i<<": segmented curve\n";
                segmented_curve_objects(i)=new OPENGL_SEGMENTED_CURVE_2D<T>(*segmented_curve,color_map->Lookup(color_map_index--));
                segmented_curve_objects(i)->draw_velocities=draw_velocities;
                segmented_curve_objects(i)->velocity_scale=velocity_scale;} // apply current parameters
            else if(TRIANGULATED_AREA<T>* triangulated_area=dynamic_cast<TRIANGULATED_AREA<T>*>(structure)){
                if(first_time) LOG::cout<<"object "<<i<<": triangulated area\n";
                triangulated_area->mesh.Initialize_Segment_Mesh(); // to enable segment selection
                triangulated_area_objects(i)=new OPENGL_TRIANGULATED_AREA<T>(*triangulated_area,true);}
            else PHYSBAM_FATAL_ERROR(LOG::sprintf("Weird object %d",i));}}
    for(int i=0;i<deformable_body_collection.structures.m;i++){
        std::string suffix=LOG::sprintf("_%d",i);
        if(File_Exists(viewer_dir.Frame_File("stress_map_of_triangulated_area")+suffix)){
            if(first_time) LOG::cout<<"adding stress map to triangulated area"<<std::endl;
            ARRAY<OPENGL_COLOR > *color_map=new ARRAY<OPENGL_COLOR >;
            Read_From_File(viewer_dir.Frame_File("stress_map_of_triangulated_area")+suffix,*color_map);
            triangulated_area_objects(i)->Set_Color_Map(color_map);}}
    valid=true;
    first_time=false;
}
//#####################################################################
// Function Set_Frame
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Set_Frame()
{
    Reinitialize();
}
//#####################################################################
// Function Set_Draw
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Set_Draw(bool draw_input)
{
    OPENGL_COMPONENT<T>::Set_Draw(draw_input);
    if(draw_input) active_list.Fill(true);
    Reinitialize();
}
//#####################################################################
// Function Display
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Display() const
{
    if(!draw||!valid) return;
    bool draw_triangulated_areas=false,draw_triangles_of_material=false;
    switch(display_mode){
        case 0:draw_triangulated_areas=true;break;
        case 1:draw_triangles_of_material=true;break;
        case 2:draw_triangles_of_material=true;break;
        case 3:draw_triangulated_areas=true;break;
        case 4:draw_triangulated_areas=true;draw_triangles_of_material=true;break;}

    glPushName(0);
    glPushName(0);
    for(int i=0;i<segmented_curve_objects.m;i++){
        if(!active_list(i)) continue;
        if(segmented_curve_objects(i)){
            glLoadName(i);
            segmented_curve_objects(i)->Display();}}
    glPopName();
    glPopName();

    glPushName(1);
    glPushName(0);
    if(draw_triangulated_areas){
        for(int i=0;i<triangulated_area_objects.m;i++){
            if(!active_list(i)) continue;
            if(triangulated_area_objects(i)){
                glLoadName(i);
                triangulated_area_objects(i)->Display();}}}
    glPopName();
    glPopName();

    glPushName(2);
    glPushName(0);
    if(draw_triangles_of_material){
        for(int i=0;i<triangles_of_material_objects.m;i++){
            if(!active_list(i)) continue;
            if(triangles_of_material_objects(i)){
                glLoadName(i);
                triangles_of_material_objects(i)->Display();}}}
    glPopName();
    glPopName();

    glPushName(5);
    glPushName(0);
    for(int i=0;i<segmented_curve_objects.m;i++){
        if(!active_list(i)) continue;
        if(free_particles_objects(i) && display_mode!=1){
            glLoadName(i);
            free_particles_objects(i)->Display();}}
    glPopName();
    glPopName();

    if(draw_velocities && deformable_body_collection.particles.store_velocity) velocity_field.Display();
}
//#####################################################################
// Function Increase_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Increase_Vector_Size()
{
    Set_Vector_Size(velocity_scale*(T)1.1);
}
//#####################################################################
// Function Decrease_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Decrease_Vector_Size()
{
    Set_Vector_Size(velocity_scale/(T)1.1);
}
//#####################################################################
// Function Set_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Set_Vector_Size(T scale)
{
    velocity_scale=scale;
    for(int i=0;i<segmented_curve_objects.m;i++)
        if(active_list(i) && segmented_curve_objects(i))
            segmented_curve_objects(i)->velocity_scale=velocity_scale; // apply to objects which already exist
    for(int i=0;i<triangulated_area_objects.m;i++)
        if(active_list(i) && triangulated_area_objects(i))
            triangulated_area_objects(i)->velocity_scale=velocity_scale;
    velocity_field.Set_Vector_Size(velocity_scale);
    for(int i=0;i<embedded_curve_objects.m;i++)
        if(active_list(i) && embedded_curve_objects(i))
            embedded_curve_objects(i)->velocity_scale=velocity_scale; // apply to objects which already exist
}
//#####################################################################
// Function Toggle_Draw_Velocities
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Toggle_Draw_Velocities()
{
    draw_velocities=!draw_velocities;
    for(int i=0;i<segmented_curve_objects.m;i++)
        if(active_list(i) && segmented_curve_objects(i))
            segmented_curve_objects(i)->draw_velocities=draw_velocities;
    for(int i=0;i<triangulated_area_objects.m;i++)
        if(active_list(i) && triangulated_area_objects(i))
            triangulated_area_objects(i)->draw_velocities=draw_velocities;
    for(int i=0;i<embedded_curve_objects.m;i++)
        if(active_list(i) && embedded_curve_objects(i))
            embedded_curve_objects(i)->draw_velocities=draw_velocities;
}
//#####################################################################
// Function Cycle_Display_Mode
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Cycle_Display_Mode()
{
    display_mode=(display_mode+1)%5;
}
//#####################################################################
// Function Use_Bounding_Box
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Use_Bounding_Box() const
{
    return draw && valid && deformable_body_collection.structures.m>0;
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Bounding_Box() const
{
    RANGE<VECTOR<T,3> > box;
    if(draw && valid && deformable_body_collection.structures.m>0){
        for(int i=0;i<segmented_curve_objects.m;i++)
            if(active_list(i) && segmented_curve_objects(i))
                box.Enlarge_To_Include_Box(segmented_curve_objects(i)->Bounding_Box());
        for(int i=0;i<triangulated_area_objects.m;i++)
            if(active_list(i) && triangulated_area_objects(i))
                box.Enlarge_To_Include_Box(triangulated_area_objects(i)->Bounding_Box());
        for(int i=0;i<triangles_of_material_objects.m;i++)
            if(active_list(i) && triangles_of_material_objects(i))
                box.Enlarge_To_Include_Box(triangles_of_material_objects(i)->Bounding_Box());}
    return box;
}
//#####################################################################
// Function Get_Selection_Priority
//#####################################################################
template<class T> int OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Get_Selection_Priority(ARRAY_VIEW<GLuint> indices)
{
    switch(indices(0)){
        case 0:return segmented_curve_objects(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
        case 1:return triangulated_area_objects(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
        case 2:return triangles_of_material_objects(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
        case 4:return embedded_curve_objects(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
        case 5:return free_particles_objects(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
        default:return 10;}
}
//#####################################################################
// Function Get_Selection
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Set_Selection(ARRAY_VIEW<GLuint> indices,int modifiers)
{
    if(modifiers&GLUT_ACTIVE_CTRL){
        active_list(indices(1))=false;
        return false;}

    selected_index=indices(1);
    switch(indices(0)){
        case 0:
            selected_segmented_curve=indices(1);
            selected_object=segmented_curve_objects(indices(1));
            break;
        case 1:
            selected_triangulated_area=indices(1);
            selected_object=triangulated_area_objects(indices(1));
            break;
        case 2:
            selected_triangles_of_material=indices(1);
            selected_object=triangles_of_material_objects(indices(1));
            break;
        case 4:
            selected_embedded_curve=indices(1);
            selected_object=embedded_curve_objects(indices(1));
            break;
        case 5:
            selected_free_particles=indices(1);
            selected_object=free_particles_objects(indices(1));
            break;
        default:PHYSBAM_FATAL_ERROR();}
    return selected_object->Set_Selection(indices.Array_View(2,indices.m-2),modifiers);
}
//#####################################################################
// Function Clear_Highlight
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Clear_Selection()
{
    for(int i=0;i<triangulated_area_objects.m;i++)if(triangulated_area_objects(i))triangulated_area_objects(i)->Clear_Selection();
    for(int i=0;i<triangles_of_material_objects.m;i++)if(triangles_of_material_objects(i))triangles_of_material_objects(i)->Clear_Selection();
    for(int i=0;i<free_particles_objects.m;i++)if(free_particles_objects(i))free_particles_objects(i)->Clear_Selection();
    for(int i=0;i<embedded_curve_objects.m;i++)if(embedded_curve_objects(i))embedded_curve_objects(i)->Clear_Selection();
}
//#####################################################################
// Function Print_Selection_Info
//#####################################################################
template<class T> void OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Print_Selection_Info(std::ostream& output_stream) const
{
    output_stream<<"Deformable object "<<selected_index<<std::endl;
    selected_object->Print_Selection_Info(output_stream);
}
//#####################################################################
// Function Print_Selection_Info
//#####################################################################
template<class T> bool OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Destroy_Selection_After_Frame_Change()
{
    if(invalidate_deformable_objects_selection_each_frame){
        Clear_Selection();
        return true;}
    return false;
}
//#####################################################################
// Function Selection_Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<T>::
Selection_Bounding_Box() const
{
    return World_Space_Box(selected_object->Bounding_Box());
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<float>;
template class OPENGL_COMPONENT_DEFORMABLE_BODY_COLLECTION_2D<double>;
}
