//#####################################################################
// Copyright 2004-2009, Eran Guendelman, Geoffrey Irving, Tamar Shinar, Eftychios Sifakis, Rachel Weinstein, Wen Zheng.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <OpenGL/OpenGL_Components/OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D.h>
#include <Core/Arrays/INDIRECT_ARRAY.h>
#include <Core/Matrices/MATRIX_MXN.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_TRANSFORMED.h>
#include <Geometry/Implicit_Objects/LEVELSET_IMPLICIT_OBJECT.h>
#include <Geometry/Meshes/TRIANGULATED_AREA.h>
#include <Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <Rigids/Rigid_Bodies/RIGID_BODY_COLLECTION.h>
#include <OpenGL/OpenGL/OPENGL_AXES.h>
#include <OpenGL/OpenGL/OPENGL_LEVELSET_2D.h>
#include <OpenGL/OpenGL/OPENGL_SEGMENTED_CURVE_2D.h>
#include <OpenGL/OpenGL/OPENGL_SHAPES.h>
#include <OpenGL/OpenGL/OPENGL_TRIANGULATED_AREA.h>
#include <OpenGL/OpenGL_Components/OPENGL_COMPONENT_TRIANGULATED_AREA_BASED_VECTOR_FIELD.h>
using namespace PhysBAM;
//#####################################################################
// Constructor
//#####################################################################
template<class T> OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D(const VIEWER_DIR& viewer_dir)
    :OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D(viewer_dir,*new RIGID_BODY_COLLECTION<TV>)
{
    need_destroy_rigid_body_collection=true;
}
//#####################################################################
// Constructor
//#####################################################################
template<class T> OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D(const VIEWER_DIR& viewer_dir,RIGID_BODY_COLLECTION<TV>& rigid_body_collection)
    :OPENGL_COMPONENT<T>(viewer_dir,"Rigid Geometry Collection 2D"),valid(false),show_object_names(false),output_positions(true),draw_velocity_vectors(false),
    draw_individual_axes(false),draw_node_velocity_vectors(false),draw_segmented_curve(true),draw_triangulated_area(false),draw_implicit_curve(false),
    need_destroy_rigid_body_collection(false),selected_curve(-1),selected_area(-1),selected_joint_id(-1),
    rigid_body_collection(rigid_body_collection),
    velocity_field(velocity_vectors,positions,OPENGL_COLOR::Cyan(),0.25,true,true),node_velocity_field(node_velocity_vectors,node_positions,OPENGL_COLOR::Magenta(),0.25,true,true)
{
    viewer_callbacks.Set("toggle_velocity_vectors",{[this](){Toggle_Velocity_Vectors();},"Toggle velocity vectors"});
    viewer_callbacks.Set("toggle_individual_axes",{[this](){Toggle_Individual_Axes();},"Toggle individual axes"});
    viewer_callbacks.Set("toggle_output_positions",{[this](){Toggle_Output_Positions();},"Toggle output positions"});
    viewer_callbacks.Set("toggle_show_object_names",{[this](){Toggle_Show_Object_Names();},"Toggle show object names"});
    viewer_callbacks.Set("toggle_node_velocity_vectors",{[this](){Toggle_Node_Velocity_Vectors();},"Toggle node velocity vectors"});
    viewer_callbacks.Set("toggle_draw_mode",{[this](){Toggle_Draw_Mode();},"Toggle draw mode"});
    viewer_callbacks.Set("increase_vector_size",{[this](){Increase_Vector_Size();},"Increase vector size"});
    viewer_callbacks.Set("decrease_vector_size",{[this](){Decrease_Vector_Size();},"Decrease vector size"});
    viewer_callbacks.Set("reset_vector_size",{[this](){Set_Vector_Size(1);},"Reset vector size"});

    has_init_destroy_information=true;
}
//#####################################################################
// Destructor
//#####################################################################
template<class T> OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
~OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D()
{
    if(need_destroy_rigid_body_collection) delete &rigid_body_collection;
}
//#####################################################################
// Function Set_Draw_Mode
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Draw_Mode(const int mode)
{
    draw_segmented_curve=(mode&0x01)!=0;
    draw_triangulated_area=(mode&0x02)!=0;
    draw_implicit_curve=(mode&0x04)!=0;
}
//#####################################################################
// Function Get_Draw_Mode
//#####################################################################
template<class T> int OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Get_Draw_Mode() const
{
    return 0x04*(int)(draw_implicit_curve)+0x02*(int)(draw_triangulated_area)+0x01*(int)(draw_segmented_curve);
}
//#####################################################################
// Function Reinitialize
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Reinitialize(const bool read_geometry)
{
    if(!draw) return;
    valid=false;
    if(!File_Exists(viewer_dir.Frame_File("rigid_body_particles"))) return;

    rigid_body_collection.Read(viewer_dir,&needs_init,&needs_destroy); // TODO: avoiding reading triangulated areas
    if(has_init_destroy_information) for(int i=0;i<needs_destroy.m;i++) Destroy_Geometry(needs_destroy(i));

    int max_number_of_bodies(max(opengl_segmented_curve.Size(),rigid_body_collection.rigid_body_particles.Size()));
    // only enlarge array as we read in more geometry to memory
    opengl_segmented_curve.Resize(max_number_of_bodies);
    opengl_triangulated_area.Resize(max_number_of_bodies);
    opengl_levelset.Resize(max_number_of_bodies);
    extra_components.Resize(max_number_of_bodies);
    opengl_axes.Resize(max_number_of_bodies);
    draw_object.Resize(max_number_of_bodies);draw_object.Fill(true);
    use_object_bounding_box.Resize(max_number_of_bodies);use_object_bounding_box.Fill(true);

    // Initialize bodies which have become active
    if(has_init_destroy_information) for(int i=0;i<needs_init.m;i++){
            int id=needs_init(i);PHYSBAM_ASSERT(rigid_body_collection.Is_Active(id));
            Create_Geometry(id);}

    // Only display real bodies (not ghost bodies)
    if(File_Exists(viewer_dir.Frame_File("partition"))) {
        ARRAY<int> particles_of_this_partition;
        Read_From_File(viewer_dir.Frame_File("partition"),particles_of_this_partition);
        for(int i=0;i<max_number_of_bodies;i++)
            draw_object(i)=false;
        for(int i=0;i<particles_of_this_partition.Size();i++)
            draw_object(particles_of_this_partition(i))=true;}

    // Update active bodies / remove inactive bodies
    for(int id=0;id<rigid_body_collection.rigid_body_particles.Size();id++){
        if(rigid_body_collection.Is_Active(id)) Update_Geometry(id);
        else Destroy_Geometry(id);}
    for(int id=rigid_body_collection.rigid_body_particles.Size();id<opengl_segmented_curve.Size();id++) Destroy_Geometry(id);
    if(File_Exists(viewer_dir.Frame_File("colors")))
        Read_From_File(viewer_dir.Frame_File("colors"),colors);
    for(int id=0;id<colors.m;id++){
        if(colors(id)==0) Set_Object_Color(id,OPENGL_COLOR::Green());
        if(colors(id)==1) Set_Object_Color(id,OPENGL_COLOR::Magenta());}

    valid=true;
    Update_Object_Labels();
}
//#####################################################################
// Function Create_Geometry
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Create_Geometry(const int id)
{
    RIGID_BODY<TV>& rigid_body=rigid_body_collection.Rigid_Body(id);

    if(!opengl_axes(id)){opengl_axes(id)=new OPENGL_AXES<T>;}
    if(rigid_body.simplicial_object && !opengl_segmented_curve(id)){
        opengl_segmented_curve(id)=new OPENGL_SEGMENTED_CURVE_2D<T>(*rigid_body.simplicial_object);
        opengl_segmented_curve(id)->draw_vertices=true;
        opengl_segmented_curve(id)->Enslave_Transform_To(*opengl_axes(id));}

    // add triangulated area
    TRIANGULATED_AREA<T>* triangulated_area=rigid_body.template Find_Structure<TRIANGULATED_AREA<T>*>();
    if(triangulated_area && !opengl_triangulated_area(id)){
        triangulated_area->mesh.Initialize_Segment_Mesh();
        opengl_triangulated_area(id)=new OPENGL_TRIANGULATED_AREA<T>(*triangulated_area);
        opengl_triangulated_area(id)->Enslave_Transform_To(*opengl_axes(id));}

    if(rigid_body.implicit_object && !opengl_levelset(id)){ // ASSUMES LEVELSET_IMPLICIT_CURVE!
        IMPLICIT_OBJECT<TV>* object_space_implicit_object=rigid_body.implicit_object->object_space_implicit_object;
        if(LEVELSET_IMPLICIT_OBJECT<TV>* levelset_implicit_object=dynamic_cast<LEVELSET_IMPLICIT_OBJECT<TV>*>(object_space_implicit_object)){
            opengl_levelset(id)=new OPENGL_LEVELSET_2D<T>(levelset_implicit_object->levelset);
            // TODO: fix me
            //opengl_levelset(id)->Set_Color_Mode(OPENGL_LEVELSET_2D<T>::COLOR_GRADIENT);
            opengl_levelset(id)->draw_cells=true;opengl_levelset(id)->draw_curve=false;opengl_levelset(id)->draw_area=false;
            opengl_levelset(id)->Enslave_Transform_To(*opengl_axes(id));}}

    // add extra components
    if(opengl_triangulated_area(id)){
        std::string filename=viewer_dir.Frame_File("accumulated_impulses_")+std::to_string(id);
        if(File_Exists(filename)){
            LOG::cout<<"Adding accumulated impulses to rigid body "<<id<<std::endl;
            OPENGL_COMPONENT_TRIANGULATED_AREA_BASED_VECTOR_FIELD<T>* component=
                new OPENGL_COMPONENT_TRIANGULATED_AREA_BASED_VECTOR_FIELD<T>(viewer_dir,*triangulated_area,filename);
            component->opengl_vector_field.Enslave_Transform_To(*opengl_axes(id));
            extra_components(id).Append(component);}}
}
//#####################################################################
// Function Update_Geometry
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Update_Geometry(const int id)
{
    if(opengl_axes(id)) *opengl_axes(id)->frame=Convert_2d_To_3d(rigid_body_collection.Rigid_Body(id).Frame());
    if(opengl_triangulated_area(id)){
        std::string color_map_filename=viewer_dir.Frame_File("stress_map_of_triangulated_area_")+std::to_string(id);
        if(File_Exists(color_map_filename)){
            if(!opengl_triangulated_area(id)->color_map) opengl_triangulated_area(id)->color_map=new ARRAY<OPENGL_COLOR>;
            Read_From_File(color_map_filename,*opengl_triangulated_area(id)->color_map);}
        else if(opengl_triangulated_area(id)->color_map){delete opengl_triangulated_area(id)->color_map;opengl_triangulated_area(id)->color_map=0;}}
    for(int i=0;i<extra_components(id).m;i++) extra_components(id)(i)->Set_Frame();
}
//#####################################################################
// Function Destroy_Geometry
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Destroy_Geometry(const int id)
{
    if(opengl_segmented_curve(id)){delete opengl_segmented_curve(id);opengl_segmented_curve(id)=0;}
    if(opengl_triangulated_area(id)){delete opengl_triangulated_area(id)->color_map;delete opengl_triangulated_area(id);opengl_triangulated_area(id)=0;}
    if(opengl_levelset(id)){delete opengl_levelset(id);opengl_levelset(id)=0;}
    if(opengl_axes(id)){delete opengl_axes(id);opengl_axes(id)=0;}
    extra_components(id).Delete_Pointers_And_Clean_Memory();
    draw_object(id)=false;
}
//#####################################################################
// Function Update_Object_Labels
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Update_Object_Labels()
{
    int number_of_drawn_bodies=draw_object.Count_Matches(true);
    if(draw_velocity_vectors){
        positions.Resize(number_of_drawn_bodies);
        velocity_vectors.Resize(number_of_drawn_bodies);}
    if(draw_node_velocity_vectors){
        node_positions.Resize(number_of_drawn_bodies*4);
        node_velocity_vectors.Resize(number_of_drawn_bodies*4);}

    int idx=0;
    for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++){
        if(draw_object(i)){
            if(draw_velocity_vectors || draw_node_velocity_vectors){
                if(draw_velocity_vectors){
                    positions(idx)=rigid_body_collection.rigid_body_particles.frame(i).t;
                    velocity_vectors(idx)=rigid_body_collection.rigid_body_particles.twist(i).linear;}
                if(draw_node_velocity_vectors){
                    //only valid for squares . . .
                    RIGID_BODY<TV>* rigid_body=&rigid_body_collection.Rigid_Body(i);

                    node_positions(idx*4)=rigid_body->World_Space_Point(VECTOR<T,2>(1,1));
                    node_velocity_vectors(idx*4)=rigid_body->Pointwise_Object_Velocity(rigid_body->World_Space_Vector(VECTOR<T,2>(1,1)));

                    node_positions(idx*4+1)=rigid_body->World_Space_Point(VECTOR<T,2>(1,-1));
                    node_velocity_vectors(idx*4+1)=rigid_body->Pointwise_Object_Velocity(rigid_body->World_Space_Vector(VECTOR<T,2>(1,-1)));

                    node_positions(idx*4+2)=rigid_body->World_Space_Point(VECTOR<T,2>(-1,1));
                    node_velocity_vectors(idx*4+2)=rigid_body->Pointwise_Object_Velocity(rigid_body->World_Space_Vector(VECTOR<T,2>(-1,1)));

                    node_positions(idx*4+3)=rigid_body->World_Space_Point(VECTOR<T,2>(-1,-1));
                    node_velocity_vectors(idx*4+3)=rigid_body->Pointwise_Object_Velocity(rigid_body->World_Space_Vector(VECTOR<T,2>(-1,-1)));}

                idx++;}
            if(opengl_segmented_curve(i)){
                if(output_positions){
                    opengl_segmented_curve(i)->Set_Name(
                        LOG::sprintf("%s <%.3f %.3f> [w=%.3f]",
                            rigid_body_collection.Rigid_Body(i).name.c_str(),
                            rigid_body_collection.rigid_body_particles.frame(i).t.x,
                            rigid_body_collection.rigid_body_particles.frame(i).t.y,
                            rigid_body_collection.rigid_body_particles.twist(i).angular.x));}
                else opengl_segmented_curve(i)->Set_Name(rigid_body_collection.Rigid_Body(i).name);}}}
    for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++){
        if(draw_object(i)){
            if(opengl_segmented_curve(i)){
                if(output_positions){
                rigid_body_collection.Rigid_Body(i).Update_Angular_Velocity();}}}}
}
//#####################################################################
// Function Set_Frame
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Frame()
{
    
    Reinitialize();
}
//#####################################################################
// Function Set_Draw
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Draw(bool draw_input)
{
    OPENGL_COMPONENT<T>::Set_Draw(draw_input);
}
//#####################################################################
// Function Use_Bounding_Box
//#####################################################################
template<class T> bool OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Use_Bounding_Box() const
{
    int num_drawable_objects=0;
    for(int i=0;i<opengl_segmented_curve.Size();i++)
        if(draw_object(i) && use_object_bounding_box(i))
            num_drawable_objects++;
    return (draw && num_drawable_objects>0);
}
//#####################################################################
// Function Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Bounding_Box() const
{
    RANGE<VECTOR<T,3> > box;
    if(draw)
        for(int i=0;i<opengl_segmented_curve.Size();i++)
            if(draw_object(i) && use_object_bounding_box(i) && opengl_segmented_curve(i)){
                box=RANGE<VECTOR<T,3> >::Combine(box,opengl_segmented_curve(i)->Bounding_Box());}
    return box;
}
//#####################################################################
// Function Get_Selection_Priority
//#####################################################################
template<class T> int OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Get_Selection_Priority(ARRAY_VIEW<GLuint> indices)
{
    if(!indices.m) return -1;
    const static int priority[]={0,0,95,95};
    if(indices(0)==0) // segmented curve
        return opengl_segmented_curve(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
    if(indices(0)==1) // triangulated area
        return opengl_triangulated_area(indices(1))->Get_Selection_Priority(indices.Array_View(2,indices.m-2));
    return priority[indices(0)];
}
//#####################################################################
// Function Get_Selection
//#####################################################################
template<class T> bool OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Selection(ARRAY_VIEW<GLuint> indices,int modifiers)
{
    if(indices(0)==0 && (modifiers&GLUT_ACTIVE_CTRL)){
        draw_object(indices(1))=false;
        return false;}
    if(indices(0)==0){
        selected_curve=indices(1);
        return opengl_segmented_curve(indices(1))->Set_Selection(indices.Array_View(2,indices.m-2),modifiers);}
    else if(indices(0)==1){
        selected_area=indices(1);
        return opengl_triangulated_area(indices(1))->Set_Selection(indices.Array_View(2,indices.m-2),modifiers);}
    else return false;
    return true;
}
//#####################################################################
// Function Clear_Highlight
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Clear_Selection()
{
    if(selected_curve>=0){
        opengl_segmented_curve(selected_curve)->Clear_Selection();
        selected_curve=-1;}
    if(selected_area>=0){
        opengl_triangulated_area(selected_area)->Clear_Selection();
        selected_area=-1;}
    selected_joint_id=-1;
}
//#####################################################################
// Function Print_Selection_Info
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Print_Selection_Info(std::ostream &output_stream) const
{
    int selected_body=selected_curve>=0?selected_curve:selected_area;
    if(selected_body>=0){
        output_stream<<"Rigid body "<<selected_body<<std::endl;

        // handle case of body no longer being active
        if(!rigid_body_collection.Is_Active(selected_body) || !opengl_segmented_curve(selected_body)){
            output_stream<<"INACTIVE"<<std::endl;
            return;}

        RIGID_BODY<TV> *body=const_cast<RIGID_BODY<TV>*>(&rigid_body_collection.Rigid_Body(selected_body));
        //body->Update_Angular_Velocity();

        if(!body->name.empty()){output_stream<<"Name = "<<body->name<<std::endl;}
        rigid_body_collection.rigid_body_particles.Print(output_stream,selected_body);

        MATRIX<T,3> body_transform=body->Frame().Matrix();

        if(selected_curve>=0){
            opengl_segmented_curve(selected_curve)->Print_Selection_Info(output_stream,&body_transform);
            int vertex=opengl_segmented_curve(selected_curve)->selected_vertex;
            if(vertex>=0){
                TV position=body->simplicial_object->particles.X(vertex);
                output_stream<<"Pointwise velocity = "<<body->Pointwise_Object_Velocity(body->World_Space_Point(position))<<std::endl;}}
        if(selected_area>=0)
            opengl_triangulated_area(selected_area)->Print_Selection_Info(output_stream,&body_transform);}
}
//#####################################################################
// Function Set_Draw_Object
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Draw_Object(int i,bool draw_it)
{
    if(i>draw_object.Size()) draw_object.Resize(i);
    draw_object(i)=draw_it;
}
//#####################################################################
// Function Set_Object_Color
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Object_Color(int i,const OPENGL_COLOR &color_input)
{
    if(!opengl_segmented_curve(i)) return;
    opengl_segmented_curve(i)->color=color_input;
    opengl_segmented_curve(i)->vertex_color=color_input;
}
//#####################################################################
// Function Get_Draw_Object
//#####################################################################
template<class T> bool OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Get_Draw_Object(int i) const
{
    return draw_object(i);
}
//#####################################################################
// Function Set_Use_Object_Bounding_Box
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Use_Object_Bounding_Box(int i,bool use_it)
{
    use_object_bounding_box(i)=use_it;
}
//#####################################################################
// Function Toggle_Velocity_Vectors
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Velocity_Vectors()
{
    draw_velocity_vectors=!draw_velocity_vectors;
    Reinitialize();
}
//#####################################################################
// Function Toggle_Individual_Axes
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Individual_Axes()
{
    draw_individual_axes=!draw_individual_axes;
    Reinitialize();
}
//#####################################################################
// Function Toggle_Output_Positions
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Output_Positions()
{
    output_positions=!output_positions;
    Update_Object_Labels();
}
//#####################################################################
// Function Toggle_Show_Object_Names
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Show_Object_Names()
{
    show_object_names=!show_object_names;
}
//#####################################################################
// Function Toggle_Node_Velocity_Vectors
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Node_Velocity_Vectors()
{
    draw_node_velocity_vectors=!draw_node_velocity_vectors;
    Reinitialize();
}
//#####################################################################
// Function Increase_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Increase_Vector_Size()
{
    velocity_field.Scale_Vector_Size(1.1);
    node_velocity_field.Scale_Vector_Size(1.1);
}
//#####################################################################
// Function Decrease_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Decrease_Vector_Size()
{
    velocity_field.Scale_Vector_Size(1/1.1);
    node_velocity_field.Scale_Vector_Size(1/1.1);
}
//#####################################################################
// Function Reset_Vector_Size
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Set_Vector_Size(T scale)
{
    velocity_field.Set_Vector_Size(scale);
    node_velocity_field.Set_Vector_Size(scale);
}
//#####################################################################
// Function Toggle_Draw_Mode
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Toggle_Draw_Mode()
{
    int mode=Get_Draw_Mode();
    mode=(mode+1)%8;
    Set_Draw_Mode(mode);
}
//#####################################################################
// Selection Bounding_Box
//#####################################################################
template<class T> RANGE<VECTOR<T,3> > OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Selection_Bounding_Box() const
{
    if(selected_curve>=0) return opengl_segmented_curve(selected_curve)->Selection_Bounding_Box();
    if(selected_area>=0) return opengl_triangulated_area(selected_area)->Selection_Bounding_Box();
    return RANGE<VECTOR<T,3> >::Centered_Box();
}
//#####################################################################
// Function Display
//#####################################################################
template<class T> void OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<T>::
Display() const
{
    if(draw){
        glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT);
        glDisable(GL_DEPTH_TEST);
        glDisable(GL_LIGHTING);

        GLint mode=0;
        glGetIntegerv(GL_RENDER_MODE,&mode);

        if(draw_segmented_curve){
            glPushName(0);
            glPushName(0);
            for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++){
                glLoadName(Value(i));
                if(draw_object(i) && opengl_segmented_curve(i)) opengl_segmented_curve(i)->Display();}
            glPopName();
            glPopName();}
        if(draw_triangulated_area){
            glPushName(1);
            glPushName(0);
            for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++){
                glLoadName(Value(i));
                if(draw_object(i) && opengl_triangulated_area(i)) opengl_triangulated_area(i)->Display();}
            glPopName();
            glPopName();}
        if(draw_implicit_curve){
            glPushName(2);
            glPushName(0);
            for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++){
                glLoadName(Value(i));
                if(draw_object(i) && opengl_levelset(i)) opengl_levelset(i)->Display();}
            glPopName();
            glPopName();}
        if(draw_individual_axes)
            for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++)
                if(draw_object(i) && opengl_axes(i)) opengl_axes(i)->Display();

        if(mode!=GL_SELECT)
        {
            if(draw_velocity_vectors) velocity_field.Display();
            if(draw_node_velocity_vectors) node_velocity_field.Display();

            for(int i=0;i<extra_components.Size();i++)
                for(int j=0;j<extra_components(i).m;j++)
                    extra_components(i)(j)->Display();

            if(show_object_names){
                glColor3f(1,1,1);
                for(int i=0;i<rigid_body_collection.rigid_body_particles.Size();i++)
                    if(draw_object(i) && rigid_body_collection.Rigid_Body(i).name.length())
                        OpenGL_String(rigid_body_collection.rigid_body_particles.frame(i).t,rigid_body_collection.Rigid_Body(i).name);}
        }
        glPopAttrib();}
}
//#####################################################################
namespace PhysBAM{
template class OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<float>;
template class OPENGL_COMPONENT_RIGID_BODY_COLLECTION_2D<double>;
}
