//#####################################################################
// Copyright 2006-2009, Craig Schroeder, Tamar Shinar, Eftychios Sifakis, Jonathan Su.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Rigids/Rigid_Bodies/RIGID_BODY_COLLECTION.h>
#include <Core/Arrays/INDIRECT_ARRAY.h>
#include <Core/Log/DEBUG_UTILITIES.h>
#include <Core/Log/LOG.h>
#include <Core/Read_Write/FILE_UTILITIES.h>
#include <Core/Utilities/DEBUG_CAST.h>
#include <Core/Utilities/VIEWER_DIR.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT.h>
#include <Geometry/Implicit_Objects/IMPLICIT_OBJECT_TRANSFORMED.h>
#include <Geometry/Meshes/STRUCTURE.h>
#include <Geometry/Meshes/STRUCTURE_LIST.h>
#include <Rigids/Rigid_Bodies/RIGID_BODY.h>
#include <climits>
namespace PhysBAM{
template<class TV>
struct ALLOCATE_BODY_HELPER:public ALLOCATE_HELPER<TV>
{
    RIGID_BODY_COLLECTION<TV>& collection;
    ALLOCATE_BODY_HELPER(RIGID_BODY_COLLECTION<TV>& collection_input):collection(collection_input) {}
    RIGID_BODY<TV>* Create(int index=0) override {return new RIGID_BODY<TV>(collection,true,index);}
    virtual ~ALLOCATE_BODY_HELPER(){}
};
//#####################################################################
// Constructor
//#####################################################################
template<class TV> RIGID_BODY_COLLECTION<TV>::
RIGID_BODY_COLLECTION()
    :rigid_body_particles(*new RIGID_BODY_PARTICLES<TV>()),structure_list(*new STRUCTURE_LIST<TV,int>),always_create_structure(false),
    structure_hash(*new HASHTABLE<std::string,int>),
    allocate_helper(new ALLOCATE_BODY_HELPER<TV>(*this))
{
    if(!allocate_helper) allocate_helper=new ALLOCATE_BODY_HELPER<TV>(*this);
}
//#####################################################################
// Destructor
//#####################################################################
template<class TV> RIGID_BODY_COLLECTION<TV>::
~RIGID_BODY_COLLECTION()
{
    delete &rigid_body_particles;
    delete &structure_hash;
    delete allocate_helper;
    delete &structure_list;
}
//#####################################################################
// Function Rigid_Body
//#####################################################################
template<class TV> RIGID_BODY<TV>& RIGID_BODY_COLLECTION<TV>::
Rigid_Body(const int particle_index)
{
    return *rigid_body_particles.rigid_body(particle_index);
}
//#####################################################################
// Function Rigid_Body
//#####################################################################
template<class TV> const RIGID_BODY<TV>& RIGID_BODY_COLLECTION<TV>::
Rigid_Body(const int particle_index) const
{
    return *rigid_body_particles.rigid_body(particle_index);
}
//#####################################################################
// Function Add_Rigid_Body
//#####################################################################
// structures already added to their respective lists
template<class TV> int RIGID_BODY_COLLECTION<TV>::
Add_Rigid_Body(RIGID_BODY<TV>* rigid_body,const int simplicial_boundary_id,const int implicit_object_id,const int simplicial_interior_id)
{
    int id=rigid_body->particle_index;
    if(simplicial_boundary_id>=0) rigid_body_particles.structure_ids(id)(0)=simplicial_boundary_id;
    if(implicit_object_id>=0) rigid_body_particles.structure_ids(id)(1)=implicit_object_id;
    if(simplicial_interior_id>=0) rigid_body_particles.structure_ids(id)(2)=simplicial_interior_id;
    for(int i=0;i<rigid_body_particles.structure_ids(id).m;i++)
        if(rigid_body_particles.structure_ids(id)(i)>=0 && !structure_list.Element(rigid_body_particles.structure_ids(id)(i)))
            PHYSBAM_FATAL_ERROR();
    return id;
}
//#####################################################################
// Function Add_Rigid_Body
//#####################################################################
// adds body's segmented curve and implicit curve to their respective lists
template<class TV> int RIGID_BODY_COLLECTION<TV>::
Add_Rigid_Body_And_Geometry(RIGID_BODY<TV>* rigid_body)
{
    int id=rigid_body->particle_index;
    assert(rigid_body->structures.m<=3);
    rigid_body_particles.structure_ids(id)=VECTOR<int,3>(-1,-1,-1);
    for(int s=0;s<rigid_body->structures.m;s++) rigid_body_particles.structure_ids(id)(s)=structure_list.Add_Element(rigid_body->structures(s));
    return id;
}
//#####################################################################
// Function Add_Rigid_Body
//#####################################################################
template<class TV> int RIGID_BODY_COLLECTION<TV>::
Add_Rigid_Body(const std::string& basename,const T scaling_factor,const bool read_simplicial_boundary,const bool read_implicit_object,
    const bool read_simplicial_interior,const bool read_rgd_file)
{
    return Add_Rigid_Body(false,basename,scaling_factor,read_simplicial_boundary,read_implicit_object,read_simplicial_interior,read_rgd_file);
}
//#####################################################################
// Function Add_Rigid_Body
//#####################################################################
template<class TV> int RIGID_BODY_COLLECTION<TV>::
Add_Rigid_Body(const bool thin_shell,const std::string& basename,const T scaling_factor,const bool read_simplicial_boundary,const bool read_implicit_object,
    const bool read_simplicial_interior,const bool read_rgd_file)
{
    RIGID_BODY<TV>* rigid_body=new RIGID_BODY<TV>(*this,true);
    rigid_body->thin_shell=thin_shell;

    // rigid body
    std::string rgd=TV::m==2?"rgd2d":"rgd";
    if(read_rgd_file){
        try{Read_From_File(basename+"."+rgd,rigid_body->Mass(),rigid_body->Inertia_Tensor(),rigid_body->Frame());}
        catch(FILESYSTEM_ERROR&){LOG::cout<<"Note: No "<<rgd<<" file for "<<basename<<" (using default values)"<<std::endl;}}
    if(scaling_factor!=1) rigid_body->Rescale(scaling_factor);
    rigid_body->Update_Angular_Velocity();

    int id=Add_Rigid_Body(rigid_body,basename,scaling_factor,read_simplicial_boundary,read_implicit_object,read_simplicial_interior,read_rgd_file);
    return id;
}
//#####################################################################
// Function Add_Rigid_Body
//#####################################################################
template<class TV> int RIGID_BODY_COLLECTION<TV>::
Add_Rigid_Body(RIGID_BODY<TV>* rigid_body,const std::string& basename,const T scaling_factor,
    const bool read_simplicial_boundary,const bool read_implicit_object,const bool read_simplicial_interior,const bool read_rgd_file)
{
    int id=rigid_body->particle_index;

    // structures
    ARRAY<int> structure_ids;
    TV structure_center=rigid_body_particles.frame(id).t;
    if(TV::m==2){
        if(read_simplicial_boundary && !Find_Or_Read_Structure(structure_ids,basename+".curve2d",scaling_factor,structure_center))
            LOG::cout<<"Note: No curve2d file for "<<basename<<std::endl;
        if(read_implicit_object && !Find_Or_Read_Structure(structure_ids,basename+".phi2d",scaling_factor,structure_center))
            LOG::cout<<"Note: No phi2d file for "<<basename<<std::endl;
        if(read_simplicial_interior && !Find_Or_Read_Structure(structure_ids,basename+".tri2d",scaling_factor,structure_center))
            LOG::cout<<"Note: No tri2d file for "<<basename<<std::endl;}
    else{
        if(read_simplicial_boundary && !Find_Or_Read_Structure(structure_ids,basename+".tri",scaling_factor,structure_center))
            LOG::cout<<"Note: No tri file for "<<basename<<std::endl;
        if(read_implicit_object && !Find_Or_Read_Structure(structure_ids,basename+".phi",scaling_factor,structure_center) && 
            !Find_Or_Read_Structure(structure_ids,basename+".oct",scaling_factor,structure_center))
            LOG::cout<<"Note: No phi or oct file for "<<basename<<std::endl;
        if(read_simplicial_interior && !Find_Or_Read_Structure(structure_ids,basename+".tet",scaling_factor,structure_center))
            LOG::cout<<"Note: No tet file for "<<basename<<std::endl;}
    assert(structure_ids.m<=3);
    rigid_body_particles.structure_ids(id)=VECTOR<int,3>(-1,-1,-1);
    for(int i=0;i<structure_ids.m;i++){
        if(structure_ids(i)>=0){
            rigid_body_particles.structure_ids(id)(i)=structure_ids(i);
            Rigid_Body(id).Add_Structure(*structure_list.Element(structure_ids(i)));}}

    return id;
}
//#####################################################################
// Function Update_Angular_Velocity
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Update_Angular_Velocity()
{
    for(int p=0;p<rigid_body_particles.Size();p++) if(Is_Active(p)) Rigid_Body(p).Update_Angular_Velocity();
}
//#####################################################################
// Function Update_Angular_Momentum
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Update_Angular_Momentum()
{
    for(int p=0;p<rigid_body_particles.Size();p++) if(Is_Active(p)) Rigid_Body(p).Update_Angular_Momentum();
}
//#####################################################################
// Function Update_Angular_Velocity
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Update_Angular_Velocity(const ARRAY<int>& particle_indices)
{
    for(int i=0;i<particle_indices.m;i++) Rigid_Body(particle_indices(i)).Update_Angular_Velocity();
}
//#####################################################################
// Function Update_Angular_Momentum
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Update_Angular_Momentum(const ARRAY<int>& particle_indices)
{
    for(int i=0;i<particle_indices.m;i++) Rigid_Body(particle_indices(i)).Update_Angular_Momentum();
}
//#####################################################################
// Function Register_Analytic_Replacement_Structure
//#####################################################################
template<class TV> bool RIGID_BODY_COLLECTION<TV>::
Register_Analytic_Replacement_Structure(const std::string& filename,const T scaling_factor,STRUCTURE<TV>* structure)
{
    std::string hashname=LOG::sprintf("%s@%.6f",filename.c_str(),scaling_factor); // mangle hash name
    if(structure_hash.Contains(hashname)) return false;
    int id=structure?structure_list.Add_Element(structure):0;
    structure_hash.Insert(hashname,id);
    return true;
}
template<class T> void
Wrap_Structure_Helper(STRUCTURE<VECTOR<T,1> >*& structure,const VECTOR<T,1>& center)
{}
template<class TV> void
Wrap_Structure_Helper(STRUCTURE<TV>*& structure,const TV& center)
{   // TODO(jontg): This shouldn't be necessary
}
//#####################################################################
// Function Find_Or_Read_Structure
//#####################################################################
template<class TV> bool RIGID_BODY_COLLECTION<TV>::
Find_Or_Read_Structure(ARRAY<int>& structure_ids,const std::string& filename,const T scaling_factor,const TV& center)
{
    int id;
    if(!File_Exists(filename)) return false;
    std::string hashname=LOG::sprintf("%s@%.6f",filename.c_str(),scaling_factor); // mangle hash name
    if(!always_create_structure&&structure_hash.Get(hashname,id)){ // already read in
        if(!structure_list.Is_Active(id)) PHYSBAM_FATAL_ERROR();} // // only works if the referenced geometry is still in memory
    else{ // read in for the first time
        STRUCTURE<TV>* structure=0;
        structure=STRUCTURE<TV>::Create_From_File(filename);
        if(scaling_factor!=1){
            Wrap_Structure_Helper(structure,center);
            structure->Rescale(scaling_factor);}
        id=structure_list.Add_Element(structure);
        if(!always_create_structure) structure_hash.Insert(hashname,id);}
    structure_ids.Append(id);
    return true;
}
//#####################################################################
// Function Destroy_Unreferenced_Geometry
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Destroy_Unreferenced_Geometry() 
{
    ARRAY<bool> referenced(structure_list.Number_Of_Active_Elements());
    for(int i=0;i<rigid_body_particles.Size();i++) for(int j=0;j<rigid_body_particles.structure_ids(i).m;j++) if(rigid_body_particles.structure_ids(i)(j)>=0)
        referenced(structure_list.Element_Index(rigid_body_particles.structure_ids(i)(j)))=true;
    for(int i=structure_list.Number_Of_Active_Elements()-1;i>=0;i--) if(!referenced(i)) structure_list.Remove_Element(structure_list.Active_Element_Id(i));
}
//#####################################################################
// Function Destroy_Unreferenced_Geometry
//#####################################################################
template<class TV> RIGID_BODY<TV>* RIGID_BODY_COLLECTION<TV>::
New_Body(int index)
{
    return allocate_helper->Create(index);
}
//#####################################################################
// Function Update_Level_Set_Transforms
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Update_Level_Set_Transforms()
{
    for(int i=0;i<rigid_body_particles.Size();i++)
        if(rigid_body_particles.rigid_body(i))
            if(rigid_body_particles.rigid_body(i)->implicit_object)
                rigid_body_particles.rigid_body(i)->implicit_object->transform=&rigid_body_particles.frame(i);
}
//#####################################################################
// Function Read
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Read(const VIEWER_DIR& viewer_dir,ARRAY<int>* needs_init,ARRAY<int>* needs_destroy)
{
    structure_list.Read(viewer_dir,"rigid_body_structure");
    ARRAY<RIGID_BODY<TV>*> bodies(rigid_body_particles.rigid_body);
    rigid_body_particles.rigid_body.Fill(0);
    Read_From_File(viewer_dir.Frame_File("rigid_body_particles"),rigid_body_particles);
    while(rigid_body_particles.rigid_body.m<bodies.m) delete bodies.Pop_Value();
    rigid_body_particles.rigid_body.Prefix(bodies.m)=bodies;

    ARRAY<int> needs_init_default;
    if(needs_init) needs_init->Remove_All();
    if(needs_destroy) needs_destroy->Remove_All();
    char version;int next_id=0;ARRAY<int> active_ids;
    std::string active_list_name=viewer_dir.Common_File("rigid_body_active_ids_list");
    std::string active_name=viewer_dir.Frame_File("rigid_body_active_ids");
    if(File_Exists(active_name)){
        Read_From_File(active_name,version,next_id,active_ids);
        PHYSBAM_ASSERT(version==1);
        if(needs_destroy) for(int i=next_id;i<rigid_body_particles.Size();i++) if(!rigid_body_particles.rigid_body(i)) needs_destroy->Append(i);
        rigid_body_particles.Resize(next_id);}
    else{
        for(int id=0;id<rigid_body_particles.Size();id++) if(Is_Active(id)) active_ids.Append(id);
        next_id=rigid_body_particles.Size();}
    if(rigid_body_particles.rigid_body.Subset(active_ids).Contains(0)){ // don't need to re-read these things if we will not be initializing any newly-active bodies
        std::string key_file_list=viewer_dir.Common_File("rigid_body_key_list");
        std::string key_file=viewer_dir.Frame_File("rigid_body_key");
        char version;
        if(File_Exists(key_file)){
            Read_From_File(key_file,version,rigid_body_particles.structure_ids);
            PHYSBAM_ASSERT(version==2 || version==3);}

        try{
            std::istream* input=Safe_Open_Input_Raw(viewer_dir.Common_File("rigid_body_names"),false);
            int num;
            *input>>num;
            input->ignore(INT_MAX,'\n');
            rigid_body_names.Resize(num);
            for(int i=0;i<rigid_body_names.Size();i++) std::getline(*input,rigid_body_names(i));
            delete input;}
        catch(FILESYSTEM_ERROR&){
            LOG::cerr<<"Did not find rigid body names."<<std::endl;
            rigid_body_names.Clean_Memory();}}

    ARRAY<bool> exists(next_id);exists.Subset(active_ids).Fill(true);
    for(int i=0;i<exists.m;i++) if(!exists(i) && Is_Active(i)) Deactivate_Body(i);
    if(active_ids.m>0){
        for(int i=0;i<active_ids.m;i++){int p=active_ids(i);
            // initialize new rigid body with given id, and initialize geometry
            if(!rigid_body_particles.rigid_body(p)){
                if(needs_init) needs_init->Append(p);
                RIGID_BODY<TV>* rigid_body=New_Body(p);
                if(p<rigid_body_names.Size()) rigid_body->name=rigid_body_names(p);
                for(int s=0;s<rigid_body_particles.structure_ids(p).m;s++)
                    if(rigid_body_particles.structure_ids(p)(s)>=0)
                        rigid_body->Add_Structure(*structure_list.Element(rigid_body_particles.structure_ids(p)(s)));}
            if(Is_Active(p) && rigid_body_particles.structure_ids(p)==VECTOR<int,3>(-1,-1,-1)) Deactivate_Body(p);}}
    Update_Level_Set_Transforms();
}
//#####################################################################
// Function Write
//#####################################################################
template<class TV> void RIGID_BODY_COLLECTION<TV>::
Write(const STREAM_TYPE stream_type,const VIEWER_DIR& viewer_dir) const
{
    structure_list.Write(stream_type,viewer_dir,"rigid_body_structure");
    Write_To_File(stream_type,viewer_dir.Frame_File("rigid_body_particles"),rigid_body_particles);

    // update names
    rigid_body_names.Resize(rigid_body_particles.Size());
    ARRAY<int> active_ids;
    for(int id=0;id<rigid_body_particles.Size();id++) if(Is_Active(id)){active_ids.Append(id);rigid_body_names(id)=Rigid_Body(id).name;}
    if(active_ids.m>0)
        Write_To_File(stream_type,viewer_dir.Frame_File("rigid_body_active_ids"),(char)1,rigid_body_particles.Size(),active_ids);
    std::ostream* output=Safe_Open_Output_Raw(viewer_dir.Common_File("rigid_body_names"),false);
    *output<<rigid_body_names.Size()<<std::endl;
    for(int i=0;i<rigid_body_names.Size();i++) *output<<rigid_body_names(i)<<std::endl;
    delete output;
    char version=3;
    if(rigid_body_particles.structure_ids.m>0)
        Write_To_File(stream_type,viewer_dir.Frame_File("rigid_body_key"),version,rigid_body_particles.structure_ids);
}
//#####################################################################
template class RIGID_BODY_COLLECTION<VECTOR<float,1> >;
template class RIGID_BODY_COLLECTION<VECTOR<float,2> >;
template class RIGID_BODY_COLLECTION<VECTOR<float,3> >;
template class RIGID_BODY_COLLECTION<VECTOR<double,1> >;
template class RIGID_BODY_COLLECTION<VECTOR<double,2> >;
template class RIGID_BODY_COLLECTION<VECTOR<double,3> >;
}
