//#####################################################################
// Copyright 2020, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#ifndef __SYMBOLIC__
#define __SYMBOLIC__

#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/Inverse.h>
#include <Core/Math_Tools/sqr.h>
#include <Core/Vectors/SCALAR_POLICY.h>
#include <cmath>
#include <complex>
#include <iosfwd>
#include <string>
using ::std::sin;
using ::std::cos;
using ::std::tan;
using ::std::exp;
using ::std::log;
using ::std::abs;
using ::std::sqrt;
using ::PhysBAM::sqr;
using ::PhysBAM::cube;
using ::std::hypot;
namespace PhysBAM{

using ::std::sin;
using ::std::cos;
using ::std::tan;
using ::std::asin;
using ::std::acos;
using ::std::atan;
using ::std::atan2;
using ::std::exp;
using ::std::log;
using ::std::abs;
using ::std::pow;
using ::std::sqrt;
using ::PhysBAM::sqr;
using ::PhysBAM::cube;
using ::PhysBAM::Inverse;
using ::std::hypot;
using ::std::floor;
using ::std::ceil;
using ::std::rint;

struct SYMBOLIC_BASE
{
    virtual ~SYMBOLIC_BASE()=default;
    virtual void Set(double s)=0;
    virtual void Print(std::ostream& o,double x) const=0;

    virtual SYMBOLIC_BASE* copy() const=0;
    virtual SYMBOLIC_BASE* pos() const=0;
    virtual SYMBOLIC_BASE* neg() const=0;
    virtual SYMBOLIC_BASE* add(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* sub(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* mul(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* div(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* abs() const=0;
    virtual SYMBOLIC_BASE* inv() const=0;
    virtual SYMBOLIC_BASE* sqrt() const=0;
    virtual SYMBOLIC_BASE* floor() const=0;
    virtual SYMBOLIC_BASE* ceil() const=0;
    virtual SYMBOLIC_BASE* rint() const=0;
    virtual SYMBOLIC_BASE* exp() const=0;
    virtual SYMBOLIC_BASE* log() const=0;
    virtual SYMBOLIC_BASE* hypot(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* pow(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* sin() const=0;
    virtual SYMBOLIC_BASE* cos() const=0;
    virtual SYMBOLIC_BASE* tan() const=0;
    virtual SYMBOLIC_BASE* atan2(const SYMBOLIC_BASE* a) const=0;
    virtual SYMBOLIC_BASE* asin() const=0;
    virtual SYMBOLIC_BASE* acos() const=0;
    virtual SYMBOLIC_BASE* atan() const=0;
};

extern SYMBOLIC_BASE* sym_prototype;

template<class T>
void Register_Symbolic()
{
    sym_prototype=new T;
}

struct SYMBOLIC
{
    double x;
    SYMBOLIC_BASE* z;

    SYMBOLIC(double a=0)
        :x(a),z(sym_prototype->copy())
    {z->Set(a);}

    SYMBOLIC(double a,SYMBOLIC_BASE* s)
        :x(a),z(s)
    {}

    SYMBOLIC(const SYMBOLIC& s)
        :x(s.x),z(s.z->copy())
    {}

    SYMBOLIC(SYMBOLIC&& s)
        :x(s.x),z(s.z)
    {s.z=nullptr;}

    ~SYMBOLIC(){delete z;}

    SYMBOLIC& operator=(const SYMBOLIC& s);

    SYMBOLIC& operator=(SYMBOLIC&& s)
    {
        x=s.x;
        z=s.z;
        s.z=nullptr;
        return *this;
    }

    explicit operator bool() const {return x;}
    explicit operator int() const {return x;}
    explicit operator float() const {return x;}
    explicit operator double() const {return x;}
    double Value() const {return x;}
};

#define INST_OP1(op,name)                               \
    inline SYMBOLIC operator op(const SYMBOLIC& a)      \
    {return {op a.x,a.z->name()};}

#define INST_HELP1(func)                        \
    inline SYMBOLIC func(const SYMBOLIC& a)     \
    {return {func(a.x),a.z->func()};}

#define INST_HELP1n(func,name)                  \
    inline SYMBOLIC func(const SYMBOLIC& a)     \
    {return {func(a.x),a.z->name()};}

#define INST_OP2(op,name)                                               \
    inline SYMBOLIC operator op(const SYMBOLIC& a,const SYMBOLIC& b)    \
    {return {a.x op b.x,a.z->name(b.z)};}                               \
    inline SYMBOLIC operator op(const SYMBOLIC& a,float b)              \
    {return a op SYMBOLIC(b);}                                          \
    inline SYMBOLIC operator op(const SYMBOLIC& a,double b)             \
    {return a op SYMBOLIC(b);}                                          \
    inline SYMBOLIC operator op(const SYMBOLIC& a,int b)                \
    {return a op SYMBOLIC(b);}                                          \
    inline SYMBOLIC operator op(float a,const SYMBOLIC& b)              \
    {return SYMBOLIC(a) op b;}                                          \
    inline SYMBOLIC operator op(double a,const SYMBOLIC& b)             \
    {return SYMBOLIC(a) op b;}                                          \
    inline SYMBOLIC operator op(int a,const SYMBOLIC& b)                \
    {return SYMBOLIC(a) op b;}

#define INST_HELP2(func)                                        \
    inline SYMBOLIC func(const SYMBOLIC& a,const SYMBOLIC& b)   \
    {return {func(a.x,b.x),a.z->func(b.z)};}                    \
    inline SYMBOLIC func(const SYMBOLIC& a,float b)             \
    {return func(a,SYMBOLIC(b));}                               \
    inline SYMBOLIC func(const SYMBOLIC& a,double b)            \
    {return func(a,SYMBOLIC(b));}                               \
    inline SYMBOLIC func(const SYMBOLIC& a,int b)               \
    {return func(a,SYMBOLIC(b));}                               \
    inline SYMBOLIC func(float a,const SYMBOLIC& b)             \
    {return func(SYMBOLIC(a),b);}                               \
    inline SYMBOLIC func(double a,const SYMBOLIC& b)            \
    {return func(SYMBOLIC(a),b);}                               \
    inline SYMBOLIC func(int a,const SYMBOLIC& b)               \
    {return func(SYMBOLIC(a),b);}

#define INST_CMP2(op)                                                   \
    inline bool operator op(const SYMBOLIC& a,const SYMBOLIC& b)        \
    {return a.x op b.x;}                                                \
    inline bool operator op(const SYMBOLIC& a,float b)                  \
    {return a.x op b;}                                                  \
    inline bool operator op(const SYMBOLIC& a,double b)                 \
    {return a.x op b;}                                                  \
    inline bool operator op(const SYMBOLIC& a,int b)                    \
    {return a.x op b;}                                                  \
    inline bool operator op(float a,const SYMBOLIC& b)                  \
    {return a op b.x;}                                                  \
    inline bool operator op(double a,const SYMBOLIC& b)                 \
    {return a op b.x;}                                                  \
    inline bool operator op(int a,const SYMBOLIC& b)                    \
    {return a op b.x;}                                                  \

INST_OP1(+,pos);
INST_OP1(-,neg);
INST_OP2(+,add);
INST_OP2(-,sub);
INST_OP2(*,mul);
INST_OP2(/,div);
INST_CMP2(<);
INST_CMP2(>);
INST_CMP2(<=);
INST_CMP2(>=);
INST_CMP2(==);
INST_CMP2(!=);

inline SYMBOLIC& operator+= (SYMBOLIC& a,const SYMBOLIC& b) {return a=a+b;}
inline SYMBOLIC& operator-= (SYMBOLIC& a,const SYMBOLIC& b) {return a=a-b;}
inline SYMBOLIC& operator*= (SYMBOLIC& a,const SYMBOLIC& b) {return a=a*b;}
inline SYMBOLIC& operator/= (SYMBOLIC& a,const SYMBOLIC& b) {return a=a/b;}
inline SYMBOLIC& operator+= (SYMBOLIC& a,int b) {return a=a+b;}
inline SYMBOLIC& operator+= (SYMBOLIC& a,float b) {return a=a+b;}
inline SYMBOLIC& operator+= (SYMBOLIC& a,double b) {return a=a+b;}
inline SYMBOLIC& operator-= (SYMBOLIC& a,int b) {return a=a-b;}
inline SYMBOLIC& operator-= (SYMBOLIC& a,float b) {return a=a-b;}
inline SYMBOLIC& operator-= (SYMBOLIC& a,double b) {return a=a-b;}
inline SYMBOLIC& operator*= (SYMBOLIC& a,int b) {return a=a*b;}
inline SYMBOLIC& operator*= (SYMBOLIC& a,float b) {return a=a*b;}
inline SYMBOLIC& operator*= (SYMBOLIC& a,double b) {return a=a*b;}
inline SYMBOLIC& operator/= (SYMBOLIC& a,int b) {return a=a/b;}
inline SYMBOLIC& operator/= (SYMBOLIC& a,float b) {return a=a/b;}
inline SYMBOLIC& operator/= (SYMBOLIC& a,double b) {return a=a/b;}

INST_HELP1(abs);
INST_HELP1n(Inverse,inv);
INST_HELP1(sqrt);
INST_HELP1(exp);
INST_HELP1(log);
INST_HELP1(sin);
INST_HELP1(cos);
INST_HELP1(tan);
INST_HELP1(asin);
INST_HELP1(acos);
INST_HELP1(atan);
INST_HELP1(floor);
INST_HELP1(ceil);
INST_HELP1(rint);
INST_HELP2(atan2);
INST_HELP2(hypot);
INST_HELP2(pow);

#undef INST_HELP1
#undef INST_HELP2
#undef INST_OP1
#undef INST_OP2
#undef INST_CMP2

template<> struct SCALAR_POLICY<SYMBOLIC> {typedef SYMBOLIC TYPE;};

template<class RW> void Write_Binary(std::ostream& output,const double& d);
template<class RW> void Read_Binary(std::istream& input,double& d);

template<class RW> inline void
Write_Binary(std::ostream& output,const SYMBOLIC& d)
{Write_Binary<RW>(output,d.x);}

template<class RW> inline void
Read_Binary(std::istream& input,SYMBOLIC& d)
{Read_Binary<RW>(input,d.x);}

inline std::ostream& operator<<(std::ostream& o, const SYMBOLIC& s)
{s.z->Print(o,s.x);return o;}

inline std::istream& operator>>(std::istream& i, SYMBOLIC& s)
{return i>>s.x;}

inline SYMBOLIC
Magnitude_Squared(const SYMBOLIC& s)
{return s*s;}

inline SYMBOLIC
Dot_Product(const SYMBOLIC& a1,const SYMBOLIC& a2)
{return a1*a2;}

inline SYMBOLIC wrap(const SYMBOLIC& a,const SYMBOLIC& b,const SYMBOLIC& c)
{
    if(a<b) return b;
    if(a>c) return c;
    return a;
}
}

#endif
