//#####################################################################
// Copyright 2022, Craig Schroeder.
// This file is part of PhysBAM whose distribution is governed by the license contained in the accompanying file PHYSBAM_COPYRIGHT.txt.
//#####################################################################
#include <Tools/Symbolics/SYMBOLIC_NUMBER.h>
#include <Core/Data_Structures/HASHTABLE.h>
#include <Core/Log/LOG.h>
#include <Core/Math_Tools/cube.h>
#include <Core/Math_Tools/sqr.h>
#include <Core/Utilities/DEBUG_CAST.h>
#include <Core/Vectors/SCALAR_POLICY.h>
#include <Tools/Symbolics/SYMBOLIC.h>
#include <cmath>
#include <complex>
#include <iosfwd>
#include <string>
namespace PhysBAM{

SYMBOLIC_NUMBER::SYMBOLIC_NUMBER(){s="0";}
SYMBOLIC_NUMBER::SYMBOLIC_NUMBER(const std::string& t):s(t){Cache();}

void SYMBOLIC_NUMBER::Set(double x)
{
    if(std::floor(x)!=x && std::abs(x)<1000000 && std::floor(x*128)==x*128)
    {
        for(int d=2;d<=128;d+=d)
        {
            x+=x;
            if(std::floor(x)==x)
            {
                s=LOG::sprintf("(%P/%P)",(int)x,d);
                return;
            }
        }
    }
    if(x<0) s=LOG::sprintf("(%.16P)",x);
    else s=LOG::sprintf("%.16P",x);
}

void SYMBOLIC_NUMBER::Cache()
{
    static int next_id=0;
    static HASHTABLE<std::string,int> hash;
    auto pr=hash.Insert(s,next_id);
    if(pr.second)
    {
        next_id++;
        LOG::printf("X_%d:=%s:\n",*pr.first,s.c_str());
    }
    s=LOG::sprintf("X_%d",*pr.first);
}
    
void SYMBOLIC_NUMBER::Print(std::ostream& o,double x) const
{
    o<<s;
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::copy() const
{
    return new SYMBOLIC_NUMBER(*this);
}
SYMBOLIC_BASE* SYMBOLIC_NUMBER::pos() const
{
    return new SYMBOLIC_NUMBER(*this);
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::add(const SYMBOLIC_BASE* a) const
{
    const auto& b=debug_cast<const SYMBOLIC_NUMBER*>(a)->s;
    if(s=="0") return a->copy();
    if(b=="0") return copy();
    return new SYMBOLIC_NUMBER("("+s+"+"+b+")");
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::sub(const SYMBOLIC_BASE* a) const
{
    const auto& b=debug_cast<const SYMBOLIC_NUMBER*>(a)->s;
    if(b=="0") return copy();
    if(s=="0") return a->neg();
    return new SYMBOLIC_NUMBER("("+s+"-"+b+")");
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::mul(const SYMBOLIC_BASE* a) const
{
    const auto& b=debug_cast<const SYMBOLIC_NUMBER*>(a)->s;
    if(s=="0") return copy();
    if(b=="0") return a->copy();
    if(s=="1") return a->copy();
    if(b=="1") return copy();
    return new SYMBOLIC_NUMBER("("+s+"*"+b+")");
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::div(const SYMBOLIC_BASE* a) const
{
    const auto& b=debug_cast<const SYMBOLIC_NUMBER*>(a)->s;
    if(s=="0") return copy();
    if(b=="1") return copy();
    return new SYMBOLIC_NUMBER("("+s+"/"+b+")");
}

SYMBOLIC_BASE* SYMBOLIC_NUMBER::inv() const
{
    return new SYMBOLIC_NUMBER("(1/"+s+")");
}

#define OP1(o,n) SYMBOLIC_BASE* SYMBOLIC_NUMBER::n() const      \
    {return new SYMBOLIC_NUMBER(#o "("+s+")");}
#define DEF1(f) SYMBOLIC_BASE* SYMBOLIC_NUMBER::f() const       \
    {return new SYMBOLIC_NUMBER(#f "("+s+")");}
#define OP2(o,n) SYMBOLIC_BASE* SYMBOLIC_NUMBER::n(const SYMBOLIC_BASE* a) const \
    {return new SYMBOLIC_NUMBER("("+s+#o+debug_cast<const SYMBOLIC_NUMBER*>(a)->s+")");}
#define DEF2(f) SYMBOLIC_BASE* SYMBOLIC_NUMBER::f(const SYMBOLIC_BASE* a) const \
    {return new SYMBOLIC_NUMBER(#f "("+s+","+debug_cast<const SYMBOLIC_NUMBER*>(a)->s+")");}

OP1(-,neg);
DEF1(abs);
DEF1(sqrt);
DEF1(floor);
DEF1(ceil);
DEF1(rint);
DEF1(exp);
DEF1(log);
DEF2(hypot);
DEF2(pow);
DEF1(sin);
DEF1(cos);
DEF1(tan);
DEF2(atan2);
DEF1(asin);
DEF1(acos);
DEF1(atan);

#undef OP1
#undef DEF1
#undef OP2
#undef DEF2
}
